#ifndef MIE_COMMON_H_
#define MIE_COMMON_H_
/**
	@file mie_common.h
	@brief e틤ʒ`t@C
	@author $Author: shigeo $
	@version $Revision: 1.17 $
	@date $Date: 2007/08/18 16:38:59 $
*/

#if defined(__BORLANDC__) && !defined(WIN32)
#define WIN32
#endif

#if defined(_M_IX86) || defined(__linux__) || defined(__CYGWIN__) || defined(__MINGW32__)
	#define MIE_CPU_I386
	#define MIE_LITTLE_ENDIAN
#elif defined(__MWERKS__) || defined(__APPLE__) /* Macintosh */
	#define MIE_CPU_PPC
	#define MIE_BIG_ENDIAN
#else
	#error "Define CPU type"
#endif

#if !defined(MIE_LITTLE_ENDIAN) && !defined(MIE_BIG_ENDIAN)
	#error "define endian type"
#endif

#ifndef MAX
	#define MAX(x,y) (((x) >= (y)) ? (x) : (y)) /*!< ől擾}N */
	#define MIN(x,y) (((x) <= (y)) ? (x) : (y)) /*!< ŏl擾}N */
#endif

#ifndef MIE_DEFINED_UINT32
	#define MIE_DEFINED_UINT32
	typedef unsigned int uint32;
	typedef signed int sint32;
	typedef unsigned short uint16;
	typedef signed short sint16;
	typedef unsigned char uint8;
	typedef signed char sint8;

#if defined(_MSC_VER) || defined(__BORLANDC__)
	typedef unsigned __int64 uint64;
	typedef __int64 sint64;
	#define MIE_ATOI64(x) _atoi64(x)
	#define MIE_I64 "I64"
#elif defined(__MWERKS__) || defined(__GNUC__) || defined(__CYGWIN__)
	typedef unsigned long long uint64;
	typedef long long sint64;
	#define MIE_ATOI64(x) atoll(x)
	#define MIE_I64 "ll"
#else
	#error "Define int64"
#endif

#endif /* DEFINED_UINT32 */

#if defined(_MSC_VER)
	#ifdef __INTEL_COMPILER // __ICL
		#define MIE_COMPILER_ICC
	#endif
	#if _MSC_VER <= 1200
		#define MIE_COMPILER_VC
		#define MIE_COMPILER_VC6
	#else
		#define MIE_COMPILER_VC
		#define MIE_COMPILER_VC7
	#endif
#endif

#ifndef LE2uint16
#ifdef _MSC_VER
#pragma warning(push)
#pragma warning(disable : 4035) /* ߂l܂ */
#pragma warning(disable : 869) /* gpϐ */

__inline uint16 MIE_SWAP16BIT(uint16 x)
{
	_asm {
		movzx eax, word ptr x;
		bswap eax;
		shr eax, 16;
	}
}
__inline uint32 MIE_SWAP32BIT(uint32 x)
{
	_asm {
		mov eax, x;
		bswap eax;
	}
}

#pragma warning(pop)
#else
	#define MIE_SWAP16BIT(x) ( (((x)<<8)&0xFF00) | (((x)>>8)&0x00FF) ) /*!< 16bit swap */
	#define MIE_SWAP32BIT(x) ( ((x)<<24) | (((x)<<8)&0x00FF0000) | (((x)>>8)&0x0000FF00) | (((x)>>24)&0xFF) ) /*!< 32bit swap */
#endif

#ifdef MIE_LITTLE_ENDIAN
	#define LE2uint16(x) (x) /*!< 16bit little endianւ̕ϊ */
	#define LE2uint32(x) (x) /*!< 32bit little endianւ̕ϊ */
	#define BE2uint16(x) MIE_SWAP16BIT(x) /*!< 16bit big endianւ̕ϊ */
	#define BE2uint32(x) MIE_SWAP32BIT(x) /*!< 32bit big endianւ̕ϊ */
	#define MIE_PACK(x,y,z,w) ((((uint32)(x))<<24)|(((uint32)(y))<<16)|(((uint32)(z))<<8)|((uint32)(w))) /*!< 32bit pack */
#else
	#define BE2uint16(x) (x) /*!< 16bit big endianւ̕ϊ */
	#define BE2uint32(x) (x) /*!< 32bit big endianւ̕ϊ */
	#define LE2uint16(x) MIE_SWAP16BIT(x) /*!< 16bit little endianւ̕ϊ */
	#define LE2uint32(x) MIE_SWAP32BIT(x) /*!< 32bit little endianւ̕ϊ */
	#define MIE_PACK(x,y,z,w) (((uint32)(x))|(((uint32)(y))<<8)|(((uint32)(z))<<16)|(((uint32)(w))<<24)) /*!< 32bit pack */
#endif

#endif /* LE2uint16 */

/* inline macro */
#undef INLINE
#if defined(__GNUC__) || defined(__MWERKS__)
	#define INLINE inline /*!< inline */
	#define STATIC_INLINE static inline /*!< static inline */
#elif defined(_MSC_VER)
	#define INLINE __inline
	#define STATIC_INLINE static __inline
#else
	#define INLINE
	#define STATIC_INLINE static
#endif

#ifdef NDEBUG
	#define PRINTF(x) /*! printf debug */
#else
	#define PRINTF(x) printf x
#endif

#ifdef WIN32
#define SNPRINTF _snprintf /*! snprintf */
#else
#define SNPRINTF snprintf
#endif

#ifdef _MSC_VER
	#pragma warning(disable : 4514) /* inline֐폜 */
	#pragma warning(disable : 4711) /* auto inline */
	#pragma warning(disable : 4786) /* ʎq */
	#pragma warning(disable : 4503) /* O */
	#ifdef MIE_COMPILER_VC6
		#pragma warning(disable : 4127) /* 萔(for "if" trick) */
	#endif
	#if !defined(NOMINMAX) && defined(__cplusplus)
		#pragma message("you should define NOMINMAX for std::min, std::max")
	#endif
#endif
#ifdef __BORLANDC__
	#pragma warn -8027
#endif

/* misc */

/*
	clock 
*/
#ifdef __cplusplus
extern "C" {
#endif

void clkbegin(void);
void clkend(void);
void clkput(void);

#ifdef __cplusplus
}
#endif

#if defined(__cplusplus) && defined(MIE_COMPILER_VC6)
	#ifndef for
		#define for if(0);else for
	#endif
	#if !defined(_STLPORT_VERSION)
		#undef min
		#undef max
		namespace std
		{
			template<class T>
			const T&  min(const T& a, const T& b) { return ( a < b ) ? a : b; }
			template<class T>
			const T&  max(const T& a, const T& b) { return ( a > b ) ? a : b; }
		}
	#endif
#endif

#if 1
	#define MALLOC(x) malloc(x)
	#define FREE(x) free(x)
#else
	#error ""
#endif

#ifndef MIE_DEFINED_ALLOCA
	#define MIE_DEFINED_ALLOCA
	#if defined(_MSC_VER)
		#define ALLOCA_(x) _alloca(x)
	#elif defined(__GNUC__)
		#define ALLOCA_(x) alloca(x)
	#else
		#error "define ALLOCA_"
	#endif
	#define ALLOCA(x) (void *)(((int)ALLOCA_((x) + 16) + 15) & ~15)
	#if defined(WIN32)
		#include <malloc.h>
	#endif
#endif

#define NUM_OF_ARRAY(x) (sizeof(x)/sizeof(x[0]))

#endif /* MIE_COMMON_H_ */
