#ifndef MIE_CPUCLOCK_H_
#define MIE_CPUCLOCK_H_
/**
	@file mie_cpuclock.h
	@brief CPU^C}[𗘗pxvNX
	@author $Author: shigeo $
	@version $Revision: 1.8 $
	@date $Date: 2007/06/22 02:18:25 $
*/
#ifdef _WIN32
	#include <windows.h>
#else
	#include <sys/time.h>
	#include <unistd.h>
#endif
#include "mie_common.h"

extern "C" void MIE_DoBusyLoop(int count); // defined in x86/busyloopa.asm
extern "C" uint64 MIE_GetRDTSC();

namespace MIE {

/**
	CPUNbNNX
*/
class CPUclock {
public:
	CPUclock()
		: prevClock_(0)
		, CPUclockHz_(1)
		, countForUsec_(1)
	{
	}

	/**
		CPUNbN̑
	*/
	void detectCPUclock()
	{
		/*
			Pentium MłidleԂfullɈڍs܂ŎԂ̂
			肷܂0.5b҂Ă݂
		*/
		double totalSec = 0;
		do {
			totalSec += tryDetectCPUclock();
		} while (totalSec < 0.5);
		fprintf(stderr, "CPU %.1fGHz\n", getCPUclockHz() * 1e-9);
	}

	double getCPUclockHz() const { return CPUclockHz_; }

	/**
		rW[EFCg
		@param times [in] ҂
		@note times * usec(1e-6sec)҂
	*/
	inline void busyLoopUsec(int times) const
	{
		MIE_DoBusyLoop(int(countForUsec_ * times));
	}

	/**
		X[v
		@param msec [in] ~bҋ@
	*/
	inline void sleepMsec(int msec) const
	{
#ifdef _WIN32
		::Sleep(msec);
#else
		usleep(msec * 1000);
#endif
	}

	/**
		NbNϐ
	*/
	inline void clear()
	{
		prevClock_ = MIE_GetRDTSC();
	}
	/**
		clear()̌oߎ(nsec)
	*/
	inline double getPassedNsec() const
	{
		return (MIE_GetRDTSC() - prevClock_) * 1.0e9 / CPUclockHz_;
	}

	/**
		oߎ(nsec:Βl݂̂Ӗ)
	*/
	inline double getRelativeNsec() const
	{
		return MIE_GetRDTSC() * 1.0e9 / CPUclockHz_;
	}
private:
	uint64 prevClock_;
	double CPUclockHz_; // CPUNbN(Hz)
	double countForUsec_; // usec(=100*10^(-9)sec)̂߂̃JEg
	double tryDetectCPUclock()
	{
		// usec҂߂̃[v𐔂
		const int testCount = 1000000;
		uint64 begin, end;
		double measureSec;
#ifdef WIN32
		/* Winł͖{]| */
		LARGE_INTEGER freq, b, e;
		if (QueryPerformanceFrequency(&freq) == 0) {
			/* Maybe this cpu is older than normal Pentium */
			return 1;
		}

		QueryPerformanceCounter(&b);
		begin = MIE_GetRDTSC();

		MIE_DoBusyLoop(testCount);

		QueryPerformanceCounter(&e);
		end = MIE_GetRDTSC();
		measureSec = (e.QuadPart - b.QuadPart) / double(freq.QuadPart);
#else
		struct timeval b, e;
		gettimeofday(&b, 0);

		begin = MIE_GetRDTSC();

		MIE_DoBusyLoop(testCount);

		end = MIE_GetRDTSC();
		gettimeofday(&e, 0);
		measureSec = ((e.tv_usec > b.tv_usec) ? e.tv_usec - b.tv_usec : e.tv_usec + 1e6 - b.tv_usec) * 1e-6;
#endif

		CPUclockHz_ = double(end - begin) / measureSec;
		countForUsec_ = testCount / measureSec * 1e-6;
		return measureSec;
	}
};

} // end of namespace

#endif /* MIE_CPUCLOCK_H_ */
