#ifndef MIE_ETHERFRAME_H_
#define MIE_ETHERFRAME_H_
/*!
	@file mie_etherframe.h
	@brief C[Tt[/IPf[^ONX
	@author $Author: shigeo $
	@version $Revision: 1.35 $
	@date $Date: 2007/06/22 02:18:25 $
*/
#include "mie_socket.h"
#include "mie_util.h"
#include <assert.h>

namespace MIE {

enum {
	MAX_PACKET_SIZE = 1500
};

/*!
	bufSize(byte)buf̃`FbNTvZ
	@param buf [in] ̓f[^
	@param bufSize [in] f[^TCY
*/
inline uint16 CalcChecksum(const uint8 *buf, int bufSize)
{
	const uint16 *p = reinterpret_cast<const uint16*>(buf);
	uint32 sum = 0;
	while (bufSize > 1) {
		sum += *p++;
		bufSize -= 2;
	}
	if (bufSize == 1) {
		sum += *reinterpret_cast<const uint8*>(p);
	}
	sum = (sum >> 16) + (sum & 0xFFFF);
	sum += (sum >> 16);
	uint16 ret = static_cast<uint16>(~sum);
	ret = BE2uint16(ret);
	return ret;
}

/*!
	t[̍\̊{NX
*/
class NetworkContainerBase {
protected:
	const uint8* rp_; /* read only */
	uint8* wp_; /* writable */
	int totalSize_;
public:
	NetworkContainerBase()
		: rp_(0)
		, wp_(0)
		, totalSize_(0)
	{
	}

	virtual int getHeaderSize() const = 0;

	/*!
		@param data [in] f[^{
	*/
	void set(const char *data, int totalSize = 0)
	{
		rp_ = reinterpret_cast<const uint8*>(data);
		wp_ = const_cast<uint8*>(rp_);
		totalSize_ = totalSize;
	}

	void copyFrom(const NetworkContainerBase *rhs, int totalSize)
	{
		memcpy(wp_, rhs->rp_, totalSize);
	}

	const char *getData() const { return const_cast<NetworkContainerBase*>(this)->getData(); }
	char *getData() { return reinterpret_cast<char*>(&wp_[getHeaderSize()]); }
};

/*!
	ARPpPbg
*/
class ArpPacket : public NetworkContainerBase {
private:
	mutable MIE::MacAddress srcMacAddress_;
	mutable MIE::MacAddress destMacAddress_;
	mutable MIE::IpAddress srcIpAddress_;
	mutable MIE::IpAddress destIpAddress_;

	enum {
		POS_HW_TYPE = 0,
		POS_PROTOCOL = 2,
		POS_HLEN = 4,
		POS_PLEN = 5,
		POS_OPERATION = 6,
		POS_SRC_MAC_ADDRESS = 8,
		POS_SRC_ADDRESS = 14,
		POS_DEST_MAC_ADDRESS = 18,
		POS_DEST_ADDRESS = 24
	};
	enum {
		HW_TYPE = 0x0001,
		PROTOCOL = 0x0800,
		HLEN = 6,
		PLEN = 4
	};
public:
	enum {
		SIZE = 28
	};
	enum Operation {
		REQUEST = 1,
		RESPONSE = 2,
	};
	ArpPacket()
	{
	}
	/*!
		@param data [in] f[^
	*/
	ArpPacket(const char *data)
	{
		set(data);
	}

	/*!
		wb_TCY擾
	*/
	int getHeaderSize() const { return 0; }
	/*!
		wb_NA
		HW_TYPE = 0x0001,
		PROTOCOL = 0x0800,
		HLEN = 6,
		PLEN = 4,
		OPERATION = 0
	*/
	void clearHeader()
	{
		MIE::Util::set16bitAsBE(HW_TYPE, &wp_[POS_HW_TYPE]);
		MIE::Util::set16bitAsBE(PROTOCOL, &wp_[POS_PROTOCOL]);
		wp_[POS_HLEN] = HLEN;
		wp_[POS_PLEN] = PLEN;
		MIE::Util::set16bitAsBE(0, &wp_[POS_OPERATION]);
	}

	/*!
		`FbN
	*/
	bool isValid() const
	{
		if (MIE::Util::get16bitAsBE(&rp_[POS_HW_TYPE]) != HW_TYPE) {
			DPRINTF(("arp err:HW type=%d\n", MIE::Util::get16bitAsBE(&rp_[POS_HW_TYPE])));
			return false;
		}
		if (MIE::Util::get16bitAsBE(&rp_[POS_PROTOCOL]) != PROTOCOL) {
			DPRINTF(("arp err:protocol=%d\n", MIE::Util::get16bitAsBE(&rp_[POS_PROTOCOL])));
			return false;
		}
		if (rp_[POS_HLEN] != HLEN) {
			DPRINTF(("arp err:hlen=%d\n", rp_[POS_HLEN]));
			return false;
		}
		if (rp_[POS_PLEN] != PLEN) {
			DPRINTF(("arp err:plen=%d\n", rp_[POS_PLEN]));
			return false;
		}
		int ope = MIE::Util::get16bitAsBE(&rp_[POS_OPERATION]);
		if ((ope != REQUEST) && (ope != RESPONSE)) {
			DPRINTF(("arp err:ope=%d\n", ope));
			if ((ope == 3) || (ope == 4)) {
				fprintf(stderr, "not support RAPP\n");
			}
			return false;
		}
		return true;
	}

	/*!
		̕
		@param operation [in] ԍ
	*/
	static const char *toString(Operation ope)
	{
		const char *msg;
		switch (ope) {
		case REQUEST:
			msg = "request";
			break;
		case RESPONSE:
			msg = "response";
			break;
		default:
			msg = "unknown";
			break;
		}
		return msg;
	}

	/*!
		샂[h擾
	*/
	Operation getOperation() const
	{
		return static_cast<Operation>(MIE::Util::get16bitAsBE(&rp_[POS_OPERATION]));
	}

	/*!
		샂[hݒ
	*/
	void setOperation(const Operation& ope)
	{
		MIE::Util::set16bitAsBE(ope, &wp_[POS_OPERATION]);
	}

	/*!
		MMACAhX擾
	*/
	const MacAddress& getDestinationMacAddress() const
	{
		destMacAddress_.set(&rp_[POS_DEST_MAC_ADDRESS]);
		return destMacAddress_;
	}

	/*!
		MMACAhXݒ
	*/
	void setDestinationMacAddress(const MacAddress& mac)
	{
		memcpy(&wp_[POS_DEST_MAC_ADDRESS], mac.get(), MacAddress::LENGTH);
	}

	/*!
		MMACAhX擾
	*/
	const MacAddress& getSourceMacAddress() const
	{
		srcMacAddress_.set(&rp_[POS_SRC_MAC_ADDRESS]);
		return srcMacAddress_;
	}

	/*!
		MMACAhXݒ
	*/
	void setSourceMacAddress(const MacAddress& mac)
	{
		memcpy(&wp_[POS_SRC_MAC_ADDRESS], mac.get(), MacAddress::LENGTH);
	}

	/*! MAhX擾 */
	const IpAddress& getSourceIpAddress() const
	{
		srcIpAddress_.set(ntohl(MIE::Util::get32bitAsBE(&rp_[POS_SRC_ADDRESS])));
		return srcIpAddress_;
	}

	/*!
		MAhXݒ
	*/
	void setSourceIpAddress(const IpAddress& address)
	{
		unsigned long addr = ntohl(address.get().s_addr);
		MIE::Util::set32bitAsBE(addr, &wp_[POS_SRC_ADDRESS]);
	}

	/*! MAhX擾 */
	const IpAddress& getDestinationIpAddress() const
	{
		/* QQQ swap点 */
		destIpAddress_.set(ntohl(MIE::Util::get32bitAsBE(&rp_[POS_DEST_ADDRESS])));
		return destIpAddress_;
	}

	/*!
		MAhXݒ
	*/
	void setDestinationIpAddress(const IpAddress& address)
	{
		unsigned long addr = ntohl(address.get().s_addr);
		MIE::Util::set32bitAsBE(addr, &wp_[POS_DEST_ADDRESS]);
	}
};

/*!
	UDPpPbgNX
*/
class UdpPacket : public NetworkContainerBase {
	enum {
		POS_SRC_PORT = 0,
		POS_DEST_PORT = 2,
		POS_LENGTH = 4,
		POS_CHECKSUM = 6,
		POS_DATA = 8
	};
public:
	UdpPacket()
	{
	}
	/*!
		@param data [in] f[^
		@pram totalSize [in] wb_܂ރTCY
	*/
	UdpPacket(const char* data, int totalSize = 0)
	{
		set(data, totalSize);
	}

	/*! wb_TCY擾 */
	int getHeaderSize() const { return POS_DATA; }

	/*! ]|[gԍ擾 */
	int getSourcePort() const { return MIE::Util::get16bitAsBE(&rp_[POS_SRC_PORT]); }

	/*! ]|[gԍ擾 */
	int getDestinationPort() const { return MIE::Util::get16bitAsBE(&rp_[POS_DEST_PORT]); }

	/*! wb_݂̒(byte)擾 */
	int getTotalSize() const { return MIE::Util::get16bitAsBE(&rp_[POS_LENGTH]); }

	/*! `FbNT擾 */
	uint16 getChecksum() const { return MIE::Util::get16bitAsBE(&rp_[POS_CHECKSUM]); }

	/*! f[^TCY擾 */
	int getDataSize() const { return totalSize_ - POS_DATA; }

	/*!
		]|[gԍݒ
		@param port [in] |[gԍ
	*/
	void setSourcePort(int port)
	{
		MIE::Util::set16bitAsBE(port, &wp_[POS_SRC_PORT]);
	}

	/*!
		]|[gԍݒ
		@param port [in] |[gԍ
	*/
	void setDestinationPort(int port)
	{
		MIE::Util::set16bitAsBE(port, &wp_[POS_DEST_PORT]);
	}

	/*!
		f[^TCYݒ
		@param dataSize [in] f[^TCY(byte)
	*/
	bool setDataSize(int dataSize)
	{
		if (dataSize < 0 || dataSize > MAX_PACKET_SIZE - POS_DATA) {
			return false;
		}
		MIE::Util::set16bitAsBE(dataSize + POS_DATA, &wp_[POS_LENGTH]);
		return true;
	}

	/*!
		`FbNT蓮ݒ
	*/
	void setChecksum(uint16 checksum)
	{
		MIE::Util::set16bitAsBE(checksum, &wp_[POS_CHECKSUM]);
	}

	/*!
		`FbNTݒ
	*/
	void setChecksum(struct in_addr src, struct in_addr dest, uint16 protocol)
	{
		setChecksum(0);
		setChecksum(calcChecksum(src, dest, protocol));
	}

	/*! `FbNTvZ */
	uint16 calcChecksum(struct in_addr src, struct in_addr dest, uint16 protocol) const
	{
		uint32 sum = 0, v;
		v = ntohl(src.s_addr);
//		v = src.s_addr;
		sum += (v & 0xFFFF) + (v >> 16);
		v = ntohl(dest.s_addr);
//		v = dest.s_addr;
		sum += (v & 0xFFFF) + (v >> 16);
//		sum += ntohs(protocol);
//		sum += ntohs(getTotalSize());
		sum += protocol << 8;
		sum += getTotalSize();
		return CalcChecksum(rp_, getTotalSize());
	}
};

/*!
	TCPpPbgNX
*/
class TcpPacket : public NetworkContainerBase {
	enum {
		POS_SRC_PORT = 0, /* 16bit */
		POS_DEST_PORT = 2, /* 16bit */
		POS_SEQ_NUM = 4, /* 32bit */
		POS_ACK_NUM = 8, /* 32bit */
		POS_DATA_OFFSET = 12, /* [data offset:reserved] = [4:4] */
		POS_CONTROL_BIT = 13, /* [reserved:control bits] = [2:6] */
		POS_WINDOW = 14, /* 16bit */
		POS_CHECKSUM = 16, /* 16bit */
	};
public:
	TcpPacket()
	{
	}
	/*!
		@param data [in] f[^
		@pram totalSize [in] wb_܂ރTCY
	*/
	TcpPacket(const char* data, int totalSize = 0)
	{
		set(data, totalSize);
	}

	/*! wb_TCY擾 */
	int getHeaderSize() const { return (rp_[POS_DATA_OFFSET] >> 4) * 4; }

	/*! ]|[gԍ擾 */
	int getSourcePort() const { return MIE::Util::get16bitAsBE(&rp_[POS_SRC_PORT]); }

	/*! ]|[gԍ擾 */
	int getDestinationPort() const { return MIE::Util::get16bitAsBE(&rp_[POS_DEST_PORT]); }

	/*! `FbNT擾 */
	uint16 getChecksum() const { return MIE::Util::get16bitAsBE(&rp_[POS_CHECKSUM]); }

	/*!
		wb_܂܂ȂTCY̎擾
	*/
	int getDataSize() const { return totalSize_ - getHeaderSize(); }
};

/*!
	ICMPpPbgNX
*/
class IcmpPacket : public NetworkContainerBase {
public:
	enum Type {
		ECHO_REPLY = 0,
		DESTINATION_UNREACHABLE = 3,
		SOURCE_QUENCH = 4,
		ECHO_REQUEST = 8
	};

	enum Code {
		UNUSED = 0,

		/*! for Type 3 */
		NET_UNREACHABLE = 0,
		HOST_UNREACHABLE = 1,
		PROTOCOL_UNREACHABLE = 2,
		PORT_UNREACHABLE = 3,
		FRAGMENTATION_NEEDED_AND_DF_SET = 4,
		SOURCE_ROUTE_FAILED = 5
	};

	enum {
		POS_TYPE = 0,
		POS_CODE = 1,
		POS_CHECKSUM = 2,
		POS_DATA = 4,
		POS_MTU = 4,
//		POS_IDENTIFIER = 4,
//		POS_SEQUENCE_NUMBER = 6,
	};

	IcmpPacket()
	{
	}

	/*!
		wb_NA(8byte)
	*/
	void clearHeader()
	{
		memset(wp_, 0, POS_DATA);
	}

	/*!
		^Cv擾
	*/
	Type getType() const { return static_cast<Type>(rp_[POS_TYPE]); }

	/*!
		^Cvݒ
		@param type [in] ^Cv
	*/
	void setType(Type type)
	{
		wp_[POS_TYPE] = static_cast<uint8>(type);
	}

	/*!
		R[h擾
	*/
	Code getCode() const { return static_cast<Code>(rp_[POS_CODE]); }

	/*!
		R[hݒ
		@param code [in] R[h
	*/
	void setCode(Code code)
	{
		wp_[POS_CODE] = static_cast<uint8>(code);
	}

	/*!
		MTUݒ
		@param mtu [in] MTU(max 0xFFFF)
	*/
	bool setMTU(int mtu)
	{
		if (mtu > 0xFFFF) {
			return false;
		}
		MIE::Util::set32bitAsBE(mtu, &wp_[POS_MTU]);
		return true;
	}

	/*!
		
		@param type [in] ^Cv
	*/
	static const char *toString(Type type)
	{
		const char *msg;
		switch (type) {
		case ECHO_REPLY:
			msg = "echo reply";
			break;
		case DESTINATION_UNREACHABLE:
			msg = "destination unreachable";
			break;
		case SOURCE_QUENCH:
			msg = "source quench";
			break;
		case ECHO_REQUEST:
			msg = "echo request";
			break;
		default:
			msg = "unknown";
			break;
		}
		return msg;
	}

	/*!
		`FbNT擾
	*/
	uint16 getChecksum() const
	{
		return MIE::Util::get16bitAsBE(&rp_[POS_CHECKSUM]);
	}
	/*!
		`FbNTvZ
	*/
	uint16 calcChecksum() const
	{
		return CalcChecksum(rp_, totalSize_);
	}

	/*!
		`FbNTݒ
	*/
	void setChecksum()
	{
		MIE::Util::set16bitAsBE(0, &wp_[POS_CHECKSUM]);
		uint16 ret = calcChecksum();
		MIE::Util::set16bitAsBE(ret, &wp_[POS_CHECKSUM]);
	}

	/*!
		f[^TCYݒ(wb_܂)
	*/
	void setTotalSize(int totalSize)
	{
		totalSize_ = totalSize;
	}

	/*!
		f[^TCY擾(wb_܂)
	*/
	int getTotalSize() const { return totalSize_; }

	/*!
		wb_TCY擾
	*/
	int getHeaderSize() const { return POS_DATA; }
};

/*!
	IPf[^ONX
*/
class IpDatagram : public NetworkContainerBase {
private:
	mutable MIE::IpAddress srcIpAddress_;
	mutable MIE::IpAddress destIpAddress_;
	mutable MIE::UdpPacket udp_;
	mutable MIE::IcmpPacket icmp_;
	mutable MIE::TcpPacket tcp_;

	enum IPv4Header {
		POS_VERSION = 0,
		POS_TOS = 1,
		POS_TOTAL_LENGTH = 2,
		POS_IDENTIFICATION = 4,
		POS_FRAGMENT = 6,
		POS_TTL = 8,
		POS_PROTOCOL = 9,
		POS_HEADER_CHECKSUM = 10,
		POS_SRC_ADDRESS = 12,
		POS_DEST_ADDRESS = 16,
		POS_OPTION = 20
	};
	enum {
		DF = 0x4000,
		MF = 0x2000,
		FRAGMENT_OFFSET_MASK = (1 << 13) - 1
	};
	uint16 getFragment() const { return MIE::Util::get16bitAsBE(&rp_[POS_FRAGMENT]); }

public:
	/*!
		vgR
	*/
	enum Protocol {
		ICMP = 1,
		TCP = 6,
		UDP = 17
	};
	enum {
		VERSION_IPV4 = 4, /*! IPV4o[W */
		DEFAULT_TTL = 64 /*! ftHgTTL */
	};

	/*!
		vgR̕
		@param protocol [in] vgR
	*/
	static const char *toString(Protocol protocol)
	{
		const char *msg;
		switch (protocol) {
		case ICMP:
			msg = "icmp";
			break;
		case TCP:
			msg = "tcp";
			break;
		case UDP:
			msg = "udp";
			break;
		default:
			msg = "unknown";
			break;
		}
		return msg;
	}

	IpDatagram()
	{
	}

	/*!
		@param data [in] f[^
		@param dataSize [in] f[^TCY
	*/
	IpDatagram(const char* data, int dataSize = 0)
	{
		set(data, dataSize);
	}

	/*!
		f[^Owb_NA
		ftHg
		version:4(VERSION_IPV4)
		IHL:5
		TOS:0
		total length:0
		identification:0
		fragment:0
		TTL:DEFAULT_TTL
		protocol:UDP
	*/
	void clearHeader()
	{
		setVersion(VERSION_IPV4);
		setIHL(5);
		setTypeOfService(0);
		setIdentification(0);
		clearFragment();
		setTTL(DEFAULT_TTL);
		setProtocol(UDP);
	}

	/* QQQ:GfBAˑ? */
	/*! o[W擾 */
	int getVersion() const { return rp_[POS_VERSION] >> 4; }

	/*!
		o[Wݒ
		@param version [in] o[W
	*/
	void setVersion(int version)
	{
		if (version < 0 || version > 15) return;
		wp_[POS_VERSION] = static_cast<uint8>((version << 4) | getIHL());
	}

	/*!
		IHL(32bitP)擾(0`15)
	*/
	int getIHL() const { return rp_[POS_VERSION] & 0x0F; }

	/*!
		IHL(32bitP)ݒ
		@param ihl [in] ihl
	*/
	void setIHL(int ihl)
	{
		if (ihl < 0 || ihl > 15) return;
		wp_[POS_VERSION] = static_cast<uint8>((getVersion() << 4) | ihl);
	}

	/*! T[rX^Cv̎擾 */
	int getTypeOfService() const { return rp_[POS_TOS]; }

	/*!
		T[rX^Cv̐ݒ
		@param tos [in] ^Cv
	*/
	void setTypeOfService(int tos)
	{
		wp_[POS_TOS] = static_cast<uint8>(tos);
	}


	/*! IvV邩 */
	bool hasOption() const { return getIHL() > 5; }

	/*! ʔԍ擾 */
	int getIdentification() const { return MIE::Util::get16bitAsBE(&rp_[POS_IDENTIFICATION]); }

	/*!
		ʔԍݒ
		@param id [in] ʔԍ
	*/
	void setIdentification(uint16 id)
	{
		MIE::Util::set16bitAsBE(id, &wp_[POS_IDENTIFICATION]);
	}

	/*!
		DF(Don't Fragment)擾
		@retval true ֎~
		@retval false \
	*/
	bool getDF() const { return (getFragment() & DF) != 0; }

	/*!
		DFݒ
		@param df [in] DFtO
	*/
	void setDF(bool df)
	{
		uint16 val = getFragment();
		MIE::Util::set16bitAsBE((val & ~DF) | (df ? DF : 0), &wp_[POS_FRAGMENT]);
	}

	/*!
		MF(More Fragment)擾
		@retval true r̃tOg
		@retval false Ō̃tOg
	*/
	bool getMF() const { return (getFragment() & MF) != 0; }

	/*!
		MFݒ
		@param mf [in] MFtO
	*/
	void setMF(bool mf)
	{
		uint16 val = getFragment();
		MIE::Util::set16bitAsBE((val & ~MF) | (mf ? MF : 0), &wp_[POS_FRAGMENT]);
	}

	/*!
		tOƃtOgItZbg̃NA
	*/
	void clearFragment()
	{
		MIE::Util::set16bitAsBE(0, &wp_[POS_FRAGMENT]);
	}

	/*! tOgItZbg擾(ꂽl8{) */
	int getFragmentOffset() const { return (getFragment() & FRAGMENT_OFFSET_MASK) * 8; }

	/*!
		tOgItZbgݒ
		@param fragmentOffset [in] ItZbg(8̔{ł邱)
	*/
	bool setFragmentOffset(uint16 fragmentOffset)
	{
		if (fragmentOffset > FRAGMENT_OFFSET_MASK) {
			return false;
		}
		if ((fragmentOffset & 7) != 0) {
			return false;
		}
		uint32 val = (getFragmentOffset() & ~FRAGMENT_OFFSET_MASK) | (fragmentOffset / 8);
		MIE::Util::set16bitAsBE(val, &wp_[POS_FRAGMENT]);
		return true;
	}

	/*! TTL擾 */
	int getTTL() const { return rp_[POS_TTL]; }

	/*!
		TTLݒ
		@param ttl [in] TTL
	*/
	void setTTL(uint8 ttl) { wp_[POS_TTL] = ttl; }

	/*! vgR擾 */
	Protocol getProtocol() const { return static_cast<Protocol>(rp_[POS_PROTOCOL]); }

	/*!
		vgRݒ
		@param protocol [in] vgR
	*/
	void setProtocol(Protocol protocol)
	{
		wp_[POS_PROTOCOL] = static_cast<uint8>(protocol);
	}

	/*! wb_`FbNT擾 */
	uint16 getHeaderChecksum() const { return MIE::Util::get16bitAsBE(&rp_[POS_HEADER_CHECKSUM]); }

	/*!
		`FbNTvZ
	*/
	uint16 calcHeaderChecksum() const
	{
		return CalcChecksum(rp_, getHeaderSize());
	}

	/*!
		`FbNTݒ
	*/
	void setHeaderChecksum()
	{
		/* `FbNT0ɂ */
		MIE::Util::set16bitAsBE(0, &wp_[POS_HEADER_CHECKSUM]);
		uint16 sum;
		sum = calcHeaderChecksum();
		MIE::Util::set16bitAsBE(sum, &wp_[POS_HEADER_CHECKSUM]);
	}

	/*!
		`FbNTmF
	*/
	bool isCorrectHeaderChecksum()
	{
		return calcHeaderChecksum() == 0;
	}

	/*! MAhX擾 */
	const IpAddress& getSourceIpAddress() const
	{
		srcIpAddress_.set(ntohl(MIE::Util::get32bitAsBE(&rp_[POS_SRC_ADDRESS])));
		return srcIpAddress_;
	}

	/*!
		MAhXݒ
	*/
	void setSourceIpAddress(const IpAddress& address)
	{
		unsigned long addr = ntohl(address.get().s_addr);
		MIE::Util::set32bitAsBE(addr, &wp_[POS_SRC_ADDRESS]);
	}

	/*! MAhX擾 */
	const IpAddress& getDestinationIpAddress() const
	{
		destIpAddress_.set(ntohl(MIE::Util::get32bitAsBE(&rp_[POS_DEST_ADDRESS])));
		return destIpAddress_;
	}

	/*!
		MAhXݒ
	*/
	void setDestinationIpAddress(const IpAddress& address)
	{
		unsigned long addr = ntohl(address.get().s_addr);
		MIE::Util::set32bitAsBE(addr, &wp_[POS_DEST_ADDRESS]);
	}

	/*!
		wb_TCY擾
	*/
	int getHeaderSize() const { return getIHL() * 4; }
	/*!
		wb_܂ޑŜ̒̎擾
	*/
	int getTotalSize() const { return MIE::Util::get16bitAsBE(&rp_[POS_TOTAL_LENGTH]); }
	/*!
		wb_܂܂ȂTCY̎擾
	*/
	int getDataSize() const { return getTotalSize() - getHeaderSize(); }

	/*!
		wb_TCY̐ݒ
		@param headerSize [in] wb_TCY
	*/
	void setHeaderSize(int headerSize)
	{
		setIHL(headerSize / 4);
	}

	/*!
		wb_܂ޑŜ̒̐ݒ
		@param totalSize [in] Ŝ̒
	*/
	void setTotalSize(int totalSize)
	{
		MIE::Util::set16bitAsBE(totalSize, &wp_[POS_TOTAL_LENGTH]);
	}

	/*!
		wb_܂܂ȂTCY̐ݒ
		@param dataSize [in] f[^TCY
	*/
	void setDataSize(int dataSize)
	{
		setTotalSize(dataSize + getHeaderSize());
	}

	/*!
		UDPpPbg擾(const)
		@note getProtocol() == UDPł邱ƂmF邱
	*/
	const UdpPacket* getUdpPacket() const
	{
		return const_cast<IpDatagram*>(this)->getUdpPacket();
	}

	/*!
		UDPpPbg擾(nonconst)
		@note getProtocol() == UDPł邱ƂmF邱
	*/
	UdpPacket* getUdpPacket()
	{
		udp_.set(getData(), getDataSize());
		return &udp_;
	}

	/*!
		TCPpPbg擾(const)
		@note getProtocol() == TCPł邱ƂmF邱
	*/
	const TcpPacket* getTcpPacket() const
	{
		return const_cast<IpDatagram*>(this)->getTcpPacket();
	}

	/*!
		TCPpPbg擾(nonconst)
		@note getProtocol() == TCPł邱ƂmF邱
	*/
	TcpPacket* getTcpPacket()
	{
		tcp_.set(getData(), getDataSize());
		return &tcp_;
	}
	/*!
		ICMPpPbg擾(const)
	*/
	const IcmpPacket& getIcmpPacket() const
	{
		return const_cast<IpDatagram*>(this)->getIcmpPacket();
	}

	/*!
		ICMPpPbg擾(nonconst)
	*/
	IcmpPacket& getIcmpPacket()
	{
		icmp_.set(getData(), getDataSize());
		return icmp_;
	}
};

/*!
	C[Tt[NX
*/
class EtherFrame : public NetworkContainerBase {
private:
	mutable MIE::MacAddress srcMacAddress_;
	mutable MIE::MacAddress destMacAddress_;
	mutable MIE::IpDatagram ipDatagram_;
	mutable MIE::ArpPacket arp_;

	enum {
		POS_DEST_MAC_ADDRESS = 0,
		POS_SRC_MAC_ADDRESS = 6,
		POS_TYPE = 12,
		POS_IP = 14
	};

public:
	/*!
		t[^Cv
	*/
	enum Type {
		IPv4 = 0x0800,
		ARP = 0x0806,
		IPv6 = 0x06DD
	};
	/*!
		wb_TCY
	*/
	enum {
		HEADER_SIZE = POS_IP
	};

	/*!
		Type̕
		@param type [in] ^Cv
	*/
	static const char* toString(Type type)
	{
		const char *msg;
		switch (type) {
		case IPv4:
			msg = "IPv4";
			break;
		case ARP:
			msg = "ARP";
			break;
		case IPv6:
			msg = "IPv6";
			break;
		default:
			msg = "unknwon";
			break;
		}
		return msg;
	}

	EtherFrame()
	{
	}

	/*!
		@param data [in] f[^
		@pram dataSize [in] TCY
		@note QQQ:̓f[^TCYƓ̃f[^TCY̐`FbNKv
	*/
	EtherFrame(const char* data, int dataSize = 0)
	{
		set(data, dataSize);
	}

	/*!
		wb_t@C擾
	*/
	int getHeaderSize() const { return HEADER_SIZE; }

	/*!
		f[^TCY擾
	*/
	int getDataSize() const
	{
		if (totalSize_ == 0) return 0;
		if (totalSize_ < HEADER_SIZE) {
			return -1;
		}
		return totalSize_ - HEADER_SIZE;
	}

	/*!
		t[wb_NA
		ftHg
		macAhX:0
		type:IPv4
	*/
	void clearHeader()
	{
		memset(wp_, 0, POS_TYPE);
		setType(IPv4);
	}

	/*!
		MMACAhX擾
	*/
	const MacAddress& getDestinationMacAddress() const
	{
		destMacAddress_.set(&rp_[POS_DEST_MAC_ADDRESS]);
		return destMacAddress_;
	}

	/*!
		MMACAhXݒ
	*/
	void setDestinationMacAddress(const MacAddress& mac)
	{
		memcpy(&wp_[POS_DEST_MAC_ADDRESS], mac.get(), MacAddress::LENGTH);
	}

	/*!
		MMACAhX擾
	*/
	const MacAddress& getSourceMacAddress() const
	{
		srcMacAddress_.set(&rp_[POS_SRC_MAC_ADDRESS]);
		return srcMacAddress_;
	}

	/*!
		MMACAhXݒ
	*/
	void setSourceMacAddress(const MacAddress& mac)
	{
		memcpy(&wp_[POS_SRC_MAC_ADDRESS], mac.get(), MacAddress::LENGTH);
	}

	/*!
		^Cv擾
	*/
	Type getType() const { return static_cast<Type>(MIE::Util::get16bitAsBE(&rp_[POS_TYPE])); }

	/*!
		^Cvݒ
	*/
	void setType(Type type)
	{
		MIE::Util::set16bitAsBE(type, &wp_[POS_TYPE]);
	}

	/*!
		IPf[^O擾(const)
	*/
	const IpDatagram& getIpDatagram() const
	{
		ipDatagram_.set(getData(), getDataSize());
		return ipDatagram_;
	}

	/*!
		IPf[^O擾(nonconst)
	*/
	IpDatagram& getIpDatagram()
	{
		ipDatagram_.set(getData(), getDataSize());
		return ipDatagram_;
	}

	/*!
		ARP擾(const)
	*/
	const ArpPacket& getArpPacket() const
	{
		arp_.set(getData(), getDataSize());
		return arp_;
	}

	/*!
		ARP擾(nonconst)
	*/
	ArpPacket& getArpPacket()
	{
		arp_.set(getData(), getDataSize());
		return arp_;
	}
};

} // end of namespace

#endif /* MIE_ETHERFRAME_H_ */
