#ifndef MIE_FILE_H_
#define MIE_FILE_H_
/**
	@file mie_file.h
	@brief t@C̊ȈՃNX
	@author $Author: shigeo $
	@version $Revision: 1.7 $
	@date $Date: 2007/06/22 02:18:25 $
*/
#include "mie_common.h"
#include <string>
#include <cassert>
#ifdef WIN32
#ifndef _WIN32_WINNT
	#define _WIN32_WINNT 0x0500
//	#pragma message("This program is for only Windows 2k/Xp because of SetFilePointerEx")
#endif
#include <windows.h>
#endif

namespace MIE {

/*!
	t@C{NX
	@note ʒ萔ݒp
*/
class FileBase {
public:
	/**
		t@C
	*/
	enum Mode {
		Read			= 0x00000001,
		Write			= 0x00000002,
		ShareRead		= 0x00000010,
		ShareWrite		= 0x00000020,
		OpenExisting	= 0x00000100,
		CreateAlways	= 0x00000200,
		FORBIT_EXTRA_COMMA
	};
	/**
		t@C̈ʒu
	*/
	enum PositionOffset {
		Top,
		Current,
		End
	};
};

#ifdef WIN32

/*!
	t@CNX
*/
class File : public FileBase {
private:
	HANDLE hdl_;
public:
	/**
		RXgN^
	*/
	File()
		: hdl_(INVALID_HANDLE_VALUE)
	{
	}

	/**
		fXgN^
	*/
	~File()
	{
		close();
	}

	/**
		nh擾
		@note gׂł͂Ȃ
	*/
	inline HANDLE getHandle() const { return hdl_; }

	/**
		Ƀt@CJĂ
	*/
	inline bool isValid() const { return hdl_ != INVALID_HANDLE_VALUE; }

	/**
		t@CJ
		@param fName [in] t@C
		@param mode [in] t@C[h
	*/
	bool open(const char *fName, int mode)
	{
		DWORD access = 0, share = 0, disposition = 0;

		if (isValid()) return false;
		if (mode & Read) {
			access |= GENERIC_READ;
		}
		if (mode & Write) {
			access |= GENERIC_WRITE;
		}
		if (mode & ShareRead) {
			share |= FILE_SHARE_READ;
		}
		if (mode & ShareWrite) {
			share |= FILE_SHARE_WRITE;
		}
		if (mode & OpenExisting) {
			disposition = OPEN_EXISTING;
		}
		if (mode & CreateAlways) {
			disposition = CREATE_ALWAYS;
		}
		hdl_ = CreateFile(fName, access, share, NULL, disposition, FILE_ATTRIBUTE_NORMAL, NULL);
		return isValid();
	}

	/**
		ǂݍ݃[hŊJ
		@param inName [in] t@C
	*/
	bool openR(const char *inName)
	{
		return open(inName, Read | OpenExisting);
	}
	/**
		݃[hŊJ
		@param outName [in] t@C
	*/
	bool openW(const char *outName)
	{
		return open(outName, Write | CreateAlways);
	}

	/**
		t@C
	*/
	bool close()
	{
		if (isValid()) {
			bool isOK;
			isOK = CloseHandle(hdl_) != 0;
			hdl_ = INVALID_HANDLE_VALUE;
			return isOK;
		}
		return false;
	}

	/**
		t@C
		@param data [in] ݃f[^
		@param dataSize [in] f[^TCY
		@retval  񂾒l
		@retval  G[
	*/
	int write(const void *data, int dataSize)
	{
		DWORD writeSize;
		int isOK;
		isOK = WriteFile(hdl_, data, dataSize, &writeSize, NULL);
		if (isOK) {
			return writeSize;
		} else {
			return -1;
		}
	}

	/**
		t@Cǂݍ
		@param data [out] ǂݍ݃f[^
		@param dataSize [in] ǂݍݍő咷
		@retval  ǂݍ񂾒l
		@retval  G[
	*/
	int read(void *data, int dataSize)
	{
		DWORD readSize;
		int isOK;
		isOK = ReadFile(hdl_, data, dataSize, &readSize, NULL);
		if (isOK) {
			return readSize;
		} else {
			return -1;
		}
	}

	/**
		t@Cʒu̕ύX
		@param pos [in] t@C̈ʒu
		@param offset [in] t@C̈ʒu
	*/
	bool seek(sint64 pos, PositionOffset offset = Top)
	{
		LARGE_INTEGER largePos;
		largePos.QuadPart = pos;
		DWORD posMode = FILE_BEGIN;
		switch (offset) {
		case Top:
			posMode = FILE_BEGIN;
			break;
		case Current:
			posMode = FILE_CURRENT;
			break;
		case End:
			posMode = FILE_END;
			break;
		}

		return SetFilePointerEx(hdl_, largePos, NULL, offset) != 0;
	}

	/**
		t@C|C^擪ɖ߂
	*/

	bool rewind() { return seek(0); }

	/**
		t@CTCY擾
		@retval ܂0 t@CTCY
		@retval -1 G[
	*/
	sint64 getSize() const
	{
		LARGE_INTEGER size;
		int isOK;
		isOK = GetFileSizeEx(hdl_, &size);
		if (isOK) {
			return size.QuadPart;
		} else {
			return -1;
		}
	}
};

/**
	fBNgTNX
	g
	for (PathIterator itr(path); itr.isValid(); itr.next()) {
		if (itr.isDirectory()) continue;
		printf("filename=%s\n", itr.getFileName());
	}
*/
class PathIterator {
private:
	HANDLE hdl_;
	WIN32_FIND_DATA info_;
	std::string base_;
	mutable std::string name_;

	void close()
	{
		FindClose(hdl_);
		hdl_ = INVALID_HANDLE_VALUE;
	}

public:
	PathIterator(const char *path = 0)
		: hdl_(INVALID_HANDLE_VALUE)
	{
		const int sufLen = 4;
		char tmp[1024 + sufLen];
		if (path && *path) {
			strncpy(tmp, path, sizeof(tmp) - sufLen);
			strcat(tmp, "\\*.*");
			base_ = path;
		} else {
			strcpy(tmp, ".\\*.*");
			base_ = ".";
		}
		hdl_ = FindFirstFile(tmp, &info_);
	}

	~PathIterator()
	{
		if (isValid()) {
			close();
		}
	}

	inline bool isValid() const
	{
		return hdl_ != INVALID_HANDLE_VALUE;
	}

	void next()
	{
		if (isValid()) {
			if (FindNextFile(hdl_, &info_) == 0) {
				close();
			}
		}
	}

	bool isDirectory() const
	{
		return (info_.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) != 0;
	}

	const char* getFileName() const
	{
		if (isValid()) {
			return info_.cFileName;
		}
		return 0;
	}

	const char* getFullName() const
	{
		if (isValid()) {
			name_ = base_;
			name_ += "\\";
			name_ += info_.cFileName;
			return name_.c_str();
		}
		return "";
	}
};

static inline bool IsDirectory(const char *path)
{
	DWORD attr;
	attr = GetFileAttributes(path);
	return (attr & FILE_ATTRIBUTE_DIRECTORY) != 0;
}

#else

#include <stdio.h>
#include <sys/stat.h>

#if defined(__gcc__) || defined(__linux__)

	#if !defined(_FILE_OFFSET_BITS) || (_FILE_OFFSET_BITS != 64)
		#error "add -D_FILE_OFFSET_BITS=64"
	#endif

#elif defined(__MINGW32__)

	#define ftello ftell
	#define fseeko fseek

#else

	#error "not supported"

#endif


/*!
	t@CNX
*/
class File : public FileBase {
private:
	FILE *fp_;
public:
	/**
		RXgN^
	*/
	File()
		: fp_(0)
	{
	}

	/**
		fXgN^
	*/
	~File()
	{
		close();
	}

	/**
		Ƀt@CJĂ
	*/
	inline bool isValid() const { return fp_ != 0; }

	/**
		t@CJ
		@param fName [in] t@C
		@param mode [in] t@C[h
		@note modeRead, WritêݗL
	*/
	bool open(const char *fName, int mode)
	{
		if ((mode & ShareRead) || (mode & ShareWrite)) {
			return false;
		}
		if (isValid()) return false;
		if (mode & Read) {
			fp_ = fopen(fName, "rb");
		}
		if (mode & Write) {
			fp_ = fopen(fName, "wb");
		}
		return isValid();
	}

	/**
		ǂݍ݃[hŊJ
		@param inName [in] t@C
	*/
	bool openR(const char *inName)
	{
		return open(inName, Read | OpenExisting);
	}
	/**
		݃[hŊJ
		@param outName [in] t@C
	*/
	bool openW(const char *outName)
	{
		return open(outName, Write | CreateAlways);
	}

	/**
		t@C
	*/
	bool close()
	{
		if (isValid()) {
			bool isOK;
			isOK = fclose(fp_) == 0;
			fp_ = 0;
			return isOK;
		}
		return false;
	}

	/**
		t@C
		@param data [in] ݃f[^
		@param dataSize [in] f[^TCY
		@retval  񂾒l
		@retval  G[
	*/
	int write(const void *data, int dataSize)
	{
		int writeSize;
		writeSize = fwrite(data, 1, dataSize, fp_);
		return writeSize;
	}

	/**
		t@Cǂݍ
		@param data [out] ǂݍ݃f[^
		@param dataSize [in] ǂݍݍő咷
		@retval  ǂݍ񂾒l
		@retval  G[
	*/
	int read(void *data, int dataSize)
	{
		int readSize;
		readSize = fread(data, 1, dataSize, fp_);
		return readSize;
	}

	/**
		t@Cʒu̕ύX
		@param pos [in] t@C̈ʒu
		@param offset [in] t@C̈ʒu
	*/
	bool seek(sint64 pos, PositionOffset offset = Top)
	{
		int posMode = SEEK_SET;
		switch (offset) {
		case Top:
			posMode = SEEK_SET;
			break;
		case Current:
			posMode = SEEK_CUR;
			break;
		case End:
			posMode = SEEK_END;
			break;
		}
		return fseeko(fp_, pos, offset) == 0;
	}

	/**
		t@C|C^擪ɖ߂
	*/

	bool rewind() { return seek(0); }

	/**
		t@CTCY擾
		@retval ܂0 t@CTCY
		@retval -1 G[
	*/
	sint64 getSize() const
	{
		struct stat stat;
		if (fstat(fileno(fp_), &stat) == 0) {
			return stat.st_size;
		} else {
			return -1;
		}
	}
};

class Path {
private:
	char path_[1024];
public:
	Path()
	{
		*path_ = '\0';
	}
	Path(const char* path)
	{
		strncpy(path_, path, sizeof(path_));
	}
	bool isDirectory() const
	{
		struct stat info;
		if (stat(path_, &info) == 0) {
			return (info.st_mode & S_IFDIR) != 0;
		} else {
			return false;
		}
	}
};

#endif

/**
	eLXgƂĈNX
*/
class FileAsText : public File {
public:
	FileAsText()
	{
	}
	/**
		sǂݍ
		s̏I(0x0D) 0x0A
		@param buf [in] ǂݍݐ
		@param bufSize [in] ǂݍݐTCY(̐ł邱)
		@note őbufSize - 1ǂݍ'\0'^[~l[g
		x
	*/
	bool getLine(char *buf, size_t bufSize)
	{
		assert(bufSize > 0);
		size_t curr = 0;
		while (curr < bufSize - 1) {
			char c;
			if (read(&c, 1) == 0) {
				break;
			}
			if (c == 0x0D) {
				read(&c, 1);
			}
			if (c == 0x0A) {
				break;
			}
			buf[curr++] = c;
		}
		buf[curr] = '\0';
		return curr > 0;
	}
	/**
		s
		@param buf [in] 
		@note s
	*/
	bool putLine(const char *buf)
	{
		int size = strlen(buf);
		if (write(buf, size) == size) {
			char c = 0x0A;
			return write(&c, 1) == 1;
		}
		return false;
	}
};

} // end of namespace

#endif /* MIE_FILE_H_ */
