#ifndef MIE_SOCKET_H_
#define MIE_SOCKET_H_
/**
	@file mie_socket.h
	@brief \Pbg̊ȈՃNX
	@author $Author: shigeo $
	@version $Revision: 1.39 $
	@date $Date: 2007/08/25 10:05:30 $
*/
#include <vector>

/*
	function   macro               win    linux cygwin
	normal     none                 o       o     o
	raw socket MIE_USE_RAW_SOCKET  winpcap  o     x
	+ filter   MIE_USE_PCAP        winpcap pcap   x
	+ gigapcap MIE_USE_GPCAP        x      gpcap  x

	MIE_USE_PCAP is equal to MIE_USE_RAW_SOCKET at win
	writeAsRaw is not supported by MIE_USE_PCAP on linux
*/

#ifdef _WIN32
	#if defined(MIE_USE_PCAP) && !defined(MIE_USE_RAW_SOCKET)
		#define MIE_USE_RAW_SOCKET
	#endif
	#if !defined(MIE_USE_PCAP) && defined(MIE_USE_RAW_SOCKET)
		#define MIE_USE_PCAP
	#endif
#else
	#ifndef MIE_USE_RAW_SOCKET
		#define MIE_USE_RAW_SOCKET // default for linux
	#endif
	#if defined(MIE_USE_GPCAP) && !defined(MIE_USE_PCAP)
		#define MIE_USE_PCAP
	#endif
	#if defined(__CYGWIN__) || defined(__APPLE__)
		#undef MIE_USE_RAW_SOCKET
		#undef MIE_USE_PCAP
	#endif
#endif

#ifdef _WIN32

	#include <winsock2.h>
	#include <ws2tcpip.h> // for socklen_t

	#define S_WRITE(sd, buf, len) send(sd, buf, len, 0)
	#define S_READ(sd, buf, len) recv(sd, buf, len, 0)
	#define S_CLOSE(sd) closesocket(sd)
	#if _MSC_VER > 1200
		#pragma warning(push)
		#pragma warning(disable : 4996) /* sscanf */
	#endif

#elif defined(__unix__) || defined(__GNUC__)

	#include <unistd.h>
	#include <sys/types.h>
	#include <sys/socket.h>
	#include <netinet/in.h>
	#include <netdb.h>
	#include <arpa/inet.h>
	#include <memory.h>

	#define S_WRITE(sd, buf, len) ::write(sd, buf, len)
	#define S_READ(sd, buf, len) ::read(sd, buf, len)
	#define S_CLOSE(sd) ::close(sd)

#endif /* __unix__ */

#define DEBUG_MIE_SOCKET

#ifdef NDEBUG
	#undef DEBUG_MIE_SOCKET
#endif


#ifdef DEBUG_MIE_SOCKET
	#include <stdio.h>
	#define DPRINTF(x) printf x
#else
	#define DPRINTF(x)
#endif

namespace MIE {

/**
	MACAhXNX
*/
class MacAddress {
public:
	enum {
		LENGTH = 6,
		STRING_LENGTH = 17
	};
private:
	unsigned char address_[LENGTH]; /*!< MACAhX */
	mutable char string_[STRING_LENGTH + 1]; /*!< \ */
	mutable bool isUpdated_; /*!< XVtO */
	void update() const
	{
		const unsigned char *p = address_;
		sprintf(string_, "%02X:%02X:%02X:%02X:%02X:%02X", p[0], p[1], p[2], p[3], p[4], p[5]);
		isUpdated_ = false;
	}
public:
	MacAddress()
		: isUpdated_(true)
	{
		memset(address_, 0, sizeof(address_));
		update();
	}
	/**
		MACAhX̐ݒ
		@param address [in] MACAhX(6byte)
	*/
	void set(const unsigned char *address)
	{
		memcpy(address_, address, LENGTH);
		isUpdated_ = true;
	}

	/*!
		MACAhXݒ("00-11-22-33-44-55")
		@note '-'':'ł悢
	*/
	bool set(const char *addressString)
	{
		int len = strlen(addressString);
		if (len != STRING_LENGTH) {
			return false;
		}
		unsigned int t[6];
		int n;
		n = sscanf(addressString, "%02X-%02X-%02X-%02X-%02X-%02X", &t[0],  &t[1],  &t[2],  &t[3],  &t[4],  &t[5]);
		if (n != LENGTH) {
			n = sscanf(addressString, "%02X:%02X:%02X:%02X:%02X:%02X", &t[0],  &t[1],  &t[2],  &t[3],  &t[4],  &t[5]);
			if (n != LENGTH) {
				return false;
			}
		}
		for (int i = 0; i < LENGTH; i++) {
			address_[i] = static_cast<unsigned char>(t[i]);
		}
		isUpdated_ = true;
		return true;
	}

	/** MACAhX擾 */
	const unsigned char *get() const { return address_; }
	/** 擾 */
	const char *toString() const
	{
		if (isUpdated_) {
			update();
		}
		return string_;
	}
	bool operator==(const MacAddress& rhs) const
	{
		return memcmp(address_, rhs.address_, LENGTH) == 0;
	}
	bool operator!=(const MacAddress& rhs) const
	{
		return !(*this == rhs);
	}
};

/**
	IPAhXNX
*/
class IpAddress {
private:
	struct in_addr address_; /*!< IPAhX */
	mutable char string_[16]; /*!< \ */
	mutable bool isUpdated_; /*!< XVtO */

	void update() const
	{
		unsigned char a0, a1, a2, a3;
		unsigned long addr = ntohl(address_.s_addr);
		a0 = (unsigned char)(addr >> 24);
		a1 = (unsigned char)(addr >> 16);
		a2 = (unsigned char)(addr >>  8);
		a3 = (unsigned char)(addr);
		sprintf(string_, "%d.%d.%d.%d", a0, a1, a2, a3);
		isUpdated_ = false;
	}
public:
	IpAddress()
		: isUpdated_(true)
	{
		memset(&address_, 0, sizeof(address_));
	}
	IpAddress(const struct in_addr& address)
	{
		set(address);
	}
	IpAddress(unsigned int addrUint32)
	{
		set(addrUint32);
	}
	IpAddress(const char* address)
	{
		set(address);
	}
	/**
		IPAhX̐ݒ(IPv4AhX\)
		@param address [in] 32rbgIPv4AhX
	*/
	bool set(const struct in_addr& address)
	{
		address_ = address;
		isUpdated_ = true;
		return true;
	}
	/**
		IPAhX̐ݒ(32rbg)
		@param addrUint32 [in] 32rbgl
	*/
	bool set(unsigned int addrUint32)
	{
		address_.s_addr = addrUint32;
		isUpdated_ = true;
		return true;
	}
	/**
		IPAhX̐ݒ(10ihbg:ex. 192.168.100.5)
	*/
	bool set(const char* address)
	{
		bool isOK;
#if 1
		if (!address || !*address) {
			return false;
		} else {
			unsigned int a, b, c, d;
			int n;
			n = sscanf(address, "%d.%d.%d.%d", &a, &b, &c, &d);
			if ((n == 4) && (a < 256) && (b < 256) && (c < 256) && (d < 256)) {
				unsigned int adr = (a << 24) | (b << 16) | (c << 8) | d;
				address_.s_addr = htonl(adr);
				isOK = true;
			} else {
				isOK = false;
			}
		}
#else
		/* only for UNIX */
		isOK = inet_aton(address, &address_) != 0;
#endif
		if (!isOK) {
			struct hostent *host;
			host = gethostbyname(address);
			if (!host) {
				return false;
			}
			memcpy(&address_.s_addr, host->h_addr_list[0], host->h_length);
		}
		isUpdated_ = true;
		return true;
	}
	/** IPAhX擾 */
	const struct in_addr& get() const { return address_; }
	/** 擾 */
	const char *toString() const
	{
		if (isUpdated_) {
			update();
		}
		return string_;
	}

	bool operator==(const IpAddress& rhs) const
	{
		return address_.s_addr == rhs.address_.s_addr;
	}
	bool operator!=(const IpAddress& rhs) const
	{
		return !(*this == rhs);
	}

	/** MACAhX擾 */
	bool getMacAddress(MacAddress& macAddress) const;
};

/**
	lbg[NC^tF[XNX
*/
class NetworkInterface {
friend class Socket;
private:
	int index_; /*!< C^tF[Xԍ */
	char name_[512]; /*!< C^tF[X̖O */
	char description_[512]; /*!< C^tF[X̐ */
	MacAddress macAddress_;
	IpAddress ipAddress_; /*!< IPAhX */
	IpAddress gwIpAddress_; /*!< Q[gEFCAhX */
public:
	NetworkInterface()
		: index_(0)
	{
		*name_ = '\0';
		*description_ = '\0';
	}
	/** C^tF[Xԍ擾 */
	int getIndex() const { return index_; }
	/** C^tF[X擾 */
	const char* getName() const { return name_; }
	/** C^tF[X̃Rg */
	const char* getDescription() const { return description_; }
	/** MACAhX擾 */
	const MacAddress& getMacAddress() const { return macAddress_; }
	/** IPAhX擾 */
	const IpAddress& getIpAddress() const { return ipAddress_; }
	/** ftHgQ[gEFC擾 */
	const IpAddress& getGwIpAddress() const { return gwIpAddress_; }
	/** \ */
	void dump() const
	{
		printf(" name=%s\n", name_);
		if (*description_) {
			printf(" desc=%s\n", description_);
		}
		printf(" IP=%s\n", ipAddress_.toString());
		printf(" MAC=%s\n", macAddress_.toString());
//		printf(" GW=%s\n", gwIpAddress_.toString());
	}
};

//typedef std::vector<MIE::NetworkInterface> NetworkInterfaces;
struct NetworkInterfaces : public std::vector<MIE::NetworkInterface> {
	void dump() const
	{
		for (size_t i = 0; i < size(); i++) {
			printf("index %d ----\n", i);
			(*this)[i].dump();
		}
		printf("---\n");
	}
};

/**
	\PbgNX
*/
class Socket {
private:
	int sd_;
	int listenSd_;
	bool isTCP_;
	bool isServer_;
	struct sockaddr_in addr_;
#ifdef MIE_USE_RAW_SOCKET
	bool isRaw_;

	void *rawFp_;
	/* raw\Pbgpread() */
	int readAsRaw(char *buf, int bufSize);
	/* raw\Pbgpwrite() */
	int writeAsRaw(const char *buf, int bufSize);
	bool closeAsRaw();
#endif

public:
	enum {
		NOERR = 1,
		TIMEOUT = -1,
		CLOSED = -2,
		ERR = -3
	};
	/** \PbgVXe(xĂяo΂悢) */
	static bool initSystem();

	/** \PbgVXeI */
	static bool termSystem();

	Socket()
		: sd_(-1)
		, listenSd_(-1)
		, isTCP_(true)
		, isServer_(false)
#ifdef MIE_USE_RAW_SOCKET
		, isRaw_(false)
		, rawFp_(0)
#endif
	{
		initSystem();
		memset((char *)&addr_, 0, sizeof(addr_));
	}

	~Socket()
	{
		if (listenSd_ != -1) S_CLOSE(listenSd_);
		close();
	}

	bool close()
	{
#ifdef MIE_USE_RAW_SOCKET
		if (isRaw_) {
			return closeAsRaw();
		}
#endif
		if (sd_ == -1) return true;

		S_CLOSE(sd_);
		sd_ = -1;
		return true;
	}

	/**
		f[^
		@param buf [in] obt@
		@param bufSize [in] obt@TCY
		@retval 񂾃TCY
	*/
	int write(const char *buf, int bufSize)
	{
#ifdef MIE_USE_RAW_SOCKET
		if (isRaw_) {
			return writeAsRaw(buf, bufSize);
		}
#endif
		if (isTCP_) {
			return S_WRITE(sd_, buf, bufSize);
		} else {
			return sendto(sd_, buf, bufSize, 0, (struct sockaddr *)&addr_, sizeof(addr_));
		}
	}

	/**
		f[^ǂݍ
		@param buf [out] obt@
		@param bufSize [in] ő发݉\ȃTCY
		@retval 񂾃TCY
	*/
	int read(char *buf, int bufSize)
	{
#ifdef MIE_USE_RAW_SOCKET
		if (isRaw_) {
			return readAsRaw(buf, bufSize);
		}
#endif
		if (isTCP_) {
			return S_READ(sd_, buf, bufSize);
		} else {
			socklen_t addrLen = sizeof(addr_);
			return recvfrom(sd_, buf, bufSize, 0, (struct sockaddr *)&addr_, &addrLen);
		}
	}

	/**
		NCAgƂă|[g
		@param address [in] AhX
		@param port [in] |[gԍ
		@param isTCP [in] trueȂTCP, falseȂUDP
	*/
	bool openAsClient(const char *address, const int port, bool isTCP = true)
	{
		struct hostent *host;

		isTCP_ = isTCP;

		sd_ = socket(AF_INET, isTCP ? SOCK_STREAM : SOCK_DGRAM, 0);
		if (sd_ == -1) {
			return false;
		}
		addr_.sin_family = AF_INET;
		addr_.sin_port = htons((short)port);

		host = gethostbyname(address);
		if (!host) {
			return false;
		}
		memcpy(&addr_.sin_addr, host->h_addr_list[0], host->h_length);

#ifndef NDEBUG
		{
			unsigned char *p = (unsigned char *)&addr_.sin_addr;
			DPRINTF(("address=%d.%d.%d.%d:port=%d\n", p[0], p[1], p[2], p[3], port));
		}
#endif

		if (isTCP) {
			int ret;
			ret = ::connect(sd_, (struct sockaddr *)&addr_, sizeof(addr_));
			if (ret) {
				DPRINTF(("can't connect\n"));
				return false;
			}
		}
		return true;
	}

	/**
		T[o[Ƃă|[g
		@param port [in] |[gԍ
		@param isTCP [in] trueȂTCP, falseȂUDP
	*/
	bool openAsServer(const int port, bool isTCP = true)
	{
		isTCP_ = isTCP;

		sd_ = ::socket(AF_INET, isTCP ? SOCK_STREAM : SOCK_DGRAM, 0);
		if (sd_ == -1) {
			return false;
		}
		addr_.sin_family = AF_INET;
		addr_.sin_port = htons((short)port);

		addr_.sin_addr.s_addr = htonl(INADDR_ANY);

		int ret;
		ret = ::bind(sd_, (struct sockaddr *)&addr_, sizeof(addr_));
		if (ret != 0) return false;

		if (isTCP) {
			ret = ::listen(sd_, 1);
			if (ret) {
				DPRINTF(("can't listen\n"));
				return false;
			}
			// fBXNv^
			listenSd_ = sd_;
			sd_ = -1;
		}
		isServer_ = true;
		return true;
	}

	/**
		҂󂯏(TCPT[op)
	*/
	bool accept()
	{
		if (!isTCP_) return false;
		if (!isServer_) return false;
		close();
		socklen_t addrLen = sizeof(addr_);
		sd_ = ::accept(listenSd_, (struct sockaddr *)&addr_, &addrLen);
		return sd_ >= 0;
	}

#ifdef MIE_USE_RAW_SOCKET
	/**
		RAW socket
	*/
	bool openAsRaw(const NetworkInterface& netIF, int timeoutMsec = 0);
#ifdef MIE_USE_PCAP
	/**
		network packet filter(NPF)
	*/
	bool setFilter(const char *rule);
	int getDropNum() const;
#endif
#endif

	/**
		intpsetsockopt()
	*/
	bool setSocketOption(int optname, int value)
	{
		return setsockopt(sd_, SOL_SOCKET, optname, reinterpret_cast<const char*>(&value), sizeof(value)) == 0;
	}
	/**
		intpgetsockopt()
	*/
	bool getSocketOption(int optname, int* value) const
	{
		socklen_t len = sizeof(*value);
		return getsockopt(sd_, SOL_SOCKET, optname, reinterpret_cast<char*>(value), &len) == 0;
	}
	/**
		Mobt@TCY̎擾
		@retval  obt@TCY(byte)
		@retval -1 G[
	*/
	int getReceiveBufferSize() const
	{
		int val;
		if (getSocketOption(SO_RCVBUF, &val)) {
			return val;
		} else {
			return -1;
		}
	}
	/**
		Mobt@̐ݒ
		@param size [in] obt@TCY(byte)
	*/
	bool setReceiveBufferSize(int size)
	{
		return setSocketOption(SO_RCVBUF, size);
	}
	/**
		Mobt@TCY̎擾
		@retval  obt@TCY(byte)
		@retval -1 G[
	*/
	int getSendBufferSize() const
	{
		int val;
		if (getSocketOption(SO_SNDBUF, &val)) {
			return val;
		} else {
			return -1;
		}
	}
	/**
		Mobt@̐ݒ
		@param size [in] obt@TCY(byte)
	*/
	bool setSendBufferSize(int size)
	{
		return setSocketOption(SO_SNDBUF, size);
	}

private:
#ifdef WIN32
	bool setTimeout(int type, int msec)
	{
		return setSocketOption(type, msec);
	}
	int getTimeout(int type) const
	{
		int val;
		if (getSocketOption(type, &val)) {
			return val;
		} else {
			return -1;
		}
	}
#else
	bool setTimeout(int type, int msec)
	{
		struct timeval t;
		t.tv_sec = msec / 1000;
		t.tv_usec = (msec % 1000) * 1000;
		return setsockopt(sd_, SOL_SOCKET, type, reinterpret_cast<const char*>(&t), sizeof(t)) == 0;
	}
	int getTimeout(int type) const
	{
		struct timeval t;
		socklen_t len = sizeof(t);
		if (getsockopt(sd_, SOL_SOCKET, type, reinterpret_cast<char*>(&t), &len) == 0) {
			return t.tv_sec * 1000 + t.tv_usec / 1000; /* msec */
		} else {
			return -1;
		}
	}
#endif
public:
	/**
		M^CAEg̐ݒ
		@param msec [in] ~b
	*/
	bool setSendTimeout(int msec)
	{
		return setTimeout(SO_SNDTIMEO, msec);
	}
	/**
		M^CAEg̐ݒ
		@param msec [in] ~b
	*/
	bool setReceiveTimeout(int msec)
	{
		return setTimeout(SO_RCVTIMEO, msec);
	}
	/**
		M^CAEg̎擾
	*/
	int getSendTimeout() const
	{
		return getTimeout(SO_SNDTIMEO);
	}
	/**
		M^CAEg̎擾
	*/
	int getReceiveTimeout() const
	{
		return getTimeout(SO_RCVTIMEO);
	}

#ifdef MIE_USE_RAW_SOCKET
	/**
		lbg[NC^tF[X̎擾
		@param ifs [out] C^tF[XXg
	*/
	static bool getNetworkInterfaces(NetworkInterfaces& ifs);
#endif

};

} // end of namespace

#if _MSC_VER > 1200
	#pragma warning(pop)
#endif

#endif /* MIE_SOCKET_H_ */

