#ifndef MIE_THREAD_H_
#define MIE_THREAD_H_
/**
	@file mie_thread.h
	@brief Xbh̊ȈՃNX
	@author $Author: shigeo $
	@version $Revision: 1.23 $
	@date $Date: 2007/06/22 02:18:25 $
	@note LinuxłłCriticalSection == mutex
	QQQ:Ŝ_beginthread()ł͂Ȃ_beginthreadex()gׂ
*/

#include <assert.h>
#ifdef _WIN32
	#include <windows.h>
	#include <process.h>
#else
	#include <unistd.h>
	#include <pthread.h>
	#include <sys/time.h> // for gettimeofday
	#include <stdio.h> // for fprintf
	#include <errno.h>
#endif

namespace MIE {

/*
	ċAmutex̑
	ʏ̓IuWFNgbNXbhēxbNꍇCubN
	邪C̃tOtrueɂƁCJE^₵Ďg̓X[D
*/
const bool isRecursiveMutex = false;

#ifdef _WIN32

	#ifndef _MT
		#error "Compile under multithread mode runtime"
	#endif

	typedef HANDLE mie_thread_t;
	#define MIE_INFINITE INFINITE
	typedef void MIE_THREAD_ENTRY_TYPE(void*);

	/**
		Xbhnh𐶐
		@param hdl [out] Xbhnh
		@param entryFct [in] XbhJnGg
		@param arg [in] [Up[^
	*/
	static inline bool MIE_THREAD_Begin(mie_thread_t& hdl, MIE_THREAD_ENTRY_TYPE entryFct, void *arg)
	{
		int ret = _beginthread(entryFct, 0, arg);
		if (ret == -1) return false;
		hdl = reinterpret_cast<mie_thread_t>(ret);
		return true;
	}

	/**
		XbhI҂
		@param hdl [in] Xbhnh
		@param timeoutMsec [in] I҂(msec)
	*/
	static inline int MIE_THREAD_Join(mie_thread_t hdl, int timeoutMsec)
	{
		return WaitForSingleObject(hdl, timeoutMsec) != WAIT_OBJECT_0;
	}

	/**
		ԑ҂
		@param [in] msec ҂(msec)
	*/
	static inline void MIE_Sleep(int msec)
	{
		Sleep(msec);
	}

	typedef CRITICAL_SECTION mie_cs_t;
	static inline void MIE_CS_Init(mie_cs_t& cs) { InitializeCriticalSection(&cs); }
	static inline void MIE_CS_Lock(mie_cs_t& cs) { EnterCriticalSection(&cs); }
	static inline void MIE_CS_Unlock(mie_cs_t& cs) { LeaveCriticalSection(&cs); }
	static inline void MIE_CS_Term(mie_cs_t& cs) { DeleteCriticalSection(&cs); }

	#define mie_mutex_t HANDLE
	static inline void MIE_MUTEX_Init(mie_mutex_t& mutex)
	{
		if (isRecursiveMutex) {
			mutex = CreateMutex(NULL, FALSE, NULL);
		} else {
			mutex = CreateSemaphore(NULL /* no security */, 1 /* l */, 0x7FFFFFFF /* ől */, NULL /* OȂ */);
		}
	}
	static inline void MIE_MUTEX_Lock(mie_mutex_t& mutex) { WaitForSingleObject(mutex, INFINITE); }
	/*
		^CAEgăbNɎsfalseԂ
		@param msec [in] msec
	*/
	static inline bool MIE_MUTEX_LockTimeout(mie_mutex_t& mutex, int msec)
	{
		DWORD ret = WaitForSingleObject(mutex, msec);
		if (ret == WAIT_OBJECT_0) {
			return true;
		}
		if (ret == WAIT_TIMEOUT) {
			return false;
		}
		/* ret == WAIT_ABANDONED */
		assert(0);
		return false;
	}
	static inline void MIE_MUTEX_Unlock(mie_mutex_t& mutex)
	{
		if (isRecursiveMutex) {
			ReleaseMutex(mutex);
		} else {
			ReleaseSemaphore(mutex, 1, NULL);
		}
	}
	static inline void MIE_MUTEX_Term(mie_mutex_t& mutex) { CloseHandle(mutex); }

	class MIE_EVENT {
		HANDLE event_;
	public:
		MIE_EVENT()
		{
			event_ = ::CreateEvent(NULL, FALSE, FALSE, NULL);
		}
		~MIE_EVENT()
		{
			if (event_) ::CloseHandle(event_);
		}
		inline void send()
		{
			if (event_) SetEvent(event_);
		}
		inline void wait()
		{
			if (event_) WaitForSingleObject(event_, INFINITE);
		}
	};

#elif defined(__linux__) || defined(__CYGWIN__) 

	typedef pthread_t mie_thread_t;
	#define MIE_INFINITE 0x7fffffff
	typedef void* MIE_THREAD_ENTRY_TYPE(void*);

	/**
		Xbhnh𐶐
		@param hdl [out] Xbhnh
		@param entryFct [in] XbhJnGg
		@param arg [in] [Up[^
		@retval 0 I
	*/
	static inline bool MIE_THREAD_Begin(mie_thread_t& hdl, MIE_THREAD_ENTRY_TYPE entryFct, void *arg)
	{
		return pthread_create(&hdl, NULL, entryFct, arg) != 0;
	}

	/**
		XbhI҂
		@param hdl [in] Xbhnh
		@param timeoutMsec [in] I҂(msec)(Linuxł͂̃p[^)
	*/
	static inline int MIE_THREAD_Join(mie_thread_t hdl, int timeoutMsec = MIE_INFINITE)
	{
		if (timeoutMsec != MIE_INFINITE) {
			fprintf(stderr, "not support timeout\n");
			return -1;
		}
		return pthread_join(hdl, NULL);
	}

	/**
		ԑ҂
		@param [in] msec ҂(msec)
	*/
	static inline void MIE_Sleep(int msec)
	{
		usleep(msec * 1000);
	}

	typedef pthread_mutex_t mie_cs_t;
	static inline void MIE_CS_Init(mie_cs_t& cs) { pthread_mutex_init(&cs, NULL); }
	static inline void MIE_CS_Lock(mie_cs_t& cs) { pthread_mutex_lock(&cs); }
	static inline void MIE_CS_Unlock(mie_cs_t& cs) { pthread_mutex_unlock(&cs); }
	static inline void MIE_CS_Term(mie_cs_t& cs) { pthread_mutex_destroy(&cs); }

	typedef pthread_mutex_t mie_mutex_t;
	static inline void MIE_MUTEX_Init(mie_mutex_t& mutex)
	{
		if (isRecursiveMutex) {
			pthread_mutexattr_t mutexattr;
			pthread_mutexattr_init(&mutexattr);
			/* pthreadW */
#ifndef PTHREAD_MUTEX_RECURSIVE_NP
			pthread_mutexattr_settype(&mutexattr, PTHREAD_MUTEX_RECURSIVE);
#else
			pthread_mutexattr_settype(&mutexattr, PTHREAD_MUTEX_RECURSIVE_NP);
#endif
			pthread_mutex_init(&mutex, &mutexattr);
		} else {
			pthread_mutex_init(&mutex, NULL);
		}
	}
	static inline void MIE_MUTEX_Lock(mie_mutex_t& mutex) { pthread_mutex_lock(&mutex); }
	static inline bool MIE_MUTEX_LockTimeout(mie_mutex_t& mutex, int msec)
	{
		timespec absTime;
		absTime.tv_sec = msec / 1000;
		absTime.tv_nsec = msec % 1000;
		return pthread_mutex_timedlock(&mutex, &absTime) == 0;
	}
	static inline void MIE_MUTEX_Unlock(mie_mutex_t& mutex) { pthread_mutex_unlock(&mutex); }
	static inline void MIE_MUTEX_Term(mie_mutex_t& mutex) { pthread_mutex_destroy(&mutex); }

	typedef pthread_cond_t mie_cond_t; 
	static inline void MIE_COND_Init(mie_cond_t *cond) { pthread_cond_init(cond, 0); }
	static inline void MIE_COND_Notify(mie_cond_t *cond) { pthread_cond_signal(cond); }
	static inline bool MIE_COND_Wait(mie_cond_t *cond, mie_mutex_t *mutex) { return pthread_cond_wait(cond, mutex) == 0; }
	static inline bool MIE_COND_TimedWait(mie_cond_t *cond, mie_mutex_t *mutex, int waitMsec) {
		const int oneSec = 1000 * 1000 * 1000;
		struct timespec ts;
		struct timeval now;
		gettimeofday(&now, 0);
		ts.tv_sec = now.tv_sec;
		ts.tv_nsec = (now.tv_usec + waitMsec * 1000) * 1000;
		if (ts.tv_nsec >= oneSec) {
			ts.tv_sec += ts.tv_nsec / oneSec;
			ts.tv_nsec %= oneSec;
		}
		int ret;
		do {
			ret = pthread_cond_timedwait(cond, mutex, &ts);
		} while (ret == EINTR);
		if (ret == 0) return true;
		if (ret == ETIMEDOUT) return false;
		assert(0);
		return false;
	}
	static inline void MIE_COND_Term(mie_cond_t *cond) { pthread_cond_destroy(cond); }

	class MIE_EVENT {
		// PTHREAD_COND apiŎł
	public:
		MIE_EVENT()
		{
			fprintf(stderr, "not implemented\n");
			assert(0);
		}
		~MIE_EVENT()
		{
		}
		inline void send()
		{
		}
		inline void wait()
		{
		}
	};

#else
	#error "not supported"
#endif

/**
	XbhNX
*/
class ThreadBase {
protected:
	mie_thread_t threadHdl_;

	/*
		Xbh[vC^tF[X
	*/
	static void threadLoopIF(void *arg)
	{
		ThreadBase* main = reinterpret_cast<ThreadBase*>(arg);
		main->threadEntry();
	}
public:
	/**
		Xbh̗D揇
	*/
	enum Priority {
		HIGHEST, /* łD揇 */
		HIGH, /* D揇 */
		NORMAL, /* ʏ */
		LOW /* Ⴂ */
	};
	virtual void threadEntry() = 0; //!< XbhJnGg|Cg
	virtual void stopThread() = 0; //!< XbhI

	/**
		RXgN^
	*/
	ThreadBase()
		: threadHdl_(0)
	{
	}
	virtual ~ThreadBase()
	{
	}

	/**
		XbhJn
	*/
	bool beginThread()
	{
		return MIE_THREAD_Begin(threadHdl_, (MIE_THREAD_ENTRY_TYPE*)threadLoopIF, this) == 0;
	}

	/**
		ԑ҂
		@param msec [in] ҂(msec)
	*/
	static inline void sleepMsec(int msec)
	{
		MIE_Sleep(msec);
	}

	/**
		ݎ擾
		@return oߎ(msec)
		@note ݂̂Ӗ(I[o[t[邩Ȃ)
	*/
	static inline unsigned int getCurrTimeMsec()
	{
#ifdef _MSC_VER
		return GetTickCount();
#else
		struct timeval t;
		gettimeofday(&t, 0);
		/* 1000000000sec = 31 years */
		return (t.tv_sec - 1000000000) * 1000 + (t.tv_usec / 1000);
#endif
	}

	/**
		Xbh̏I҂
		@param waitMsec [in] ^CAEg(ftHg)
	*/
	bool joinThread(int waitMsec = MIE_INFINITE)
	{
		if (threadHdl_) {
			if (MIE_THREAD_Join(threadHdl_, waitMsec)) {
				return false;
			}
		}
		return true;
	}
	/**
		Xbh̗D揇ʂύX
		@param priority [in] D揇
	*/
	bool setThreadPriority(Priority priority)
	{
#ifdef WIN32
		int priorityValue;
		HANDLE hdl = GetCurrentThread();
		switch (priority) {
		case HIGHEST:
			priorityValue = THREAD_PRIORITY_HIGHEST;
		case HIGH:
			priorityValue = THREAD_PRIORITY_ABOVE_NORMAL;
			break;
		case NORMAL:
			priorityValue = THREAD_PRIORITY_NORMAL;
			break;
		case LOW:
			priorityValue = THREAD_PRIORITY_BELOW_NORMAL;
			break;
		default:
			return false;
		}
		SetThreadPriority(hdl, priorityValue);
#else
		fprintf(stderr, "not support setThreadPriority\n");
#endif
		return true;
	}
	/**
		vZX̗D揇ʂύX
	*/
	static inline bool setProcessPriority(Priority priority)
	{
#ifdef WIN32
		DWORD priorityClass;
		HANDLE hdl = GetCurrentProcess();
		switch (priority) {
		case HIGHEST:
			priorityClass = HIGH_PRIORITY_CLASS;
		case HIGH:
			priorityClass = ABOVE_NORMAL_PRIORITY_CLASS;
			break;
		case NORMAL:
			priorityClass = NORMAL_PRIORITY_CLASS;
			break;
		case LOW:
			priorityClass = BELOW_NORMAL_PRIORITY_CLASS;
			break;
		default:
			return false;
		}
		SetPriorityClass(hdl, priorityClass);
#else
		fprintf(stderr, "not support setProcessPriority\n");
#endif
		return true;
	}
};

/**
	yrNX
*/
class CriticalSection {
public:
	CriticalSection()
	{
		MIE_CS_Init(cs_);
	}
	~CriticalSection()
	{
		MIE_CS_Term(cs_);
	}
	inline void lock()
	{
		MIE_CS_Lock(cs_);
	}
	inline void unlock()
	{
		MIE_CS_Unlock(cs_);
	}
	inline mie_cs_t *get()
	{
		return &cs_;
	}
private:
	CriticalSection(const CriticalSection&); /* Rs[Ȃ */
	CriticalSection& operator=(const CriticalSection&);
	mie_cs_t cs_;
};

/**
	mutex(^CAEg)
*/
class Mutex {
public:
	Mutex()
	{
		MIE_MUTEX_Init(mutex_);
	}
	~Mutex()
	{
		MIE_MUTEX_Term(mutex_);
	}
	inline void lock()
	{
		MIE_MUTEX_Lock(mutex_);
	}
	inline bool lockTimeout(int msec)
	{
		return MIE_MUTEX_LockTimeout(mutex_, msec);
	}
	inline void unlock()
	{
		MIE_MUTEX_Unlock(mutex_);
	}
	inline mie_mutex_t *get()
	{
		return &mutex_;
	}
private:
	Mutex(const Mutex&); /* Rs[Ȃ */
	Mutex& operator=(const Mutex&);
	mie_mutex_t mutex_;
};

#ifdef _MSC_VER

/**
	qIȉZ
*/
inline int AtomicAdd(int* x, int y)
{
	return InterlockedExchangeAdd((long*)x, y);
}

inline int AtomicInc(int* x)
{
	return InterlockedIncrement((long*)x);
}

#else

/**
	qIȉZ
	QQQ:asmŎȂcsKvȂ̂ňӖȂ
*/
inline int AtomicAdd(CriticalSection& cs, int* x, int y)
{
	int ret = *x;
	cs.lock();
	*x += y;
	cs.unlock();
	return ret;
}

inline int AtomicInc(CriticalSection& cs, int* x)
{
	return AtomicAdd(cs, x, 1);
}

#endif

/**
	bN
*/
template<class T>
class AutoLockT {
public:
	AutoLockT(T &cs)
		: cs_(cs)
	{
		cs_.lock();
	}
	~AutoLockT()
	{
		cs_.unlock();
	}
private:
	T& cs_;
	AutoLockT& operator=(const AutoLockT&);
};

#ifdef __linux__
class CondVar {
public:
	CondVar()
	{
		MIE_COND_Init(&cond_);
	}
	~CondVar()
	{
		MIE_COND_Term(&cond_);
	}
	inline void notify()
	{
		MIE_COND_Notify(&cond_);
	}
	inline bool wait(MIE::Mutex& mutex)
	{
		return MIE_COND_Wait(&cond_, mutex.get());
	}
	inline bool waitTimeout(MIE::Mutex& mutex, int waitMsec)
	{
		return MIE_COND_TimedWait(&cond_, mutex.get(), waitMsec);
	}
private:
	CondVar(const CondVar&); /* don't copy */
	CondVar& operator=(const CondVar&);
	mie_cond_t cond_;
};
#endif

typedef AutoLockT<CriticalSection> AutoLock; /*!< auto lock critical section */

/**
	bN(^CAEg)
*/
class AutoLockTimeout {
public:
	AutoLockTimeout(Mutex& mutex, int msec = MIE_INFINITE)
		: mutex_(mutex)
	{
		if (msec == MIE_INFINITE) {
			isLocked_ = true;
			mutex_.lock();
		} else {
			isLocked_ = mutex_.lockTimeout(msec);
		}
	}
	~AutoLockTimeout()
	{
		if (isLocked_) mutex_.unlock();
	}
	bool isLocked() const { return isLocked_; }
private:
	Mutex& mutex_;
	bool isLocked_;
	AutoLockTimeout& operator=(const AutoLockTimeout&);
};

} // end of namespace

#endif /* MIE_THREAD_H_ */

