#ifndef MIE_UTIL_H_
#define MIE_UTIL_H_
/**
	@file mie_util.h
	@brief 悭g֐Q
	@author $Author: shigeo $
	@version $Revision: 1.17 $
	@date $Date: 2007/06/22 02:18:25 $
*/
#include "mie_common.h"
#include <stdlib.h> /* for atoi */
#include <string.h> /* for strcmp */

namespace MIE {

namespace Util {

/**
	32bitlittle endianƂăɊi[
	@param src [in] 32bit
	@param dest [out] ւ̃|C^
*/
inline void set32bitAsLE(uint32 src, void *dest)
{
	uint8 *p = reinterpret_cast<uint8*>(dest);
	p[0] = uint8(src);
	p[1] = uint8(src >> 8);
	p[2] = uint8(src >> 16);
	p[3] = uint8(src >> 24);
}

/**
	32bitbig endianƂăɊi[
	@param src [in] 32bit
	@param dest [out] ւ̃|C^
*/
inline void set32bitAsBE(uint32 src, void *dest)
{
	uint8 *p = reinterpret_cast<uint8*>(dest);
	p[0] = uint8(src >> 24);
	p[1] = uint8(src >> 16);
	p[2] = uint8(src >> 8);
	p[3] = uint8(src);
}

/**
	little endianƂ32bitǂݍ
	@param src [in] ւ̃|C^
*/
inline uint32 get32bitAsLE(const void *src)
{
	const uint8 *p = reinterpret_cast<const uint8*>(src);
	return p[0] | (p[1] << 8) | (p[2] << 16) | (p[3] << 24);
}

/**
	big endianƂ32bitǂݍ
	@param src [in] ւ̃|C^
*/
inline uint32 get32bitAsBE(const void *src)
{
	const uint8 *p = reinterpret_cast<const uint8*>(src);
	return p[3] | (p[2] << 8) | (p[1] << 16) | (p[0] << 24);
}

/**
	16bitlittle endianƂăɊi[
	@param src [in] 16bit
	@param dest [out] ւ̃|C^
*/
inline void set16bitAsLE(uint32 src, void *dest)
{
	uint8 *p = reinterpret_cast<uint8*>(dest);
	p[0] = uint8(src);
	p[1] = uint8(src >> 8);
}

/**
	16bitbig endianƂăɊi[
	@param src [in] 16bit
	@param dest [out] ւ̃|C^
*/
inline void set16bitAsBE(uint32 src, void *dest)
{
	uint8 *p = reinterpret_cast<uint8*>(dest);
	p[0] = uint8(src >> 8);
	p[1] = uint8(src);
}

/**
	little endianƂ16bitǂݍ
	@param src [in] ւ̃|C^
*/
inline uint16 get16bitAsLE(const void *src)
{
	const uint8 *p = reinterpret_cast<const uint8*>(src);
	return uint16(p[0] | (p[1] << 8));
}

/**
	big endianƂ16bitǂݍ
	@param src [in] ւ̃|C^
*/
inline uint16 get16bitAsBE(const void *src)
{
	const uint8 *p = reinterpret_cast<const uint8*>(src);
	return uint16(p[1] | (p[0] << 8));
}

/**
	͊NX
	ƓK
	@note g

	bool analyze(int argc, const char *const argv[])
	{
		set(argc, argv);

		while (hasOption()) {
			if (isMatch("-width", 1)) {
				width_ = getInteger();
			} else
			if (isMatch("-tcp", 0)) {
				isTCP_ = true;
			} else
			{
				usage();
				return false;
			}
			next();
		}
		...
	}
*/
class OptionBase {
private:
	int argc_;
	const char *const* argv_;

protected:
	virtual void usage() = 0;

	inline void next()
	{
		argc_--, argv_++;
	}
	inline const char *querryNextString() const
	{
		if (argc_ > 1) {
			return argv_[1];
		} else {
			return 0;
		}
	}
	inline int getArgc() const { return argc_; }
	inline bool hasOption() const
	{
		return argc_ > 0;
	}
	/**
		IvVɃ}b`
		@param option [in] ׂIvV(:"-width")
		@param optNum [in] IvVƂ̐
		@note optNum > 0 ̂Ƃ͊enext()(optNum - 1)ĂԂ
	*/
	inline bool isMatch(const char *option, int optNum = 0)
	{
		if (argc_ > optNum && strcmp(*argv_, option) == 0) {
			if (optNum > 0) {
				next();
			}
			return true;
		}
		return false;
	}
	inline const char *getString() const { return *argv_; }
	inline int getInteger() const { return atoi(*argv_); }

	void set(int argc, const char *const* argv)
	{
		argc_ = argc;
		argv_ = argv;
		next();
	}

public:
	OptionBase()
		: argc_(0)
		, argv_(0)
	{
	}
	virtual ~OptionBase()
	{
	}
	virtual bool analyze(int argc, const char *const argv[]) = 0;
};

} // end of namespace

/**
	2i=>10iϊev[g
	#define B(x) MIE::Util::binary<x>::value
	݂ɂĎgĂ
*/
template<int n>
struct binary {
	enum {
		bit = n % 10,
		value = bit + (binary<n / 10>::value << 1)
	};
};

template<>
struct binary<0> {
	enum {
		bit = 0,
		value = 0
	};
};

} // end of namespace

#endif /* MIE_UTIL_H_ */

