/**
	ARPpPbgeXg
	ARPԓeXg
*/
#include <stdio.h>
#include <ctype.h>
#include <string>
#include "../mie_common.h"
#include "../mie_socket.h"
#include "../mie_etherframe.h"
#include "rawsocket_common.h"

int main(int argc, char *argv[])
{
	MIE::Socket::initSystem();

	/* get all network interfaces */
	MIE::NetworkInterfaces ifs;
	if (MIE::Socket::getNetworkInterfaces(ifs)) {
		ifs.dump();
	} else {
		fprintf(stderr, "getNetworkInterface error\n");
		return 1;
	}

	int idx = -1; /* no select */
	std::string ipStr;
	std::string macStr;

	// get option
	argc--, argv++;
	while (argc > 0) {
		if (argc > 1 && strcmp(*argv, "-idx") == 0) {
			argc--, argv++;
			idx = atoi(*argv);
		} else
		if (argc > 1 && strcmp(*argv, "-ip") == 0) {
			argc--, argv++;
			ipStr = *argv;
		} else
		if (argc > 1 && strcmp(*argv, "-mac") == 0) {
			argc--, argv++;
			macStr = *argv;
		} else
		{
			fprintf(stderr, "arp_test [-idx num] [-ip ip] [-mac mac]\n");
			fprintf(stderr, "specify mac/ip if you want fake answer.\n");
			return 1;
		}
		argc--, argv++;
	}

	if (idx == -1) {
		if (ifs.size() > 1) {
			printf("select network interface(0-%d)\n", ifs.size() - 1);
			scanf("%d", &idx);
		} else {
			idx = 0;
		}
	}

	/* verify idx value */
	if (idx >= static_cast<int>(ifs.size())) {
		fprintf(stderr, "select error\n");
		return 1;
	}

	MIE::Socket s;
	if (!s.openAsRaw(ifs[idx])) {
		fprintf(stderr, "can't open socket\n");
		return 1;
	}

	MIE::MacAddress srcMac;
	MIE::IpAddress srcIp;

	if (macStr.empty()) {
		srcMac = ifs[idx].getMacAddress();
	} else {
		srcMac.set(macStr.c_str());
	}
	if (ipStr.empty()) {
		srcIp = ifs[idx].getIpAddress();
	} else {
		srcIp.set(ipStr.c_str());
	}

	printf("src=%s(%s)\n", srcIp.toString(), srcMac.toString());

	char outBuf[256];
	MIE::EtherFrame outFrame(outBuf, sizeof(outBuf));
	outFrame.clearHeader();
	outFrame.setType(MIE::EtherFrame::ARP);
	outFrame.setSourceMacAddress(srcMac);
	MIE::ArpPacket& outArp = outFrame.getArpPacket();
	outArp.clearHeader();
	outArp.setOperation(MIE::ArpPacket::RESPONSE);
	outArp.setSourceIpAddress(srcIp);
	outArp.setSourceMacAddress(srcMac);

	for (;;) {
		char buf[2048]; /* 1packet\傫Ă */
		int readSize;
		readSize = s.read(buf, sizeof(buf));
		if (readSize <= 0) {
			fprintf(stderr, "can't read\n");
			continue;
		}
		MIE::EtherFrame frame(buf, readSize);

		if (frame.getType() != MIE::EtherFrame::ARP) {
			continue;
		}
		const MIE::ArpPacket& inArp = frame.getArpPacket();
		if (!inArp.isValid()) {
			printf(" illegal ARP\n");
			continue;
		}
		printf("receive ARP\n");
		printf(" Operation=%s\n", MIE::ArpPacket::toString(inArp.getOperation()));
		printf(" src MAC=%s, dest MAC=%s\n", inArp.getSourceMacAddress().toString(), inArp.getDestinationMacAddress().toString());
		printf(" src ip=%s, dest ip=%s\n", inArp.getSourceIpAddress().toString(), inArp.getDestinationIpAddress().toString());
		if (inArp.getOperation() != MIE::ArpPacket::REQUEST) {
			continue;
		}
		if (inArp.getDestinationIpAddress() != srcIp) {
			continue;
		}
		/* send ARP */
		outFrame.setDestinationMacAddress(inArp.getSourceMacAddress());
		outArp.setDestinationIpAddress(inArp.getSourceIpAddress());
		outArp.setDestinationMacAddress(inArp.getSourceMacAddress());
		printf("send ARP\n");
		const int outBufSize = MIE::ArpPacket::SIZE + MIE::EtherFrame::HEADER_SIZE;
		s.write(outBuf, outBufSize);
		dump(outBuf, outBufSize);
//
		if (0) {
			MIE::EtherFrame f(outBuf, outBufSize);
			printf("  mac dest=%s, src=%s\n", f.getDestinationMacAddress().toString(), f.getSourceMacAddress().toString());
			MIE::ArpPacket& a = f.getArpPacket();
			printf("  src ip=%s, mac=%s\n", a.getSourceIpAddress().toString(), a.getSourceMacAddress().toString());
			printf("  dst ip=%s, mac=%s\n", a.getDestinationIpAddress().toString(), a.getDestinationMacAddress().toString());
		}
///
	}
	return 0;
}
