/**
	ICMPpPbgeXg
	ICMPԓeXg
*/
#include <stdio.h>
#include <ctype.h>
#include <string>
#include <assert.h>
#include "../mie_common.h"
#include "../mie_socket.h"
#include "../mie_etherframe.h"
#include "rawsocket_common.h"

#define DEFAULT_MTU 1500

int main(int argc, char *argv[])
{
	MIE::Socket::initSystem();

	/* get all network interfaces */
	MIE::NetworkInterfaces ifs;
	if (MIE::Socket::getNetworkInterfaces(ifs)) {
		ifs.dump();
	} else {
		fprintf(stderr, "getNetworkInterface error\n");
		return 1;
	}

	int idx = -1; /* no select */
	int mtu = -1;
	std::string ipStr;
	std::string macStr;

	// get option
	argc--, argv++;
	while (argc > 0) {
		if (argc > 1 && strcmp(*argv, "-idx") == 0) {
			argc--, argv++;
			idx = atoi(*argv);
		} else
		if (argc > 1 && strcmp(*argv, "-mtu") == 0) {
			argc--, argv++;
			mtu = atoi(*argv);
		} else
		if (argc > 1 && strcmp(*argv, "-ip") == 0) {
			argc--, argv++;
			ipStr = *argv;
		} else
		if (argc > 1 && strcmp(*argv, "-mac") == 0) {
			argc--, argv++;
			macStr = *argv;
		} else
		{
			fprintf(stderr, "icmp_test [-idx num] [-mtu mtu] [-ip ip] [-mac mac]\n");
			fprintf(stderr, "specify mac/ip if you want fake answer.\n");
			return 1;
		}
		argc--, argv++;
	}

	if (idx == -1) {
		if (ifs.size() > 1) {
			printf("select network interface(0-%d)\n", ifs.size() - 1);
			scanf("%d", &idx);
		} else {
			idx = 0;
		}
	}

	/* verify idx value */
	if (idx >= static_cast<int>(ifs.size())) {
		fprintf(stderr, "select error\n");
		return 1;
	}

//	if ((mtu & 7) == 0) {
//		fprintf(stderr, "MTU must be the muptiple of 8(now mtu=%d)\n", mtu);
//		return 1;
//	}
	if (mtu <= 0) {
		mtu = DEFAULT_MTU;
	}

	MIE::Socket s;
	if (!s.openAsRaw(ifs[idx])) {
		fprintf(stderr, "can't open socket\n");
		return 1;
	}

	MIE::MacAddress srcMac;
	MIE::IpAddress srcIp;

	if (macStr.empty()) {
		srcMac = ifs[idx].getMacAddress();
	} else {
		srcMac.set(macStr.c_str());
	}
	if (ipStr.empty()) {
		srcIp = ifs[idx].getIpAddress();
	} else {
		srcIp.set(ipStr.c_str());
	}

	printf("src=%s(%s)\n", srcIp.toString(), srcMac.toString());
	printf("MTU=%d\n", mtu);

	char outBuf[2048];
	MIE::EtherFrame outFrame(outBuf);
	outFrame.clearHeader();
	outFrame.setSourceMacAddress(srcMac);
	MIE::IpDatagram& outPacket = outFrame.getIpDatagram();
	outPacket.clearHeader();
	outPacket.setDF(true);
	outPacket.setProtocol(MIE::IpDatagram::ICMP);
	outPacket.setSourceIpAddress(srcIp);
	MIE::IcmpPacket& outIcmp = outPacket.getIcmpPacket();
	outIcmp.clearHeader();

	for (;;) {
		char buf[2048]; /* 1packet\傫Ă */
		int readSize;
		readSize = s.read(buf, sizeof(buf));
		if (readSize <= 0) {
			fprintf(stderr, "can't read\n");
			continue;
		}
		MIE::EtherFrame inFrame(buf, readSize);

		if (inFrame.getType() != MIE::EtherFrame::IPv4) {
			continue;
		}

		const MIE::IpDatagram& inPacket = inFrame.getIpDatagram();
		if (inPacket.getProtocol() != MIE::IpDatagram::ICMP) {
			continue;
		}
		if (inPacket.getTotalSize() > mtu) {
			fprintf(stderr, "packetSize(%d) is large!\n", inPacket.getTotalSize());
			puts("send DESTINATION_UNREQCHABLE");

			outFrame.setDestinationMacAddress(inFrame.getSourceMacAddress());
			outPacket.setDestinationIpAddress(inPacket.getSourceIpAddress());
			/* construct header of ICMP */
			outIcmp.setType(MIE::IcmpPacket::DESTINATION_UNREACHABLE);
			outIcmp.setCode(MIE::IcmpPacket::FRAGMENTATION_NEEDED_AND_DF_SET);
			/* construct data of ICMP */
			char *data = outIcmp.getData();
			outIcmp.setMTU(mtu);
			/*
				ʏcopySize = 28
				ől68
			*/
			assert(inPacket.getHeaderSize() <= 60);
			const int copySize = inPacket.getHeaderSize() + 8;
			memcpy(&data[4], inPacket.getData() - inPacket.getHeaderSize(), copySize);
			const int totalSize = 8 /* header */ + copySize;
			outIcmp.setTotalSize(totalSize);
			outIcmp.setChecksum();
			outPacket.setDataSize(outIcmp.getTotalSize());
			outPacket.setHeaderChecksum();
			int outBufSize = outPacket.getTotalSize() + MIE::EtherFrame::HEADER_SIZE;
			s.write(outBuf, outBufSize);
			puts("end");
			continue;
		}
		const MIE::IcmpPacket& inIcmp = inPacket.getIcmpPacket();
		const MIE::IcmpPacket::Type type = inIcmp.getType();
		printf("icmp type=%s, code=%d, size=%d, checksum=%04X\n",
			MIE::IcmpPacket::toString(type), inIcmp.getCode(), inIcmp.getTotalSize(), inIcmp.getChecksum());
		switch (type) {
		case MIE::IcmpPacket::ECHO_REQUEST:
			{
				puts("send ECHO_REPLY");
				outFrame.setDestinationMacAddress(inFrame.getSourceMacAddress());
				outPacket.setDestinationIpAddress(inPacket.getSourceIpAddress());
				/* construct header of ICMP */
				outIcmp.setType(MIE::IcmpPacket::ECHO_REPLY);
				outIcmp.setCode(MIE::IcmpPacket::UNUSED);
				/* construct data of ICMP */
				memcpy(outIcmp.getData(), inIcmp.getData(), inIcmp.getTotalSize());
				outIcmp.setTotalSize(inIcmp.getTotalSize());
				outIcmp.setChecksum();
				outPacket.setDataSize(outIcmp.getTotalSize());
				outPacket.setHeaderChecksum();
				int outBufSize = outPacket.getTotalSize() + MIE::EtherFrame::HEADER_SIZE;
				s.write(outBuf, outBufSize);
//dump(outBuf, outBufSize);
			}
			break;
		default:
			fprintf(stderr, "not support(type=%d)\n", type);
			break;
		}
	}
	return 0;
}
