/**
	raw\PbgeXg
	rawf[^MeXg
*/
#include <stdio.h>
#include <ctype.h>
#include <string>
#include "../mie_common.h"
#include "../mie_socket.h"
#include "../mie_etherframe.h"
#include "rawsocket_common.h"

int main(int argc, char *argv[])
{
	MIE::Socket::initSystem();

	/* get all network interfaces */
	MIE::NetworkInterfaces ifs;
	if (MIE::Socket::getNetworkInterfaces(ifs)) {
		ifs.dump();
	} else {
		fprintf(stderr, "getNetworkInterface error\n");
		return 1;
	}

	const int maxFileCount = 5000;
	std::string baseName;
	int idx = -1; /* no select */

	// get option
	argc--, argv++;
	while (argc > 0) {
		if (argc > 1 && strcmp(*argv, "-idx") == 0) {
			argc--, argv++;
			idx = atoi(*argv);
		} else
		if (argc > 1 && strcmp(*argv, "-b") == 0) {
			argc--, argv++;
			baseName = *argv;
		} else
		{
			fprintf(stderr, "rawsocket_r_test [-idx num]\n");
			return 1;
		}
		argc--, argv++;
	}

	if (idx == -1) {
		if (ifs.size() > 1) {
			printf("select network interface(0-%d)\n", ifs.size() - 1);
			scanf("%d", &idx);
		} else {
			idx = 0;
		}
	}

	/* verify idx value */
	if (idx >= static_cast<int>(ifs.size())) {
		fprintf(stderr, "select error\n");
		return 1;
	}

	MIE::Socket in;
	if (!in.openAsRaw(ifs[idx])) {
		fprintf(stderr, "can't open socket\n");
		return 1;
	}

	int fileCount = 0;
	for (;;) {
		char buf[2048]; /* 1packet\傫Ă */
		int readSize;
		readSize = in.read(buf, sizeof(buf));
		if (readSize <= 0) {
			fprintf(stderr, "can't read\n");
			continue;
		}
		MIE::EtherFrame frame(buf, readSize);

		char name[1024];
		sprintf(name, "%s.%04d", baseName.c_str(), fileCount);
		FILE *fp = fopen(name, "wb");
		if (!fp) {
			fprintf(stderr, "can't open %s\n", name);
			return 1;
		}
		fwrite(buf, 1, readSize, fp);
		fclose(fp);
		fprintf(stderr, "save %s(size=%d)\n", name, readSize);
		fileCount++;
		if (fileCount == maxFileCount) {
			break;
		}

#if 0
		printf("receive %dbytes (type=%s)\n", readSize, MIE::EtherFrame::toString(frame.getType()));
//			printf("  src MAC=%s, ", frame.getSourceMacAddress().toString());
//			printf("dest MAC=%s\n", frame.getDestinationMacAddress().toString());
		if (frame.getType() == MIE::EtherFrame::ARP) {
			const MIE::ArpPacket& arp = frame.getArpPacket();
			if (arp.isValid()) {
				printf("  Operation=%s\n", MIE::ArpPacket::toString(arp.getOperation()));
				printf("  src MAC=%s, dest MAC=%s\n", arp.getSourceMacAddress().toString(), arp.getDestinationMacAddress().toString());
				printf("  src ip=%s, dest ip=%s\n", arp.getSourceIpAddress().toString(), arp.getDestinationIpAddress().toString());
			} else {
				printf(" illegal ARP\n");
			}
			continue;
		}
		if (frame.getType() != MIE::EtherFrame::IPv4) {
			continue;
		}

		const MIE::IpDatagram& packet = frame.getIpDatagram();

		printf("  type=%s, ", MIE::EtherFrame::toString(frame.getType()));
		printf("version=%d, IHL=%d, ", packet.getVersion(), packet.getIHL());
		printf("identification=%d\n", packet.getIdentification());
		printf("  DF=%d, MF=%d, fragment offset=%d, ", packet.getDF(), packet.getMF(), packet.getFragmentOffset());
		printf("check sum=%04X(verify:%04X)\n", packet.getHeaderChecksum(), packet.calcHeaderChecksum());
		printf("  src ip=%s, dest ip=%s\n", packet.getSourceIpAddress().toString(), packet.getDestinationIpAddress().toString());
		printf("  protocol=%s\n", MIE::IpDatagram::toString(packet.getProtocol()));
		if (packet.getVersion() != MIE::IpDatagram::VERSION_IPV4) {
			fprintf(stderr, "  not support version=%d\n", packet.getVersion());
			continue;
		}
		if (packet.getMF()) {
			fprintf(stderr, "  not support fragmented packet\n");
			continue;
		}
		if (packet.getFragmentOffset()) {
			fprintf(stderr, "  not support nonzero fragment offset packet\n");
			continue;
		}

		if (packet.getProtocol() == MIE::IpDatagram::ICMP) {
			const MIE::IcmpPacket& icmp = packet.getIcmpPacket();
			printf("  type=%s, ", MIE::IcmpPacket::toString(icmp.getType()));
			printf("len=%d, ", icmp.getTotalSize());
			printf("cksum=%04X(calc:%04X)\n", icmp.getChecksum(), icmp.calcChecksum());
			continue;
		}
		if (packet.getProtocol() == MIE::IpDatagram::UDP) {
			const MIE::UdpPacket *udp = packet.getUdpPacket();
#if 0
			if (udp->getDestinationPort() != 31010) {
				printf("  src port=%d, dest port=%d\n", udp->getSourcePort(), udp->getDestinationPort());
				continue;
			}
#endif
			printf("    src port=%d, dest port=%d\n", udp->getSourcePort(), udp->getDestinationPort());
//
{
printf("  ***cksum=%04X, calc=%04X\n", udp->getChecksum(), udp->calcChecksum(packet.getSourceIpAddress().get(), packet.getDestinationIpAddress().get(), packet.getProtocol()));
}
//
			dump(udp->getData(), udp->getDataSize());
		}
#endif
	} /* for (;;) */
	return 0;
}
