/**
	raw\PbgeXg
	UDPf[^MeXg
*/
#include <stdio.h>
#include <ctype.h>
#include <string>
#include "../mie_common.h"
#include "../mie_socket.h"
#include "../mie_etherframe.h"
#include "../mie_util.h"
#include "rawsocket_common.h"

#define DEFAULT_ADDRESS "localhost"
#define DEFAULT_PORT 31010
#define DEFAULT_SIZE 10
#define DEFAULT_COUNT 1

class Option : public MIE::Util::OptionBase {
private:
	std::string address_;
	int port_;
	int size_;
	int count_;
	int idx_;
	int ifsNum_;

	void usage()
	{
		fprintf(stderr, "rawsocket_s_test [-h] [-to address[:port]] [-s size] [-c count] [-idx idx]\n");
		fprintf(stderr, "option\n");
		fprintf(stderr, "-h                 --- help\n");
		fprintf(stderr, "-to address[:port] --- address to send(default:%s[:%d])\n", DEFAULT_ADDRESS, DEFAULT_PORT);
		fprintf(stderr, "-s size            --- packet size(default:%d)\n", DEFAULT_SIZE);
		fprintf(stderr, "-c count           --- number of packet to send(default:%d)\n", DEFAULT_COUNT);
		fprintf(stderr, "-idx idx           --- index of network interface\n");
	}

public:
	Option(int ifsNum)
		: port_(-1)
		, size_(-1)
		, count_(-1)
		, idx_(-1)
		, ifsNum_(ifsNum)
	{
	}
	/**
		IvV
	*/
	bool analyze(int argc, const char *const argv[])
	{
		set(argc, argv);

		while (hasOption()) {
			if (isMatch("-h", 0)) {
				usage();
				return false;
			} else
			if (isMatch("-to", 1)) {
				const char *address = getString();
				const char *p;
				p = strchr(address, ':');
				// ':' 
				if (p) {
					// ܂łstringɑ
					address_ = std::string(address, p - address);
					port_ = atoi(p + 1);
				} else {
					address_ = address;
				}
			} else
			if (isMatch("-s", 1)) {
				size_ = getInteger();
			} else
			if (isMatch("-c", 1)) {
				count_ = getInteger();
			} else
			if (isMatch("-idx", 1)) {
				idx_ = getInteger();
			} else
			{
				usage();
				return false;
			}
			next();
		}

		// ftHgw
		if (address_.empty()) {
			address_ = DEFAULT_ADDRESS;
		}
		if (port_ == -1) {
			port_ = DEFAULT_PORT;
		}
		if (size_ == -1) {
			size_ = DEFAULT_SIZE;
		}
		if (count_ == -1) {
			count_ = DEFAULT_COUNT;
		}
		if (idx_ == -1) {
			if (ifsNum_ > 1) {
				fprintf(stderr, "select network interface(0-%d)\n", ifsNum_ - 1);
				scanf("%d", &idx_);
			} else {
				idx_ = 0;
			}
		}
		if (idx_ >= ifsNum_) {
			fprintf(stderr, "select error\n");
			usage();
			return false;
		}

		fprintf(stderr, "address=%s:%d\n", address_.c_str(), port_);
		fprintf(stderr, "  idx=%d\n", idx_);
		fprintf(stderr, "  size=%d, count=%d\n", size_, count_);
		return true;
	}
	const char *getAddress() const { return address_.c_str(); }
	int getPort() const { return port_; }
	int getSize() const { return size_; }
	int getCount() const { return count_; }
	int getIdx() const { return idx_; }
};

int main(int argc, char *argv[])
{
	MIE::Socket::initSystem();

	/* get all network interfaces */
	MIE::NetworkInterfaces ifs;
	if (MIE::Socket::getNetworkInterfaces(ifs)) {
		ifs.dump();
	} else {
		fprintf(stderr, "getNetworkInterface error\n");
		return 1;
	}

	if (ifs.size() < 1) {
		fprintf(stderr, "no interface\n");
		return 1;
	}

	Option opt(ifs.size());
	if (!opt.analyze(argc, argv)) {
		return 1;
	}

	MIE::Socket out;
	const MIE::NetworkInterface& curIf = ifs[opt.getIdx()];

	if (!out.openAsRaw(curIf)) {
		fprintf(stderr, "can't open socket\n");
		return 1;
	}

	/* get IP/MAC address */
	const MIE::IpAddress& srcIp = curIf.getIpAddress();
	const MIE::MacAddress& srcMac = curIf.getMacAddress();

	MIE::IpAddress destIp;
	MIE::MacAddress destMac;

	destIp.set(opt.getAddress());

	if (!destIp.getMacAddress(destMac)) {
		fprintf(stderr, "can't get mac(%s)\n", destIp.toString());
		return 1;
	}

	/* construct ether frame */
	char buf[2048];
	memset(buf, 0, sizeof(buf));
	MIE::EtherFrame frame(buf, sizeof(buf));

	frame.clearHeader();
	frame.setSourceMacAddress(srcMac);
	frame.setDestinationMacAddress(destMac);

	/* construct ipDatagram */
	MIE::IpDatagram& ipDatagram = frame.getIpDatagram();
	ipDatagram.clearHeader();
	ipDatagram.setSourceIpAddress(srcIp);
	ipDatagram.setDestinationIpAddress(destIp);

/// test for DF
	ipDatagram.setDF(true);
///

	/* construct UDP packet */
	const int port = opt.getPort();
	const int udpSize = opt.getSize();
	MIE::UdpPacket* udp = ipDatagram.getUdpPacket();
	udp->setSourcePort(port);
	udp->setDestinationPort(port);
	udp->setChecksum(0);
	udp->setDataSize(udpSize);

	ipDatagram.setDataSize(udp->getTotalSize());
	ipDatagram.setHeaderChecksum();

	for (int i = 0; i < opt.getCount(); i++) {
		int dataSize = ipDatagram.getTotalSize() + MIE::EtherFrame::HEADER_SIZE;
		printf("write size=%d\n", dataSize);
		int writeSize = out.write(buf, dataSize);
		printf("writeSize=%d\n", writeSize);
	}
	return 0;
}
