#ifndef TS_COMMON_H_
#define TS_COMMON_H_
/*!
	ʗp
	@author $Author: shigeo $
	@version $Revision: 1.19 $
	@date $Date: 2007/08/30 02:09:07 $
*/
#include "mie_timecapsule.h"
#include "cipher_rijndael.h"
#include "ts_protocol.h"
#include <assert.h>

#define PUBLIC_KEY_FILE "timecapsule.pub" /* Jt@C */
#define TC_SUF "tce" /* Ít@C̊gq */

#ifndef debug
	#undef debug
	#ifndef NDEUBG
		#define debug 1
	#else
		#define debug 0
	#endif
#endif


namespace MIE_TC {

template<class Input, class Output>
bool EncodeCore(
	const PublicKey& publicKey,
	const char *decTime,
	Input& in,
	Output& out,
	const char *seed
)
{
	TimeInfo timeInfo;
	if (!timeInfo.set(decTime)) {
		if (debug) fprintf(stderr, "can 't set time [%s]\n", decTime);
		return false;
	}

	SessionKey sessionKey;
	EncodedSessionKey encodedSessionKey;
	if (!sessionKey.init(&encodedSessionKey, publicKey, timeInfo, seed)) {
		if (debug) fprintf(stderr, "can't encode session key\n");
		return false;
	}

	if (debug) {
		fprintf(stderr, "session key by generator\n%s\n", sessionKey.get());
		fprintf(stderr, "encodedSessionKey\n%s\n", encodedSessionKey.get());
	}

	/* ܂ZbVwb_Ƃď */
	{
		const char *key = encodedSessionKey.get();
		out.write(key, strlen(key));
		char term[1] = { 0 };
		out.write(term, 1); /* NUL̏ */
	}

	/* {̂̈Í */
	const char *key = sessionKey.get();
	bool isOK;
	isOK = MTT::RIJNDAEL_do(MTT::CipherRijndael::ENCODING, key, 256, in, out);
	return isOK;
}

template<class Input, class Output>
bool DecodeCore(
	Input& in,
	Output& out,
	const char *address,
	const PublicKey *publicKey = 0
)
{
	/* ÍꂽZbV擾 */
	std::string header;
	std::getline(in, header, '\0');
	if (!in) {
		if (debug) fprintf(stderr, "bad format\n");
		return false;
	}
	EncodedSessionKey encodedSessionKey(header.c_str());

	if (debug) fprintf(stderr, "encodedSessionKey:\n%s\n", encodedSessionKey.get());

	/* 擾 */
	TimeInfo timeInfo;
	if (!encodedSessionKey.getTimeInfo(&timeInfo)) {
		if (debug) fprintf(stderr, "can't get timeInfo\n");
		return false;
	}

	/* 擾 */
	SignedTimeInfo signedTimeInfo;
	if (!Protocol::GetSignedTimeInfoByHTTP(&signedTimeInfo, timeInfo, address)) {
		if (debug) fprintf(stderr, "can't getTimeSignature\n");
		return false;
	}

	/* ZbV */
	SessionKey sessionKey;
	if (!encodedSessionKey.decode(&sessionKey, signedTimeInfo, publicKey)) {
		if (debug) fprintf(stderr, "can't decode session key\n");
		return false;
	}

	if (debug) fprintf(stderr, "session key by generator\n%s\n", sessionKey.get());

	/*  */
	const char *key = sessionKey.get();
	return MTT::RIJNDAEL_do(MTT::CipherRijndael::DECODING, key, 256, in, out);
}

} // end of namespace

#endif /* TS_COMMON_H_ */
