#ifndef TS_PROTOCOL_H_
#define TS_PROTOCOL_H_

/*!
	̂Ƃ̂߂̃vgR(HTTPx[X)
	@file ts_protocol.h
	@author $:Author$
	@version $Revision: 1.23 $
	@date $Date: 2007/08/25 02:47:56 $
*/
#include "mie_timecapsule.h"
#include "mie_socket.h"
#include <string>

namespace MIE_TC {

#define TIME_SERVER_ADDRESS "mie.mydns.jp"
#define CGI_PATH "/cgi-bin/time.cgi"

/*!
	̂Ƃ邽߂̊֐Q
*/
namespace Protocol {

enum {
	PORT = 80,
	BUF_SIZE = 4096,
	debug = 0
};

/*!
	Mf[^body^O̒go
*/
const char* getBody(MIE::Socket& socket, char *buf, int bufSize)
{
	int readSize = 0;
	int remainSize = bufSize - 1; /* ŌNUL̂ */
	for (;;) {
		int tmp = socket.read(&buf[readSize], remainSize);
		if (tmp < 0) return 0;
		if (tmp == 0) break;
		readSize += tmp;
		remainSize -= tmp;
		if (remainSize <= 0) return 0;
	}
	buf[readSize] = 0;
	if (debug) fprintf(stderr, "rec [%s]\n", buf);

	const char *p;
	p = strstr(buf, "<body>");
	if (p == 0) return 0;
	p += 6; /* 6 = sizeof("<body>"); */
	char *q = const_cast<char*>(strstr(p, "</body>"));
	if (q == 0) return 0;
	*q = 0;
	return p;
}

enum RequestMode {
	RequestError,
	RequestPublicKey,
	RequestSignedTimeInfo
};

/*!
	oCif[^ƕ̔r
	@param p [in] rΏۃf[^
	@param str [in] rΏە(NULI[Ă邱)
*/
inline bool isEqualSubString(const char *p, const char *str)
{
	int len = strlen(str);
	return memcmp(p, str, len) == 0;
}

inline RequestMode GetRequestMode(TimeInfo *timeInfo, const char *buf)
{
	const char *msg_GET = "GET " CGI_PATH;

	if (isEqualSubString(buf, msg_GET)) {
		buf += strlen(msg_GET);
		if (isEqualSubString(buf, "?publickey")) {
			return RequestPublicKey;
		} else
		if (*buf == ' ') {
			timeInfo->setNow();
			return RequestSignedTimeInfo;
		} else
		if (*buf == '?') {
			if (timeInfo->set(buf + 1)) {
				TimeInfo now;
				now.setNow();
				if (timeInfo->isNewerThan(now)) {
					if (debug) fprintf(stderr, "%s is in the future.\n", timeInfo->get());
				} else {
					return RequestSignedTimeInfo;
				}
			}
		}
	}
	return RequestError;
}

inline bool SendMessageByHTTP(MIE::Socket& socket, const char *str)
{
	std::string msg;
	msg = "Content-type: text/html\r\n\r\n"
		"<html><head><title>time capsule</title></head><body>";
	msg += str;
	msg += "</body></html>\r\n";
	int len = msg.size();
	if (debug) fprintf(stderr, "send=[%s]\n", msg.c_str());
	return socket.write(msg.c_str(), len) == len;
}

/*!
	J擾
	@param name [in] Jt@C
	@param address [in] T[o
	@param cgiPath [in] cgipX(ftHgCGI_PATH)
*/
inline bool GetPublicKeyFileByHTTP(const char *name, const char *address, const char *cgiPath = 0)
{
	MIE::Socket socket;
	if (!socket.openAsClient(address, Protocol::PORT)) {
		if (debug) fprintf(stderr, "can't socket.openAsClient\n");
		return false;
	}

	/* Jv */
	std::string msg = "GET ";
	msg += cgiPath ? cgiPath : CGI_PATH;
	msg += "?publickey HTTP/1.0\r\n\r\n";
	if (debug) fprintf(stderr, "query [%s]\n", msg.c_str());
	socket.write(msg.c_str(), msg.size());

	/* ʎM */
	char buf[BUF_SIZE];
	const char *p = getBody(socket, buf, sizeof(buf));
	if (p == 0) return false;

	/* Jݒ */
	PublicKey publicKey;
	if (publicKey.set(p)) {
		if (debug) printf("rec size=%d\n", (int)strlen(p));
		/* 擾łΕۑ */
		return publicKey.save(name);
	} else {
		if (debug) fprintf(stderr, "can't get publicKey\n");
		return false;
	}
}

/*!
	擾
	@param signedTimeInfo [out] 
	@param timeInfo [in] 
	@param address [in] T[o
	@param cgiPath [in] cgipX(ftHgCGI_PATH)
*/
inline bool GetSignedTimeInfoByHTTP(SignedTimeInfo *signedTimeInfo, const TimeInfo& timeInfo, const char *address, const char *cgiPath = 0)
{
	MIE::Socket socket;
	if (!socket.openAsClient(address, PORT)) {
		if (debug) fprintf(stderr, "can't socket.openAsClient()\n");
		return false;
	}

	/* v */
	std::string msg = "GET ";
	msg += cgiPath ? cgiPath : CGI_PATH;
	msg += "?";
	msg += timeInfo.get();
	msg += " HTTP/1.0\r\n\r\n";
	if (debug) fprintf(stderr, "query [%s]\n", msg.c_str());
	socket.write(msg.c_str(), msg.size());

	/* ʎM */
	char buf[BUF_SIZE];
	const char *p = getBody(socket, buf, sizeof(buf));
	if (p == 0) return false;

	if (debug) fprintf(stderr, "singedTimeInfo [%s]\n", p);
	/* ݒ */
	return signedTimeInfo->set(p);
}

} // end of namespace

} // end of namespace

#endif /* TS_PROTOCOL_H_ */
