#include <cstdio>
#include <cstdlib>
#include "common.h"
#include "aligned_new.h"

/*
	alignment ꂽ|C^Ԃ
	*((void**)Ԃ|C^ - 1) Ɋmۂ|C^ۑĂ
*/

void* alignedMalloc(size_t size) throw()
{
	const int header = sizeof(void*); // |C^ۑ̂ߐ擪Ɋm
	const int alignMask = MEMORY_ALIGN_SIZE - 1;

	/* FAS_alloc()ĂяoĊmۂꂽ|C^̓v[Ȃ̂ŊJȂ */
	char* allocAdr = reinterpret_cast<char*>(malloc(size + alignMask + 1 + header));
	if (allocAdr == NULL) {
		assert(0);
	}

	char* returnAdr = (allocAdr + alignMask + header);

	if (sizeof(void*) == sizeof(uint32)) {
		uint32 alignedAdr = (reinterpret_cast<uint32>(returnAdr)) & ~((uint32)alignMask);
		returnAdr = reinterpret_cast<char*>(alignedAdr);
	} else

	{
		fprintf(stderr, "error in alignedMalloc(). unknown pointer size.");
		abort();
	}
	*(reinterpret_cast<void**>(returnAdr) - 1) = allocAdr;
	return reinterpret_cast<void *>(returnAdr);
}

/*
	alignment ꂽ|C^Ԃ

	*((void**)Ԃ|C^ - 1)  NULL
	blockToBlockHeader(Ԃ|C^)  BlockHeader 
	gpubN ϐnext ̃XgŊǗAgpς݃ubN̓Xg擪ɒǉ
*/

struct FASBlockHeader {
	FASBlockHeader* next;
	void* dummy_pointer;
};

const int FASBlockHeader_Size = (sizeof(FASBlockHeader) + MEMORY_ALIGN_SIZE) & ~(MEMORY_ALIGN_SIZE - 1);

static FASBlockHeader* next = NULL;

inline void* blockHeaderToBlock(FASBlockHeader* blockHeader)
{
	return reinterpret_cast<void*>(reinterpret_cast<char*>(blockHeader) + FASBlockHeader_Size);
}

inline FASBlockHeader* blockToBlockHeader(void* block)
{
	return reinterpret_cast<FASBlockHeader*>(reinterpret_cast<char*>(block) - FASBlockHeader_Size);
}

void* FAS_alloc()
{
	void* result;
	const int alignMask = MEMORY_ALIGN_SIZE - 1;
	const int blockSize = (FREQUENCY_ALLOCATE_SIZE + alignMask + 1 + FASBlockHeader_Size) & ~alignMask;
	const int blockCountPerAlloc = 512;
	const int allocSize = blockSize * blockCountPerAlloc + alignMask + 1;

	if (next) {
		result = blockHeaderToBlock(next);
		next = next->next;
	} else {
		/* v[Ȃ̂ł̃|C^͊JȂ */
		char* firstBlockHeaderAdr = reinterpret_cast<char*>(alignedMalloc(allocSize));
		FASBlockHeader* curBlockHeader  = reinterpret_cast<FASBlockHeader*>(firstBlockHeaderAdr);
		FASBlockHeader* nextBlockHeader = reinterpret_cast<FASBlockHeader*>(firstBlockHeaderAdr + blockSize);
		curBlockHeader->next = NULL;
		// initialize dummy_pointer for compatibility with alignedMalloc()
		*(reinterpret_cast<void**>(blockHeaderToBlock(curBlockHeader)) - 1) = NULL;
		next = nextBlockHeader;

		result = blockHeaderToBlock(curBlockHeader);

		char* blockHeaderAddr = reinterpret_cast<char*>(next);
		for (int i = 1; i < blockCountPerAlloc - 1; i ++, blockHeaderAddr += blockSize) {
			curBlockHeader  = reinterpret_cast<FASBlockHeader*>(blockHeaderAddr);
			nextBlockHeader = reinterpret_cast<FASBlockHeader*>(blockHeaderAddr + blockSize);
			curBlockHeader->next = nextBlockHeader;
			*(reinterpret_cast<void**>(blockHeaderToBlock(curBlockHeader)) - 1) = NULL;
		}
		curBlockHeader = reinterpret_cast<FASBlockHeader*>(blockHeaderAddr);
		curBlockHeader->next = NULL;
		*(reinterpret_cast<void**>(blockHeaderToBlock(curBlockHeader)) - 1) = NULL;
	}
	return result;
}

void FAS_free(void* ptr)
{
	FASBlockHeader* blockHeader = blockToBlockHeader(ptr);
	blockHeader->next = next;
	next = blockHeader;
}
