#ifndef MTT_FIELD_BASE_H_
#define MTT_FIELD_BASE_H_

#include "common.h"
#include "mttimp_ua.h"
#include "field_base.h"
#include "bitsbase.h"
#include "fpbase.h"

namespace MTT {

int ua_init_from_str(char const* str, uint32* ua, unsigned int ua_len);
int ua_init_from_str_mod(char const* str, uint32 const* mod, unsigned int mod_len, uint32* ua, unsigned int ua_len);
int ua_init_from_int(int dat, uint32* ua, unsigned int ua_len, int* sign);
int ua_init_from_int_mod(int dat, uint32 const* mod, unsigned int mod_len, uint32* ua, unsigned int ua_len);
int ua_convert_to_str(uint32 const* ua, unsigned int ua_len, char* str, unsigned int str_len);

struct UAS {
	enum {
		SIZE = 36
	};
	uint32 ua[SIZE];
	UAS* mod;
	int sign;
	bool normalized;
	int ua_size;

	UAS() {}
	UAS(int d, UAS* mod_ = 0)
	{
		InitFromInt(d, mod_);
	}
	UAS(const char *const str, UAS* mod_ = 0)
	{
		InitFromStr(str, mod_);
	}
	void InitFromInt(int d, UAS* mod_ = 0)
	{
		mod = mod_;
		if (mod != 0) {
			ua_init_from_int_mod(d, mod->ua, SIZE, ua, SIZE);
			sign = 1;
		} else {
			ua_init_from_int(d, ua, SIZE, &sign);
		}
		ua_size = ua_get_length(ua, SIZE);
		if (ua_size == 0) {
			sign = 0;
			normalized = true;
		} else {
			normalized = false;
		}
	}
	void InitFromStr(const char *const str, UAS* mod_ = 0)
	{
		mod = mod_;

		if (mod != 0) {
			ua_init_from_str_mod(str, mod->ua, SIZE, ua, SIZE);
		} else {
			int result = ua_init_from_str(str, ua, SIZE);
			if (result < 0) printf("%s\n", str); fflush(stdout);
			assert(result > 0);
		}
		ua_size = ua_get_length(ua, SIZE);
		if (ua_size == 0) {
			sign = 0;
			normalized = true;
		} else {
			sign = 1;
			normalized = false;
		}
	}
	void InitFromUint32(uint32 ui32, UAS* mod_ = 0)
	{
		mod = mod_;
		ua_set_uint(ui32, ua, SIZE);
		sign = 1;
		normalized = true;
		ua_size = 1;
	}

	void Clear()
	{
		ua_clear(ua, ua_size);
		ua_size = 0;
		normalized = true;
		sign = 0;
	}

	static int GetMaxLength(UAS const* p_src1, UAS const* p_src2)
	{
		return (p_src1->ua_size > p_src2->ua_size) ? p_src1->ua_size : p_src2->ua_size;
	}

	static int Compare(UAS const* p_src1, UAS const* p_src2)
	{
		return ua_cmp(p_src1->ua, p_src2->ua, GetMaxLength(p_src1, p_src2));
	}

	static void Add(UAS const* p_src1, UAS const* p_src2, UAS* p_dst)
	{
		int len = GetMaxLength(p_src1, p_src2);
		if (ua_add(p_src1->ua, p_src2->ua, p_dst->ua, len)) {
			p_dst->ua[len] = 1;
			len++;
		}
		if (p_dst->ua_size > len)
			ua_clear(p_dst->ua + len, p_dst->ua_size-len);
		p_dst->ua_size = len;
	}

	static void Sub(UAS const* p_src1, UAS const* p_src2, UAS* p_dst)
	{
		int len = GetMaxLength(p_src1, p_src2);
		if (ua_sub(p_src1->ua, p_src2->ua, p_dst->ua, len)) {
			assert(false);
		}
		if (p_dst->ua_size > len)
			ua_clear(p_dst->ua + len, p_dst->ua_size-len);
		p_dst->ua_size = ua_get_length(p_dst->ua, len);
	}

	static void ua_mul(uint32 const* p_src1, uint32 const* p_src2, uint32* p_dst, int len)
	{
		// dst  src ƏdłȂ
		// Zɂ
		int n1 = ua_get_length(p_src1, len);
		int n2 = ua_get_length(p_src2, len);
		if (n1 < n2) {
			uint32 const* p=p_src1; p_src1=p_src2; p_src2=p;
			int n=n1; n1=n2; n2=n;
		}

		len *= 2;
		if (n2== 0) {
			ua_clear(p_dst, len);
			return;
		}

		if (p_src1 == p_src2) {
			// 
			ua_clear(p_dst+n1*2, len-n1*2);
			ua_square_s(p_src1++, p_dst, n1--);
			for(int i=1; i<n2; i++)
				ua_square_add_s(p_src1++, (p_dst+=2), n1--);
		} else {
			// MZZ
			ua_clear(p_dst+n1+n2, len-n1-n2);
			ua_mul_s(p_src1, *p_src2++, p_dst++, n1);
			for(int i=1; i<n2; i++)
				ua_mul_add_s(p_src1, *p_src2++, p_dst++, n1);
		}
	}

	static void Mul(UAS* p_src1, UAS* p_src2, UAS* p_dst)
	{
		int len = GetMaxLength(p_src1, p_src2);
		if (p_src1 == p_dst || p_src2 == p_dst) {
			uint32 tmp[SIZE];
			ua_mul(p_src1->ua, p_src2->ua, tmp, len);
			ua_copy(tmp, p_dst->ua, len*2);
		} else {
			ua_mul(p_src1->ua, p_src2->ua, p_dst->ua, len);
		}
		len *= 2;
		if (p_dst->ua_size > len)
			ua_clear(p_dst->ua + len, p_dst->ua_size-len);
		p_dst->ua_size = ua_get_length(p_dst->ua, len);
	}
};

/*!
	@brief MttFmBase: L F_m ̊bƂȂNX
*/

class MttFmBase {
private:
	// @breif MttFmBase::sf_inverse(): IuWFNg̋t(*obj = 1/ *obj)
	//
	// *obj  0 ȂΌʂ͕ۏ؂ȂB

	static inline void uas_modinv(UAS* uas_obj)
	{
		UAS r0, r1, t0, t1;
		UAS *pr0 = &r0, *pr1 = &r1, *pt0 = &t0, *pt1 = &t1;

		*pr0 = *uas_obj->mod;
		pr0->mod = uas_obj->mod;
		*pr1 = *uas_obj;
		pt0->InitFromUint32(0, uas_obj->mod);
		pt1->InitFromUint32(1, uas_obj->mod);

		if (UAS::Compare(uas_obj, &MttFmBase::uas_2) < 0 ) {
			uas_obj->InitFromUint32(1, uas_obj->mod);
			return;
		}

		UAS q, tmp;
		q.InitFromUint32(0, uas_obj->mod);
		tmp.InitFromUint32(0, uas_obj->mod);
		for(;;) {
			int len = UAS::GetMaxLength(pr0, pr1);
			ua_div(pr0->ua, pr1->ua, q.ua, pr0->ua, len);
			if (q.ua_size > len)
				ua_clear(q.ua + len, q.ua_size - len);
			q.ua_size = ua_get_length(q.ua, len);
			pr0->ua_size = ua_get_length(pr0->ua, len);

			MttFmBase::sf_mul(&q, pt1, &tmp);
			MttFmBase::sf_sub(pt0, &tmp, pt0);

			if (UAS::Compare(pr0, &MttFmBase::uas_2) < 0 ) {
				*uas_obj = *pt0;
				break;
			}

			UAS *p;
			p = pr0; pr0 = pr1; pr1 = p;
			p = pt0; pt0 = pt1; pt1 = p;
		}
	}
//protected:
public: // QQQ
	// for ctor/dtor
	static inline void* sf_init_value(int v)
	{
		UAS *uas = new UAS(v, &uas_m);
		return uas;
	}
	static inline void* sf_init_str(char const* str)
	{
		UAS *uas = new UAS(str, &uas_m);
		return uas;
	}
	static inline void* sf_copy(void const* obj)
	{
		UAS const* p_uas = reinterpret_cast<UAS const*>(obj);
		UAS *uas = new UAS();
		*uas = *p_uas;
		return uas;
	}
	static inline void sf_delete(void* obj)
	{
		UAS* p_uas = reinterpret_cast<UAS*>(obj);
		delete p_uas;
	}
	
	// for access
	static inline char const* sf_get_str(void const* obj)
	{
		UAS const* p_uas = reinterpret_cast<UAS const*>(obj);
		UAS t_uas = *p_uas;
		sf_normalize(&t_uas);

		static char string_buf[UAS::SIZE*10];
		
		ua_convert_to_str(t_uas.ua, UAS::SIZE, string_buf, sizeof(string_buf));
		return string_buf;
	}
	static inline void sf_show(void const* obj)
	{
		printf("%s", sf_get_str(obj));
	}
	
	// for operator
	static inline void sf_add(void const* src1, void const* src2, void* dst)
	{
		UAS const* p_src1 = reinterpret_cast<UAS const*>( src1 );
		UAS const* p_src2 = reinterpret_cast<UAS const*>( src2 );
		UAS* p_dst = reinterpret_cast<UAS*>( dst );

		if (p_src1->sign == 0) {
			*p_dst = *p_src2;
			return;
		}

		if (p_src2->sign == 0) {
			*p_dst = *p_src1;
			return;
		}

		p_dst->normalized = false;

		if (p_src1->sign == p_src2->sign) {
			UAS::Add(p_src1, p_src2, p_dst);
			p_dst->sign = p_src1->sign;
		} else {
			int cmp = UAS::Compare(p_src1, p_src2);
			if (cmp == 0) {
				p_dst->Clear();
				return;
			} else if (cmp > 0) {
				UAS::Sub(p_src1, p_src2, p_dst);
				p_dst->sign = p_src1->sign;
			} else {
				UAS::Sub(p_src2, p_src1, p_dst);
				p_dst->sign = p_src2->sign;
			}
		}

		if (p_dst->ua[UAS::SIZE-1] > 0)
			sf_normalize(p_dst);
	}
	static inline void sf_sub(void const* src1, void const* src2, void* dst)
	{
		UAS const* p_src1 = reinterpret_cast<UAS const*>( src1 );
		UAS* p_src2 = const_cast<UAS*>(reinterpret_cast<UAS const*>( src2 ));
		UAS* p_dst = reinterpret_cast<UAS*>( dst );

		if (p_src1 == p_src2) {
			p_dst->Clear();
			return;
		}

		p_src2->sign = -p_src2->sign;
		sf_add(p_src1, p_src2, p_dst);
		if (p_src2 != p_dst)
			p_src2->sign = -p_src2->sign;
	}
	static inline void sf_mul(void const* src1, void const* src2, void* dst)
	{
		UAS* p_src1 = const_cast<UAS*>(reinterpret_cast<UAS const*>( src1 ));
		UAS* p_src2 = const_cast<UAS*>(reinterpret_cast<UAS const*>( src2 ));
		UAS* p_dst = reinterpret_cast<UAS*>( dst );

		// ZʂUAS::SIZEȉɂ
		if (p_src1->ua_size >= UAS::SIZE/2) sf_normalize(p_src1);
		if (p_src2->ua_size >= UAS::SIZE/2) sf_normalize(p_src2);

		if (p_src1->sign == 0 || p_src2->sign == 0) {
			p_dst->Clear();
			return;
		}

		UAS::Mul(p_src1, p_src2, p_dst);

		int cmp = UAS::Compare(p_dst, p_dst->mod);
		if (cmp == 0) {
			p_dst->Clear();
			return;
		}

		p_dst->sign = p_src1->sign * p_src2->sign;
		p_dst->normalized = (cmp < 0 && p_dst->sign > 0) ? true : false;
	}

	// for operation
	static inline void sf_inverse(void* obj)
	{
		UAS* p_uas = reinterpret_cast<UAS*>(obj);

		// gcd, O(log n)
		sf_normalize(p_uas);
		uas_modinv(p_uas);
	}
	static inline void sf_normalize(void* obj)
	{
		UAS* p_uas = reinterpret_cast<UAS*>(obj);

		assert(p_uas->mod != 0);

		if (p_uas->normalized) return;
		p_uas->normalized = true;

		int cmp = UAS::Compare(p_uas, p_uas->mod);

		if (cmp == 0) {
			p_uas->Clear();
			return;
		}

		if (cmp > 0) {
			uint32 tmp[UAS::SIZE];
			ua_div(p_uas->ua, p_uas->mod->ua, tmp, p_uas->ua, UAS::GetMaxLength(p_uas, p_uas->mod));
		}

		p_uas->ua_size = ua_get_length(p_uas->ua, UAS::SIZE);

		if (p_uas->ua_size == 0) {
			p_uas->sign = 0;
			return;
		}
		if (p_uas->sign < 0) {
			UAS::Sub(p_uas->mod, p_uas, p_uas);
			p_uas->sign = 1;
		}
	}
	static inline void sf_clear(void* obj)
	{
		UAS* p_uas = reinterpret_cast<UAS*>(obj);
		p_uas->Clear();
	}
	static inline void sf_change_sign(void* obj)
	{
		UAS* p_uas = reinterpret_cast<UAS*>(obj);
		p_uas->normalized = (p_uas->sign != 0) ? false : true;
		p_uas->sign = -p_uas->sign;
	}
	static inline bool sf_is_zero(const void* obj)
	{
		UAS const* p_uas = reinterpret_cast<UAS const*>(obj);
		return (p_uas->sign == 0) ? true : false;
	}
public:
	static UAS uas_m, uas_0, uas_2;
	static void setup();
};

} // end of namespace

#endif
