#ifndef MTT_FIELD_FQ_H_
#define MTT_FIELD_FQ_H_

#include "field_base.h"
#include "field_fm.h"

namespace MTT {

class MttBits;

// ==========================================================================
/*!
	@brief MttFp: L F_p (p: 512bitf)
	@file $RCSfile: field_fq.h,v $
	@version $Revision: 1.9 $
	@date $Date: 2007/08/30 02:09:07 $
*/
// F_p̉ZNX
// Z +=, -=, *=, inverse(), clear(), normalize(), sqrt() T|[g
// normalize() Ȃ normalize() Ă邱Ƃۏ؂ȂB
// ܂̂Ƃ isZero() ͐ʂԂƂ͌ȂB
// ctor̓iVAFpA{^Adouble T|[g
// FpA{^T|[g
//

class MttFp : public MttFpBase {
// public member functions
public:
	// ctors
	MttFp() : p_value_(0) {}
	MttFp(MttFp const& fp) : p_value_(sf_copy(fp.p_value_)) {}
	explicit MttFp(char const* i_str) : p_value_(sf_init_str(i_str)) {}
	explicit MttFp(int x) : p_value_(sf_init_value(x)) {}
	
	// dtors
	~MttFp() { sf_delete(p_value_); p_value_ = 0;}
	
	// accessors
	void const* getValue() const { return p_value_; }
	char const* toStr() const { return sf_get_str(p_value_); }
	void show() const { sf_show(p_value_); }
	
	// operators
	MttFp& operator=(MttFp const& fp) {
		if(&fp != this) {
			sf_delete(p_value_);
			p_value_ = sf_copy(fp.p_value_);
		}
		return *this;
	}
	MttFp& operator=(char const* i_str) {
		sf_delete(p_value_);
		p_value_ = sf_init_str(i_str);
		return *this;
	}
	MttFp& operator=(int x) {
		sf_delete(p_value_);
		p_value_ = sf_init_value(x);
		return *this;
	}
	MttFp& operator+=(MttFp const& fp) {
		sf_add(p_value_, fp.p_value_, p_value_);
		return *this;
	}
	MttFp& operator-=(MttFp const& fp) {
		sf_sub(p_value_, fp.p_value_, p_value_);
		return *this;
	}
	MttFp& operator*=(MttFp const& fp) {
		sf_mul(p_value_, fp.p_value_, p_value_);
		return *this;
	}

	// operations
	void inverse() { sf_inverse(p_value_); }
	void normalize() const { sf_normalize(const_cast<void*>(p_value_)); }
	void clear() { sf_clear(p_value_); }
	void changeSign() { sf_change_sign(p_value_); }
	bool isZero() const { return sf_is_zero(p_value_); }
	void cbrt() { sf_cbrt(p_value_); }
	void sqrt() { sf_sqrt(p_value_); }
	void pow(MttFp const& fp) { sf_pow(p_value_, fp.p_value_); }
	static inline void Add(MttFp const& src1, MttFp const& src2, MttFp& dst)
		{ if(!dst.p_value_) dst.p_value_ = sf_init_value(0);
		  sf_add(src1.p_value_, src2.p_value_, dst.p_value_); }
	static inline void Sub(MttFp const& src1, MttFp const& src2, MttFp& dst)
		{ if(!dst.p_value_) dst.p_value_ = sf_init_value(0);
		  sf_sub(src1.p_value_, src2.p_value_, dst.p_value_); }
	static inline void Mul(MttFp const& src1, MttFp const& src2, MttFp& dst)
		{ if(!dst.p_value_) dst.p_value_ = sf_init_value(0);
		  sf_mul(src1.p_value_, src2.p_value_, dst.p_value_); }

// the member
protected:
	void* p_value_;

public:
};



// ==========================================================================
// @brief MttFq: L F_q (q:p^2, p: 512bitf)
//
// F_q̉ZNX
// Z +=, -=, *=, inverse(), clear(), normalize(), sqrt() T|[g
// normalize() Ȃ normalize() Ă邱Ƃۏ؂ȂB
// ctor̓iVAFq, FpA{^(1A2)T|[g
// Fq, FpA{^Adouble(1A2)T|[g
//
// {^̕ύXɂ郊RpC悤ɐ݌v

class MttFq {
// public member functions
public:
	// ctors
	MttFq(MttFq const& fq) : coeff1_(fq.coeff1_), coeff0_(fq.coeff0_) {}
	MttFq(MttFp const& fp0, MttFp const& fp1) : coeff1_(fp1), coeff0_(fp0) {}
	MttFq(MttFp const& fp0, char const* i_str1) :
		coeff1_(i_str1), coeff0_(fp0) {}
	explicit MttFq(MttFp const& fp0, int x1 = 0) :
		coeff1_(x1), coeff0_(fp0) {}
	MttFq(char const* i_str0, MttFp const& fp1) :
		coeff1_(fp1), coeff0_(i_str0) {}
	MttFq(int x0, MttFp const& fp1) : coeff1_(fp1), coeff0_(x0) {}
	MttFq(char const* i_str0, char const* i_str1) :
		coeff1_(i_str1), coeff0_(i_str0) {}
	explicit MttFq(char const* i_str0, int x1 = 0) :
		coeff1_(x1), coeff0_(i_str0) {}
	MttFq(int x0, char const* i_str1) : coeff1_(i_str1), coeff0_(x0) {}
	explicit MttFq(int x0 = 0, int x1 = 0) : coeff1_(x1), coeff0_(x0) {}

	// accessors
	MttFp const& getCoeff1() const { return coeff1_; }
	MttFp const& getCoeff0() const { return coeff0_; }
	char const* toStr() const;
	void show() const;
	
	// operators
	MttFq& operator=(MttFq const& fq) {
	  if(&fq != this) {
		  coeff1_ = fq.coeff1_;
		  coeff0_ = fq.coeff0_;
	  }
	  return *this;
	}
	MttFq& operator=(MttFp const& fp) {
	  coeff1_.clear();
	  coeff0_ = fp;
	  return *this;
	}
	MttFq& operator=(char const* i_str) {
	  coeff1_.clear();
	  coeff0_ = i_str;
	  return *this;
	}
	MttFq& operator=(int x0) {
	  coeff1_.clear();
	  coeff0_ = x0;
	  return *this;
	}
	MttFq& operator+=(MttFq const& fq) {
	  coeff1_ += fq.coeff1_;
	  coeff0_ += fq.coeff0_;
	  return *this;
	}
	MttFq& operator+=(MttFp const& fp) {
	  coeff0_ += fp;
	  return *this;
	}
	MttFq& operator-=(MttFq const& fq) {
	  coeff1_ -= fq.coeff1_;
	  coeff0_ -= fq.coeff0_;
	  return *this;
	}
	MttFq& operator-=(MttFp const& fp) {
	  coeff0_ -= fp;
	  return *this;
	}
	MttFq& operator*=(MttFq const& fq);
	MttFq& operator*=(MttFp const& fp) {
	  coeff1_ *= fp;
	  coeff0_ *= fp;
	  return *this;
	}
	MttFq& operator^=(MttBits const& exponent); // power
	
	// operations
	void inverse();
	void normalize() const { coeff1_.normalize(); coeff0_.normalize(); }
	void clear() { coeff1_.clear(); coeff0_.clear(); }
	void changeSign() { coeff1_.changeSign(); coeff0_.changeSign(); }
	void adjoint() { coeff1_.changeSign(); }
	bool isZero() const { return coeff1_.isZero() && coeff0_.isZero(); }
	void sqrt();
	static void Add(MttFq const& src1, MttFq const& src2, MttFq& dst)
		{ MttFp::Add(src1.coeff0_, src2.coeff0_, dst.coeff0_);
		  MttFp::Add(src1.coeff1_, src2.coeff1_, dst.coeff1_); }
	static void Add(MttFp const& src1, MttFq const& src2, MttFq& dst)
		{ MttFp::Add(src1, src2.coeff0_, dst.coeff0_);
		  dst.coeff1_ = src2.coeff1_; }
	static void Add(MttFq const& src1, MttFp const& src2, MttFq& dst)
		{ MttFp::Add(src1.coeff0_, src2, dst.coeff0_);
		  dst.coeff1_ = src1.coeff1_; }
	static void Sub(MttFq const& src1, MttFq const& src2, MttFq& dst)
		{ MttFp::Sub(src1.coeff0_, src2.coeff0_, dst.coeff0_);
		  MttFp::Sub(src1.coeff1_, src2.coeff1_, dst.coeff1_); }
	static void Sub(MttFp const& src1, MttFq const& src2, MttFq& dst)
		{ MttFp::Sub(src1, src2.coeff0_, dst.coeff0_);
		  dst.coeff1_ = src2.coeff1_; dst.coeff1_.changeSign(); }
	static void Sub(MttFq const& src1, MttFp const& src2, MttFq& dst)
		{ MttFp::Sub(src1.coeff0_, src2, dst.coeff0_);
		  dst.coeff1_ = src1.coeff1_; }
	static void Mul(MttFq const& src1, MttFq const& src2, MttFq& dst) {
		MttFq ret(src1);
		ret *= src2;
		dst = ret;
		// dst  src* ƂAEg
//		dst = src1; dst *= src2;
	}
	static void Mul(MttFp const& src1, MttFq const& src2, MttFq& dst) {
		MttFp::Mul(src1, src2.coeff0_, dst.coeff0_);
		MttFp::Mul(src1, src2.coeff1_, dst.coeff1_);
	}
	static void Mul(MttFq const& src1, MttFp const& src2, MttFq& dst) {
		MttFp::Mul(src2, src1.coeff0_, dst.coeff0_);
		MttFp::Mul(src2, src1.coeff1_, dst.coeff1_);
	}
// the member
protected:
	// coeff0_ + (-1) coeff1_;
	MttFp coeff1_, coeff0_;
};

} // end of namespace

#endif
