#ifndef MTT_IMP_UA_H_
#define MTT_IMP_UA_H_

extern "C" {
bool ua_add(uint32 const* p_src1, uint32 const* p_src2, uint32* p_dst, int len);
bool ua_sub(uint32 const* p_src1, uint32 const* p_src2, uint32* p_dst, int len);
void ua_mul_s(uint32 const* p_src, uint32 src, uint32* p_dst, int len);
void ua_mul_add_s(uint32 const* p_src, uint32 src, uint32* p_dst, int len);
void ua_square_s(uint32 const* p_src, uint32* p_dst, int len);
void ua_square_add_s(uint32 const* p_src, uint32* p_dst, int len);
uint32 ua_div_s(uint32 const* p_src, uint32 src, uint32* p_dst, int len);
}

namespace MTT {

// ---------------------------------------------------------------------
// arraỷZ

template<typename T>
inline void ua_clear(T* ua, int len)
{
	for (int i = 0; i < len; i++) *ua++ = 0;
}

template<typename T>
inline void ua_copy(const T *p_src, T *p_dst, int len)
{
	for (int i = 0; i < len; i++) *p_dst++ = *p_src++;
}

template<typename T>
inline bool ua_is_zero(const T *ua, int len)
{
	int i;
	for (i = 0; i < len && *ua++ == 0; i++);
	return i == len;
}

template<typename T>
inline int ua_cmp(const T* p_src1, const T* p_src2, int len)
{
	for (int i = len - 1; i >= 0; i--) {
		if (p_src1[i] > p_src2[i]) return  1; /* large */
		if (p_src1[i] < p_src2[i]) return -1; /* small */
	}
	return 0; /* equal */
}

template<typename T>
inline void ua_set_uint(T data, T *p, int len)
{
	ua_clear(p, len);
	p[0] = data;
}

inline int ua_get_length(uint32 const* p, int len)
{
	for (len--; len >= 0 && p[len] == 0; len--);
	return len + 1;
}

inline int ua_get_bitlength(uint32 const* p, int len)
{
	len = ua_get_length(p, len) - 1;
	if (len < 0) return 0;
	int i;
	uint32 d = p[len];
	for (i = 0; i< 32 && d != 0; i++, d >>= 1);
	return len * 32 + i;
}


//nvf / nvf = nvf + ]nvf
inline void ua_div(uint32 const* p_a, uint32 const* p_b, uint32* p_q, uint32* p_r, int len)
{
	//QlFߎvZVC6++lj󂷂œK鎖
	int cmp = ua_cmp(p_a, p_b, len);
	if (cmp < 0) {
		//0 ]a
		ua_copy(p_a, p_r, len);
		ua_clear(p_q, len);
		return;
	}
	if (cmp == 0) {
		//1 ]0
		ua_set_uint(uint32(1), p_q, len);
		ua_clear(p_r, len);
		return;
	}

	int len_a = ua_get_length(p_a, len);
	int len_b = ua_get_length(p_b, len);
	if (len_b == 0) {
		//0ł̊Z
		ua_clear(p_q, len);
		ua_clear(p_r, len);
		return;
	}

	if (len_b == 1) {
		//XJ[Z
		ua_set_uint( ua_div_s(p_a, p_b[0], p_q, len_a), p_r, len );
		ua_clear(&p_q[len_a], len-len_a);
		return;
	}

	//ʌ̗]p1vf
	uint32 *a = new uint32[len_a+1];
	uint32 *b = new uint32[len_b+1];
	uint32 *t = new uint32[len_b+1];
	ua_copy(p_a, a, len_a); a[len_a] = 0;
	ua_copy(p_b, b, len_b); b[len_b] = 0;

	// 2^32
	const double radix = ((double)((uint32)1<<31) * 2.0);
	const double r_radix = 1.0/radix;
	//b̋ߎl(b菬,ȂׂLrbg𑽂)
	double apx_b;
	apx_b  = (double)b[len_b-1];
	apx_b += (len_b > 1) ? (double)b[len_b-2]*r_radix : 0.0;
	apx_b += (len_b > 2) ? (double)b[len_b-3]*r_radix*r_radix : 0.0;
	double r_apx_b = 1.0/apx_b;

	//ča̎c茅-b̌
	int len_r = len_a-len_b;
	ua_clear(&p_q[len_r], len-len_r);
	while( len_r>=0 ) {
		uint32 q;
		if ((a[len_a] | a[len_a-1]) == 0 ) {
			q = 0;
		} else {
			//a̋ߎl(O̗]E݌vZΏۂ̌Ẻʌl,傫)
			double apx_a;
			apx_a  = (double)a[len_a]*radix;
			apx_a += (double)a[len_a-1];
			apx_a += (len_a > 1) ? ((double)a[len_a-2] + 1.0)*r_radix : r_radix;
			//݂̌̏̋ߎ(傫߂Ɏ,X1傫x)
			double apx_q = (apx_a * r_apx_b);
			q = (uint32)apx_q + 1;
			if ((double)q < apx_q ) q = 0xffffffff;

			//ab*q
			bool carry = false;
			if (q < 5 ) {
				for(uint32 i=0; i<q; i++) carry |= ua_sub(&a[len_r], b, &a[len_r], len_b+1);
			} else {
				//ua_mul_sub_s~Ƃ
				ua_mul_s(b, q, t, len_b);
				carry = ua_sub(&a[len_r], t, &a[len_r], len_b+1);
			}

			//q傫ꍇ͏C(2̕␔\L)
			while (carry) {
				q--;
				carry = !ua_add(&a[len_r], b, &a[len_r], len_b+1);
			}
		}
		p_q[len_r--] = q;
		len_a--;
	}
	ua_copy(a, p_r, len_b);
	ua_clear(&p_r[len_b], len-len_b);

	delete [] a;
	delete [] b;
	delete [] t;
}

} // end of namespace

#endif
