#ifndef   MTT_ELLIPTIC_POWER_TMPL_H
#define   MTT_ELLIPTIC_POWER_TMPL_H

// ===========================================================================
//  elliptic_power_tmpl.h -- ȉ~Ȑ̔{Z̎
// ---------------------------------------------------------------------------
// 
//  {Z͕̎GɂȂ̂ŕ(c.f. elliptic_tmpl.h)


#include "types_tmpl.h"

namespace mttec {
#define Fq Fp
// ===========================================================================
//  class power_ : calculate power
template<class Fp, class Fm>
class power_ : public ec_<Fp> {
public:
	// ----------------------------------------------------------------------
	//  point generation
	void genPoint(mttfq::power_<Fp, Fm>& F, ep_<Fq>& P);
	void genMTorsionPoint(mttfq::power_<Fp, Fm>& F, ep_<Fq>& P);
	
	// ----------------------------------------------------------------------
	//  multiplication of an elliptic point
	template<class Bits>
	void pow(ep_<Fq>const& s, Bits const& e, ep_<Fq>& d);
	void pow(ep_<Fp>const& s, Fp const& e, ep_<Fq>& d) {
		e.getR().getBits(expP_);
		pow(s, expP_, d);
	}
	void pow(ep_<Fq>const& s, Fm const& e, ep_<Fq>& d) {
		e.getR().getBits(expM_);
		pow(s, expM_, d);
	}

	// ----------------------------------------------------------------------
	//  fast tate pairing
	template<class Bits>
	void powFast(miller_<Fq>const& s, Bits const& e, miller_<Fq>& d);
	void powFast(miller_<Fq>const& s, Fp const& e, miller_<Fq>& d) {
		e.getR().getBits(expP_);
		powFast(s, expP_, d);
	}
	void powFast(miller_<Fq>const& s, Fm const& e, miller_<Fq>& d) {
		e.getR().getBits(expM_);
		powFast(s, expM_, d);
	}
	
	// ----------------------------------------------------------------------
	//  full tate pairing / weil pairing
	template<class Bits>
	void pow(miller_<Fq>const& s, Bits const& e, miller_<Fq>& d);
	void pow(miller_<Fq>const& s, Fp const& e, miller_<Fq>& d) {
		e.getR().getBits(expP_);
		pow(s, expP_, d);
	}
	void pow(miller_<Fq>const& s, Fm const& e, miller_<Fq>& d) {
		e.getR().getBits(expM_);
		pow(s, expM_, d);
	}
	
	// ----------------------------------------------------------------------
	//  accessor
protected:
	miller_<Fq> wmi_[20];
	typename Fp::value_type::Bits expP_;
	typename Fm::value_type::Bits expM_;
};

// ---------------------------------------------------------------------------
//  power: Q = eP
template<class Fp, class Fm>
template<class B>
void power_<Fp, Fm>::pow(ep_<Fq>const& P, B const& b, ep_<Fq>& Q)
{
	typename MttWindow<B>::Exponent e(b);
	int mbits = e.numBits();
	ep_<Fp> *const wep_ = ec_<Fp>::wep_;
	if(1 == mbits) {
		if(e.isBitOn(0)) {
			if(&P != &Q) Q = P;
			return;
		}
		--mbits;
	}
	
	wep_[0] = P;
	Q.affine = true;
	Q.x = 1;
	Q.y = 2;
	Q.z = 0;
	if(0 == mbits) return;
	
	if(mbits > 12) {
	   // prepare the window
	   dbl(wep_[0], wep_[19]);
	   for(int i = 0; i < 10; ++i) add(wep_[i], wep_[19], wep_[i+1]);
	   add(wep_[7],  wep_[0],  wep_[18]); // 16P
	   dbl(wep_[18], wep_[17]);		   // 32P
	   add(wep_[17], wep_[0],  wep_[11]); // 33P
	   add(wep_[17], wep_[2],  wep_[12]); // 37P
	   add(wep_[17], wep_[4],  wep_[13]); // 41P
	   add(wep_[17], wep_[11], wep_[14]); // 65P
	   add(wep_[17], wep_[12], wep_[15]); // 69P
	   add(wep_[17], wep_[13], wep_[16]); // 73P
	   add(wep_[18], wep_[14], wep_[17]); // 81P
	   add(wep_[18], wep_[15], wep_[18]); // 85P
	   
	   /* @@ Montgomery affinize of wep_[0] - wep_[18] */
	   
	   typename MttWindow<B>::Signed57 binary(e);
	   binary.begin();
	   int op = 0, num = 0;
	   static int rev_p[14] =
		   {11, 12, 13, 0, 0, 0, 0, 0, 14, 15, 16, 0, 17, 18};
	   binary.next_op(op, num); 
	   while(op == binary.op_double) binary.next_op(op, num);
	   Q = wep_[num<32? (num-1)/2: rev_p[(num-33)/4]];
	   if(op == binary.op_sub) Q.y.negate();
	   
	   while(binary.next_op(op, num)) {
		   if(op == binary.op_add) {
			   add(Q, wep_[num<32? (num-1)/2: rev_p[(num-33)/4]], Q);
		   } else if(op == binary.op_sub) {
			   sub(Q, wep_[num<32? (num-1)/2: rev_p[(num-33)/4]], Q);
		   } else {
			   while(num--) dbl(Q, Q);
		   }
	   }
	} else {
		while(--mbits) {
			if(e.isBitOn(mbits)) add(Q, wep_[0], Q);
			dbl(Q, Q);
		}
		if(e.isBitOn(mbits)) add(Q, wep_[0], Q);
	}
}


// ===========================================================================
//  end of namespace
#undef Fq
}

#endif /* MTT_ELLIPTIC_POWER_TMPL_H */
