#ifndef   MTT_MILLER_TMPL_H
#define   MTT_MILLER_TMPL_H

// ===========================================================================
//  miller_tmpl.h -- Milleř̎
// ---------------------------------------------------------------------------
// 
//   Milleř̂ɂ f_P([Q]-[R])܂f_P(Q)߂
//  setDivisor, add, sub, pow  full version  milleřłA
//  setPoint, addFast, subFast, powFast ͍ version  milleřł
//  o[WłPF_p̑ȉ~Ȑ̓_łKvAQ͂
//  distortionɂȂ΂ȂȂB

#include "types_tmpl.h"

namespace mttec {
#define Fq Fp
// ===========================================================================
//  set a point and affinize assuming Q != [0]
template<class T>
inline void ec_<T>::setPoint(ep_<T>const& Q)
{
	if(Q.affine) {
		wep_[0] = Q;
		return;
	}
	wfq_[0] = inverse(Q.z);
//	wfq_[2] = wfq_[0]*(wfq_[1] = wfq_[0]*wfq_[0]);
	wfq_[1] = wfq_[0] * wfq_[0];
	wfq_[2] = wfq_[1] * wfq_[0];
	wep_[0].affine = true;
	wep_[0].x = Q.x * wfq_[1];
	wep_[0].y = Q.y * wfq_[2];
	wep_[0].z = 1;
}


// ===========================================================================
//  set a divisor [Q]-[R] and affinize assuming Q,R != [0]
template<class T>
inline void ec_<T>::setDivisor(ep_<T>const& Q, ep_<T>const& R)
{
	if(Q.affine) {
		wep_[0] = Q;
		wep_[1] = R;
		if(!R.affine) affinize(wep_[1]);
		return;
	}
	if(R.affine) {
		wep_[1] = R;
		setPoint(Q);
		return;
	}
	
	wfq_[0] = inverse(Q.z*R.z);
	wfq_[1] = wfq_[0]*R.z;
//	wfq_[3] = wfq_[1]*(wfq_[2] = wfq_[1]*wfq_[1]);
	wfq_[2] = wfq_[1] * wfq_[1];
	wfq_[3] = wfq_[2] * wfq_[1];
	wep_[0].affine = true;
	wep_[0].x = Q.x * wfq_[2];
	wep_[0].y = Q.y * wfq_[3];
	wep_[0].z = 1;
	wfq_[1] = wfq_[0]*Q.z;
//	wfq_[3] = wfq_[1]*(wfq_[2] = wfq_[1]*wfq_[1]);
	wfq_[2] = wfq_[1] * wfq_[1];
	wfq_[3] = wfq_[2] * wfq_[1];
	wep_[1].affine = true;
	wep_[1].x = R.x * wfq_[2];
	wep_[1].y = R.y * wfq_[3];
	wep_[1].z = 1;
}


// ===========================================================================
//  miller double (fast version)
template<class T>
inline void ec_<T>::dblFast(miller_<T>const& s, miller_<T>& d)
{
	// P_s == [0]
	s.z.normalize();
	if(s.z.isZero()) {
		if(&s != &d) d = s;
		return;
	}
	
	// 2P_s == [0]
	s.y.normalize();
	if(s.y.isZero()) {
		d.affine = true;
		d.x = 1;
		d.y = 2;
		d.z = 0;
		d.num = s.num*s.num;
		return;
	}
	d.affine = false;	
	
	// w0 <- lambda_n := 3(x^2 - z^4),  w4 <- xQ z^2 - x
	// w5 <- yQ yz^3,  Z <- yz
	if(s.affine) {
	   wfq_[1] = s.x*s.x;
		wfq_[0] = wfq_[1] + wfq_[1] + wfq_[1] + (-3); /* - 3 Ƃ͂Ȃ */
	   wfq_[4] = wep_[0].x - s.x;
	   wfq_[5] = wep_[0].y * s.y;
	   d.z = s.y;
	} else {
		wfq_[1] = s.z*s.z;
		d.z = s.y*s.z;
		wfq_[4] = wep_[0].x*wfq_[1] - s.x;
		wfq_[5] = wep_[0].y*d.z*wfq_[1];
		wfq_[2] = s.x + wfq_[1];
		wfq_[3] = s.x - wfq_[1];
		wfq_[0] = wfq_[2] * wfq_[3];
		wfq_[1] = wfq_[0] + wfq_[0];
		wfq_[0] += wfq_[1];
	}
	
	// w1 <- y^2,  w2 <- xy^2,  w3 <- lambda_n^2
	wfq_[1] = s.y * s.y;
	wfq_[2] = wfq_[1] * s.x;
	wfq_[3] = wfq_[0]*wfq_[0];
	
	// f_d = (f_s)^2 g_{P_s,P_S}(Q)
	wfq_[4] += wfq_[4];
//	d.num = s.num*s.num*(wfq_[0]*wfq_[4] - wfq_[5] + wfq_[1]);
	wfq_[4] *= wfq_[0];
	wfq_[4] -= wfq_[5];
	wfq_[4] += wfq_[1];
	/* wfq_[4] * s.num * s.num ƂȂ */
	d.num = s.num * s.num * wfq_[4];
	
	// X <- lambda_n^2 - 2xy^2
	d.x = wfq_[3] - (wfq_[2] + wfq_[2]);
//	d.x = wfq_[2] + wfq_[2] - wfq_[3]; /* QQQ:Ƒʖ */

	// Y <- 2(3xy^2 - lambda_n^2)lambda_n - y^4
	d.y = (wfq_[2] - d.x)*wfq_[0];
	wfq_[1] *= wfq_[1];
	d.y += d.y - wfq_[1];
}


// ===========================================================================
//  miller adding (fast version)
template<class T>
inline void ec_<T>::addFast(
	miller_<T>const& s1, miller_<T>const& s2, miller_<T>& d)
{
	// P_s1 is identical to P_s2
	if(&s1 == &s2) {
		dblFast(s1, d);
		return;
	}
	
	// P_s1 == [0]
	s1.z.normalize();
	if(s1.z.isZero()) {
		if(&s2 != &d) d = s2;
		return;
	}
	
	// P_s2 == [0]
	s2.z.normalize();
	if(s2.z.isZero()) {
		if(&s1 != &d) d = s1;
		return;
	}
	
	// w2 <- lambda_d := x1 z2^2 - z1^2 x2
	// w3 <- lambda_n := y1 z2^3 - z1^3 y2
	// w0 <- mu_d	 := x1 z2^2 + z1^2 x2
	// w1 <- mu_n	 := y1 z2^3 + z1^3 y2
	// w5 <- z1 z2
	if(s1.affine) {
		if(s2.affine) {
			wfq_[2] = s1.x - s2.x;
			wfq_[3] = s1.y - s2.y;
			wfq_[0] = s1.x + s2.x;
			wfq_[1] = s1.y + s2.y;
			wfq_[5] = 1;
		} else {
//			wfq_[1]  = s2.z*(wfq_[0] = s2.z*s2.z);
			wfq_[0] = s2.z * s2.z;
			wfq_[1] = wfq_[0] * s2.z;
			wfq_[0] *= s1.x;
			wfq_[1] *= s1.y;
			wfq_[2]  = wfq_[0] - s2.x;
			wfq_[3]  = wfq_[1] - s2.y;
			wfq_[0] += s2.x;
			wfq_[1] += s2.y;
			wfq_[5]  = s2.z;
		}
	} else {
		// w0 <- z1^2 x2,  w1 <- z1^3 y2
//		wfq_[1] = s1.z*(wfq_[2] = s1.z*s1.z);
		wfq_[2] = s1.z * s1.z;
		wfq_[1] = wfq_[2] * s1.z;
		wfq_[0] = s2.x*wfq_[2];
		wfq_[1] *= s2.y;
		if(s2.affine) {
			wfq_[2]  = s1.x - wfq_[0];
			wfq_[0] += s1.x;
			wfq_[3]  = s1.y - wfq_[1];
			wfq_[1] += s1.y;
			wfq_[5]  = s1.z;
		} else {
//			wfq_[3]  = s2.z*(wfq_[2] = s2.z*s2.z);
			wfq_[2] = s2.z * s2.z;
			wfq_[3] = wfq_[2] * s2.z;
			wfq_[4]  = s1.x*wfq_[2];
			wfq_[2]  = wfq_[4] - wfq_[0];
			wfq_[0] += wfq_[4];
			wfq_[4]  = s1.y*wfq_[3];
			wfq_[3]  = wfq_[4] - wfq_[1];
			wfq_[1] += wfq_[4];
			wfq_[5]  = s1.z*s2.z;
		}
	}
	
	// P_s1 == P_s2,  P_s1 == -P_s2
	wfq_[2].normalize();
	if(wfq_[2].isZero()) {
		wfq_[3].normalize();
		if(wfq_[3].isZero()) dblFast(s1, d);
		else {
			d.affine = true;
			d.x = 1;
			d.y = 2;
			d.z = 0;
			d.num = s1.num*s2.num;
		}
		return;
	}
	d.affine = false;
	
	// w4 <- 2lambda_d,  Z <- 2z1 z2 lambda_d
	wfq_[4] = wfq_[2] + wfq_[2];
	d.z = wfq_[4]*wfq_[5];
	
	// w6 <- 2(z1 z2)^2,  w7 <- 2(z1 z2)^3
	wfq_[6]  = wfq_[5]*wfq_[5];
	wfq_[6] += wfq_[6];
	wfq_[7]  = wfq_[6]*wfq_[5];
	
	// w6 <- lambda_n(2 (z1 z2)^2 xQ - mu_d)
	// w7 <- lambda_d(2 (z1 z2)^3 yQ - mu_n)
	wfq_[6] *= wep_[0].x;
	wfq_[6] -= wfq_[0];
	wfq_[6] *= wfq_[3];

	wfq_[7] *= wep_[0].y;
	wfq_[7] -= wfq_[1];
	wfq_[7] *= wfq_[2];

	// f_d = f_s1*f_s2*g_{P_s1,P_s2}(Q)
	d.num = s1.num*s2.num*(wfq_[6]-wfq_[7]);
//	d.num = (wfq_[6] - wfq_[7]) * s1.num * s2.num;

	// w4 <- 4lambda_d^3,  w0 <- 4lambda_d^2 mu_d
	wfq_[4] *= wfq_[4];
	wfq_[0] *= wfq_[4];
	wfq_[4] *= wfq_[2];
	
	// X <- -4lambda_d^2 mu_d + lambda_n^2
	d.x = wfq_[3]*wfq_[3] - wfq_[0];
	
	// Y <- (w0 - 2X)w3 - w4*w1
	wfq_[0] -= d.x + d.x;
//	d.y = wfq_[0]*wfq_[3] - wfq_[4]*wfq_[1];
	wfq_[0] *= wfq_[3];
	wfq_[4] *= wfq_[1];
	d.y = wfq_[0] - wfq_[4];
}


// ===========================================================================
//  miller subtracting (fast version)
template<class T>
inline void ec_<T>::subFast(
	miller_<T>const& s1, miller_<T>const& s2, miller_<T>& d)
{
	// P_s1 is identical to P_s2
	if(&s1 == &s2) {
		d.affine = true;
		d.x = 1;
		d.y = 2;
		d.z = 0;
		d.num = 1;
		return;
	}
	
	// P_s2 == [0]
	s2.z.normalize();
	if(s2.z.isZero()) {
		if(&s1 != &d) d = s1;
		return;
	}
	
	// P_s1 == [0]
	s1.z.normalize();
	if(s1.z.isZero()) {
		if(&s2 != &d) d = s2;
		d.y.negate();
		d.num.frobenius();
		return;
	}
	
	// w2 <- lambda_d := x1 z2^2 - z1^2 x2
	// w3 <- lambda_n := y1 z2^3 - z1^3 (-y2)
	// w0 <- mu_d	 := x1 z2^2 + z1^2 x2
	// w1 <- mu_n	 := y1 z2^3 + z1^3 (-y2)
	// w5 <- z1 z2;
	if(s1.affine) {
		if(s2.affine) {
			wfq_[2] = s1.x - s2.x;
			wfq_[3] = s1.y + s2.y;
			wfq_[0] = s1.x + s2.x;
			wfq_[1] = s1.y - s2.y;
			wfq_[5] = 1;
		} else {
			wfq_[0] = s2.z * s2.z;
			wfq_[1] = wfq_[0] * s2.z;
			wfq_[0] *= s1.x;
			wfq_[1] *= s1.y;
			wfq_[2]  = wfq_[0] - s2.x;
			wfq_[3]  = wfq_[1] + s2.y;
			wfq_[0] += s2.x;
			wfq_[1] -= s2.y;
			wfq_[5]  = s2.z;
		}
	} else {
		// w0 <- z1^2 x2,  w3 <- z1^3 y2
		wfq_[2] = s1.z * s1.z;
		wfq_[3] = wfq_[2] * s1.z;
		wfq_[0] = s2.x*wfq_[2];
		wfq_[3] *= s2.y;
		if(s2.affine) {
			wfq_[2]  = s1.x - wfq_[0];
			wfq_[0] += s1.x;
			wfq_[1]  = s1.y - wfq_[3];
			wfq_[3] += s1.y;
			wfq_[5]  = s1.z;
		} else {
			wfq_[2] = s2.z * s2.z;
			wfq_[1] = wfq_[2] * s2.z;
			wfq_[4]  = s1.x*wfq_[2];
			wfq_[2]  = wfq_[4] - wfq_[0];
			wfq_[0] += wfq_[4];
			wfq_[4]  = s1.y*wfq_[1];
			wfq_[1]  = wfq_[4] - wfq_[3];
			wfq_[3] += wfq_[4];
			wfq_[5]  = s1.z*s2.z;
		}
	}
	
	// P_s1 == -P_s2,  P_s1 == P_s2
	wfq_[2].normalize();
	if(wfq_[2].isZero()) {
		wfq_[3].normalize();
		if(wfq_[3].isZero()) dblFast(s1, d);
		else {
			d.affine = true;
			d.x = 1;
			d.y = 2;
			d.z = 0;
			d.num = s1.num * frobenius(s2.num);
		}
		return;
	}
	d.affine = false;
	
	// w4 <- 2lambda_d,  Z <- 2z1 z2 lambda_d
	wfq_[4] = wfq_[2] + wfq_[2];
	d.z = wfq_[4]*wfq_[5];
	
	// w6 <- 2(z1 z2)^2,  w7 <- 2(z1 z2)^3
	wfq_[6]  = wfq_[5]*wfq_[5];
	wfq_[6] += wfq_[6];
	wfq_[7]  = wfq_[6]*wfq_[5];
	
	// w6 <- lambda_n(2 (z1 z2)^2 xQ - mu_d)
	// w7 <- lambda_d(2 (z1 z2)^3 yQ - mu_n)
	wfq_[6] *= wep_[0].x;
	wfq_[6] -= wfq_[0];
	wfq_[6] *= wfq_[3];

	wfq_[7] *= wep_[0].y;
	wfq_[7] -= wfq_[1];
	wfq_[7] *= wfq_[2];

	// f_d = f_s1*f_s2*g_{P_s1,P_s2}(Q)
//	d.num = s1.num*frobenius(s2.num)*(wfq_[6]-wfq_[7]);
	wfq_[6] -= wfq_[7];
	d.num = s1.num * frobenius(s2.num) * wfq_[6];

	// w4 <- 4lambda_d^3,  w0 <- 4lambda_d^2 mu_d
	wfq_[4] *= wfq_[4];
	wfq_[0] *= wfq_[4];
	wfq_[4] *= wfq_[2];
	
	// X <- -4lambda_d^2 mu_d + lambda_n^2
	d.x = wfq_[3]*wfq_[3] - wfq_[0];
	
	// Y <- (w0 - 2X)w3 - w4*w1
	wfq_[0] -= d.x + d.x;
//	d.y = wfq_[0]*wfq_[3] - wfq_[4]*wfq_[1];
	wfq_[0] *= wfq_[3];
	wfq_[4] *= wfq_[1];
	d.y = wfq_[0] - wfq_[4];
}


// ===========================================================================
//  power miller (fast version)
template<class Fp, class Fm>
template<class Bits>
void power_<Fp, Fm>::powFast(
	miller_<Fq>const& P, Bits const& b, miller_<Fp>& Q)
	
{
	typename MttWindow<Bits>::Exponent e(b);
	int mbits = e.numBits();
	if(1 == mbits) {
		if(e.isBitOn(0)) {
			if(&P != &Q) Q = P;
			return;
		}
		--mbits;
	}
	
	wmi_[0] = P;
	Q.affine = true;
	Q.x = 1;
	Q.y = 2;
	Q.z = 0;
	if(0 == mbits) return;
	
	if(mbits > 12) {
	   // prepare the window
	   dblFast(wmi_[0], wmi_[19]);
	   for(int i = 0; i < 10; ++i) addFast(wmi_[i], wmi_[19], wmi_[i+1]);
	   addFast(wmi_[7],  wmi_[0],  wmi_[18]); // 16P
	   dblFast(wmi_[18], wmi_[17]);		   // 32P
	   addFast(wmi_[17], wmi_[0],  wmi_[11]); // 33P
	   addFast(wmi_[17], wmi_[2],  wmi_[12]); // 37P
	   addFast(wmi_[17], wmi_[4],  wmi_[13]); // 41P
	   addFast(wmi_[17], wmi_[11], wmi_[14]); // 65P
	   addFast(wmi_[17], wmi_[12], wmi_[15]); // 69P
	   addFast(wmi_[17], wmi_[13], wmi_[16]); // 73P
	   addFast(wmi_[18], wmi_[14], wmi_[17]); // 81P
	   addFast(wmi_[18], wmi_[15], wmi_[18]); // 85P
	   
	   /* @@ Montgomery affinize of wmi_[0] - wmi_[18] */
	   
	   typename MttWindow<Bits>::Signed57 binary(e);
	   binary.begin();
	   int op = 0, num = 0;
	   static int rev_p[14] =
		   {11, 12, 13, 0, 0, 0, 0, 0, 14, 15, 16, 0, 17, 18};
	   binary.next_op(op, num); 
	   while(op == binary.op_double) binary.next_op(op, num);
	   Q = wmi_[num<32? (num-1)/2: rev_p[(num-33)/4]];
	   if(op == binary.op_sub) {
		   Q.y.negate();
		   Q.num.frobenius();
	   }
	   
	   while(binary.next_op(op, num)) {
		   if(op == binary.op_add) {
			   addFast(Q, wmi_[num<32? (num-1)/2: rev_p[(num-33)/4]], Q);
		   } else if(op == binary.op_sub) {
			   subFast(Q, wmi_[num<32? (num-1)/2: rev_p[(num-33)/4]], Q);
		   } else {
			   while(num--) dblFast(Q, Q);
		   }
	   }
	} else {
		while(--mbits) {
			if(e.isBitOn(mbits)) addFast(Q, wmi_[0], Q);
			dblFast(Q, Q);
		}
		if(e.isBitOn(mbits)) addFast(Q, wmi_[0], Q);
	}
}


// ===========================================================================
//  end of namespace
#undef Fq
}

#endif /* MTT_MILLER_TMPL_H */
