#ifndef MTT_FILE_ACCESS_H_
#define MTT_FILE_ACCESS_H_
/*!
	lǂݏANZbT
*/
#include "mtt_config.h"
#include <sstream>
#include "sha1.h"

namespace MTT {

static inline void put(const Fq& x)
{
	switch (x.getFlags()) {
	case Fq::REAL_IS_ZERO:
		printf("%si", x.getI().toStr());
		break;
	case Fq::IMAG_IS_ZERO:
		printf("%s", x.getR().toStr());
		break;
	case Fq::IS_ZERO:
		printf("0");
		break;
	default:
		// VC7.1ł͎̃R[h̓oÔœɕȂƂȂ
		// printf("%s+%si\n", x.getR().toStr(), x.getI().toStr());
		printf("%s", x.getR().toStr());
		printf("+%si", x.getI().toStr());
		break;
	}
}

static inline void put(const Fm& x)
{
	if (!(x.getFlags() & Fm::IS_ZERO)) {
		printf("%s", x.getR().toStr());
	} else {
		printf("0");
	}
}

static inline void put(const Em& P)
{
	printf("["); put(P.x); printf(":"); put(P.y);
	printf(":"); put(P.z); printf("]");
}

#define PUT(x) do { printf(#x " = "); put(x); puts(""); } while(0)


/*
	QQQ:yŴ͕ݕۑ
	y == sqrt(4x(x^2 - 3)
	ȉ~̌`ʂ܂ŉê͂܂ǂƂ肠
*/

static inline std::ostream& operator<<(std::ostream& os, const Fm& x)
{
	os << "Fm=" << x.getR().toStr() << std::endl;
	return os;
}
static inline std::istream& operator>>(std::istream& is, Fm& x)
{
	std::string line;
	std::getline(is, line);
	if (line.find("Fm=") == 0) {
		x = Fm(line.substr(3).c_str());
	} else {
		is.setstate(std::ios_base::failbit);
	}
	return is;
}

static inline std::ostream& operator<<(std::ostream& os, const Fq& x)
{
	os << "Fq=" << x.getR().toStr() << ','; /* toStr()͓ɎgȂ? */
	os << x.getI().toStr() << std::endl;
	return os;
}
static inline std::istream& operator>>(std::istream& is, Fq& x)
{
	std::string line;
	std::getline(is, line);
	bool isOK = false;
	if (line.find("Fq=") == 0) {
		line = line.substr(3);
		size_t pos = line.find(',');
		if (pos != std::string::npos) {
			x = Fq(line.substr(0, pos).c_str(), line.substr(pos + 1).c_str());
			isOK = true;
		}
	}
	if (!isOK) is.setstate(std::ios_base::failbit);
	return is;
}

/* get 2 * sqrt(x * (x * x - 3)) */
static inline Fq calcYby(const Fq& x)
{
	Fq y = x;
	y *= y; y -= 3; y *= x; // x(x^2 - 3)
	MttPower<ModP, ModM>::Fq F;
	F.sqrt(y, y);
	y += y; // sqrt(4x(x^2 - 3)
	return y;
}

static inline std::ostream& operator<<(std::ostream& os, const Em& p)
{
	MttPower<ModP, ModM>::ECurve E;
	Em tmp = p;
	E.affinize(tmp);
	Fq y = calcYby(tmp.x);
	y -= tmp.y;
	y.normalize();
	os << "Em=" << tmp.x.getR().toStr() << ',' << (y.isZero() ? "1" : "0") << std::endl;
	return os;
}
static inline std::istream& operator>>(std::istream& is, Em& p)
{
	std::string line;
	std::getline(is, line);
	bool isOK = false;
	if (line.find("Em=") == 0) {
		line = line.substr(3);
		size_t pos = line.find(',');
		if (pos != std::string::npos) {
			p.x = Fp(line.substr(0, pos).c_str());
			Fq y = calcYby(p.x);
			line = line.substr(pos + 1);
			if (line == "0" || line == "1") {
				if (line == "0") {
					y = -y;
				}
				p.y = y;
				p.z = Fp(1);
				isOK = true;
			}
		}
	}
	if (!isOK) is.setstate(std::ios_base::failbit);
	return is;
}

/* hex to integer */
template<class T>
static inline void convertHextToInteger(T *ret, const std::string& str, size_t size = 0)
{
	T v(uint32(0));
	if (size == 0) size = str.size();
	const T _16(16);
	static const std::string& tbl = "0123456789abcdef";
	for (size_t i = 0; i < size; i++) {
		v *= _16;
		size_t pos = tbl.find(str[i]);
		if (pos == std::string::npos) break;
		v += T(static_cast<int>(pos));
	}
	*ret = v;
}

/*
	̏lݒ
*/
template<class T>
static inline void getRandInteger(T *ret, const std::string& seed)
{
	assert(ret);
	MIE::SHA1 sha1(seed);
	convertHextToInteger(ret, sha1.toString());
}

} // end of namespace

#endif /* MTT_FILE_ACCESS_H_ */
