#ifndef MTT_IMP_H_
#define MTT_IMP_H_
/*
	===========================================================================
	mtt_imp.h -- modulo pymł̑{Z̎
	Ƃ肠ʂ͑field_fq.hǂ
	---------------------------------------------------------------------------
*/
#include "../base/common.h"
#include "../base/field_fq.h"

namespace MTT {

template<int n>
class MTTbits : MttBitsBase {
	mutable void* rep_val_;
public:
	// ----------------------------------------------------------------------
	//  1. ctor, dtor
	MTTbits() : rep_val_(0) {}
	~MTTbits() { MttBitsBase::sf_delete(rep_val_); rep_val_ = 0; }
	
	// ----------------------------------------------------------------------
	//  2. bits operation
	int numBits() const {
		return MttBitsBase::sf_num_bits(rep_val_);
	}
	int isBitOn(int i) const {
		return MttBitsBase::sf_bit(rep_val_, i);
	}

	// ----------------------------------------------------------------------
	//  3. bit size
	enum { SIZE_IN_BITS = n };
	
	void fromField(void* p) {
		MttBitsBase::sf_delete(rep_val_);
		rep_val_ = n > 300?
			MttBitsBase::sf_init_object_p(p):
			MttBitsBase::sf_init_object_m(p);
	}
};

template<class T, int n>
class MTTfield : public T {
private:
	void* p_value_;
public:
	// class initializer
	static void init();
	char const* toStr() const {
		if (p_value_) return T::sf_get_str(p_value_);
		return "0";
	}

	// ----------------------------------------------------------------------
	//  1. ctor, dtor
	MTTfield() : p_value_(0) {}
	explicit MTTfield(int r) : p_value_(T::sf_init_value(r)) {}
	MTTfield(const char* str) : p_value_(T::sf_init_str(str)) {}
	MTTfield(MTTfield const& r) : p_value_(T::sf_copy(r.p_value_)) {}
	~MTTfield() { T::sf_delete(p_value_); p_value_ = NULL; }
	
	// ----------------------------------------------------------------------
	//  2. assign functions
	static void set(int r, MTTfield& d) {
		sf_delete(d.p_value_);
		d.p_value_ = T::sf_init_value(r);
	}
	static void set(MTTfield const& s, MTTfield& d) {
		if (&d == &s) return;
		sf_delete(d.p_value_);
		d.p_value_ = sf_copy(s.p_value_);
	}
	static void set(const char *r, MTTfield& d) {
		sf_delete(d.p_value_);
		d.p_value_ = T::sf_init_str(r);
	}
	
	// ----------------------------------------------------------------------
	//  3. unary arithmetic functions
	static void inv(MTTfield const& s, MTTfield& d) {
		if (&d != &s) set(s, d);
		sf_inverse(d.p_value_);
	}
	static void neg(MTTfield const& s, MTTfield& d) {
		if (&d != &s) set(s, d);
		sf_change_sign(d.p_value_);
	}
	
	// ----------------------------------------------------------------------
	//  4. binary arithmetic functions
	static void add(MTTfield const& s, MTTfield& d) {
		sf_add(d.p_value_, s.p_value_, d.p_value_);
	}
	static void add(int s, MTTfield& d) {
		// nativeȎ͖
		MTTfield tmp = MTTfield(s);
		add(tmp, d);
	}
	static void sub(MTTfield const& s, MTTfield& d) {
		sf_sub(d.p_value_, s.p_value_, d.p_value_);
	}
	static void sub(int s, MTTfield& d) {
		MTTfield tmp = MTTfield(s);
		sub(tmp, d);
	}
	static void mul(MTTfield const& s, MTTfield& d) {
		sf_mul(d.p_value_, s.p_value_, d.p_value_);
	}
	static void mul(int s, MTTfield& d) {
		MTTfield tmp = MTTfield(s);
		sf_mul(d.p_value_, tmp.p_value_, d.p_value_);
	}
	
	// ----------------------------------------------------------------------
	//  5. trinary arithmetic functions
	static void add(MTTfield const& s1, MTTfield const& s2, MTTfield& d) {
		if (!d.p_value_) d.p_value_ = T::sf_init_value(0);
		sf_add(s1.p_value_, s2.p_value_, d.p_value_);
	}
	static void add(int s1, MTTfield const& s2, MTTfield& d) {
		MTTfield tmp = MTTfield(s1);
		add(tmp, s2, d);
	}
	static void add(MTTfield const& s1, int s2, MTTfield& d) {
		MTTfield tmp = MTTfield(s2);
		add(s1, tmp, d);
	}
	static void sub(MTTfield const& s1, MTTfield const& s2, MTTfield& d) {
		if (!d.p_value_) d.p_value_ = T::sf_init_value(0);
		sf_sub(s1.p_value_, s2.p_value_, d.p_value_);
	}
	static void sub(int s1, MTTfield const& s2, MTTfield& d) {
		MTTfield tmp = MTTfield(s1);
		sub(tmp, s2, d);
	}
	static void sub(MTTfield const& s1, int s2, MTTfield& d) {
		MTTfield tmp = MTTfield(s2);
		add(s1, tmp, d);
	}
	static void mul(MTTfield const& s1, MTTfield const& s2, MTTfield& d) {
		if (!d.p_value_) d.p_value_ = T::sf_init_value(0);
		sf_mul(s1.p_value_, s2.p_value_, d.p_value_);
	}
	static void mul(int s1, MTTfield const& s2, MTTfield& d) {
		MTTfield tmp = MTTfield(s1);
		mul(tmp, s2, d);
	}
	static void mul(MTTfield const& s1, int s2, MTTfield& d) {
		MTTfield tmp = MTTfield(s2);
		mul(s1, tmp, d);
	}
	
	// ----------------------------------------------------------------------
	//  6. other
	void normalize() const {
		T::sf_normalize(const_cast<void*>(p_value_));
	}
	bool isZero() const {
		return T::sf_is_zero(p_value_);
	}
	
	// ----------------------------------------------------------------------
	//  7. bits
	typedef MTTbits<n> Bits;
	void getBits(Bits& b) const {
		b.fromField(p_value_);
	}
	
	static Bits sqrtexp_, factor_m_, factor_co_m_;
	static inline const Bits& GetSqrtExp() { return sqrtexp_; }
	static inline const Bits& GetFactorM() { return factor_m_; }
	static inline const Bits& GetFactorCoM() { return factor_co_m_; }
};

typedef MTTfield<MttFpBase, 512> ModP;
typedef MTTfield<MttFmBase, 160> ModM;

template<class T, int n>
MTTbits<n> MTTfield<T, n>::sqrtexp_;

template<class T, int n>
MTTbits<n> MTTfield<T, n>::factor_m_;

template<class T, int n>
MTTbits<n> MTTfield<T, n>::factor_co_m_;

} // end of namespace

#endif // MTT_IMP_H_
