#define USE_GMP
#include <cstdlib>
#include "mtt_config.h"

void rand_ModP(ModP& modp) {
	ModP::set(rand(), modp);
	for(int i = 0; i < 15; ++i) {
		ModP::mul(0x40000000, modp);
		ModP::add(rand(), modp);
	}
}

void rand_ModM(ModM& modm) {
	ModM::set(rand(), modm);
	for(int i = 0; i < 5; ++i) {
		ModM::mul(0x40000000, modm);
		ModM::add(rand(), modm);
	}
}

void rand_Fq(MTT::Fq& fq) {
	rand_ModP(fq.getR());
	fq.getFlags() = fq.getR().isZero()? MTT::Fq::REAL_IS_ZERO: 0;
	rand_ModP(fq.getI());
	fq.getFlags() |= fq.getI().isZero()? MTT::Fq::IMAG_IS_ZERO: 0;
}

void rand_Fp(MTT::Fp& fp) {
	rand_ModP(fp.getR());
	fp.getFlags() = fp.getR().isZero()? MTT::Fq::IS_ZERO: MTT::Fq::IMAG_IS_ZERO;
}

void rand_Fm(MTT::Fm& fm) {
	rand_ModM(fm.getR());
	fm.getFlags() = fm.getR().isZero()? MTT::Fm::IS_ZERO: 0;
}

void rand_Fqm(MttPower<ModP, ModM>::Fq& F, MTT::Fq& fq) {
	rand_Fq(fq);
	fq *= inverse(frobenius(fq));
	F.pow(fq, ModP::GetFactorCoM(), fq);
}

void rand_Ep(MttPower<ModP, ModM>::ECurve& E, MttPower<ModP, ModM>::Fq& F, MTT::Em& P) {
	MTT::Em R;
	rand_Fp(P.x);
	E.genPoint(F, P);
	rand_Fp(R.x);
	E.genPoint(F, R);
	E.distort(R, R);
	E.add(P, R, P);
	if(rand() > RAND_MAX/2) E.affinize(P);
}

void rand_Em(MttPower<ModP, ModM>::ECurve& E, MttPower<ModP, ModM>::Fq& F, MTT::Em& P) {
	MTT::Em R;
	rand_Fp(P.x);
	E.genMTorsionPoint(F, P);
	rand_Fp(R.x);
	E.genMTorsionPoint(F, R);
	E.distort(R, R);
	E.add(P, R, P);
	if(rand() > RAND_MAX/2) E.affinize(P);
}

void rand_EmFp(MttPower<ModP, ModM>::ECurve& E, MttPower<ModP, ModM>::Fq& F, MTT::Em& P) {
	MTT::Em R;
	rand_Fp(P.x);
	E.genMTorsionPoint(F, P);
	if(rand() > RAND_MAX/2) E.affinize(P);
}

int main()
{
	initSystem();
	srand(time(0));
	
	int i;
	int const iter = 100;
	
	// test Fq -- power1
	// check x^a * x^b = x^(a + b)
	// where a, b, c in Fm
	//	   x in Fq[m]
	MTT::Fm a, b, c;
	MTT::Fq x, x1, x2, x3;
	MttPower<ModP,ModM>::Fq F;
	
	for(i = 0; i < iter; ++i) {
		rand_Fm(a);
		rand_Fm(b);
		c = a + b;
		rand_Fqm(F, x);
		F.pow(x, a, x1);
		F.pow(x, b, x2);
		F.pow(x, c, x3);
		x3 -= x1 * x2;
		x3.normalize();
		if(!x3.isZero()) {
			printf("x^a * x^b != x^(a + b) !! where\n");
			PUT(a); PUT(b); PUT(c); PUT(x);
			break;
		}
	}
	if(i >= iter) printf("x^a * x^b == x^(a + b) OK!!\n");
	
	// test Fq -- power2
	// check (x^a)^b = x^(a * b)
	// where a, b, c in Fm
	//	   x in Fq[m]
	for(i = 0; i < iter; ++i) {
		rand_Fm(a);
		rand_Fm(b);
		c = a * b;
		rand_Fqm(F, x);
		F.pow(x, a, x1);
		F.pow(x1, b, x2);
		F.pow(x, c, x3);
		x3 -= x2;
		x3.normalize();
		if(!x3.isZero()) {
			printf("(x^a)^b != x^(ab) !! where\n");
			PUT(a); PUT(b); PUT(c); PUT(x);
			F.pow(x, ModP::GetFactorCoM(), x);
			PUT(x);
			break;
		}
	}
	if(i >= iter) printf("(x^a)^b == x^(ab) OK!!\n");
	
	MTT::Em P, Q, R, S, T;
	MttPower<ModP,ModM>::ECurve E;	
	
	// test Elliptic1
	// check P + Q = Q + P
	// where P, Q in E[m]
	for(i = 0; i < iter; ++i) {	
		rand_Ep(E, F, P);
		rand_Ep(E, F, Q);
		E.add(P, Q, R); E.affinize(R);
		E.add(Q, P, S); E.affinize(S);
		x1 = R.x - S.x; x1.normalize();
		x2 = R.y - S.y; x2.normalize();
		if(!x1.isZero() || !x2.isZero()) {
			printf("P + Q != Q + P !! where\n");
			PUT(P);
			PUT(Q);
			break;
		}
	}
	if(i >= iter) printf("P + Q == Q + P  OK!!\n");	
	
	// test Elliptic2
	// check P + (Q + R) = (P + Q) + R
	for(i = 0; i < iter; ++i) {	
		rand_Ep(E, F, P);
		rand_Ep(E, F, Q);
		rand_Ep(E, F, R);
		E.add(Q, R, S); E.add(P, S, S); E.affinize(S);
		E.add(P, Q, T); E.add(T, R, T); E.affinize(T);
		x1 = T.x - S.x; x1.normalize();
		x2 = T.y - S.y; x2.normalize();
		if(!x1.isZero() || !x2.isZero()) {
			printf("P + (Q + R) != (P + Q) + R !! where\n");
			PUT(P);
			PUT(Q);
			PUT(R);
			break;
		}
	}
	if(i >= iter) printf("P + (Q + R) == (P + Q) + R   OK!!\n");	
	
	// test Elliptic2'
	// check P - (Q + R) = (P - Q) - R
	for(i = 0; i < iter; ++i) {	
		rand_Ep(E, F, P);
		rand_Ep(E, F, Q);
		rand_Ep(E, F, R);
		E.add(Q, R, S); E.sub(P, S, S); E.affinize(S);
		E.sub(P, Q, T); E.sub(T, R, T); E.affinize(T);
		x1 = T.x - S.x; x1.normalize();
		x2 = T.y - S.y; x2.normalize();
		if(!x1.isZero() || !x2.isZero()) {
			printf("P - (Q + R) != (P - Q) - R !! where\n");
			PUT(P);
			PUT(Q);
			PUT(R);
			break;
		}
	}
	if(i >= iter) printf("P - (Q + R) == (P - Q) - R   OK!!\n");	
	
	// test Elliptic3
	// check (P + P) = 2P
	for(i = 0; i < iter; ++i) {	
		rand_Ep(E, F, P);
		Q = P;
		E.add(P, Q, R); E.affinize(R);
		E.add(P, P, S); E.affinize(S);
		x1 = R.x - S.x; x1.normalize();
		x2 = R.y - S.y; x2.normalize();
		if(!x1.isZero() || !x2.isZero()) {
			printf("P + P != 2P !! where\n");
			PUT(P);
			break;
		}
	}
	if(i >= iter) printf("P + P == 2P   OK!!\n");	
	
	// test Elliptic4
	// check (P + P) + (P + P) = ((P + P) + P) + P
	for(i = 0; i < iter; ++i) {	
		rand_Ep(E, F, P);
		E.add(P, P, Q); E.add(Q, Q, S); E.affinize(S);
		E.add(Q, P, R); E.add(R, P, R); E.affinize(R);
		x1 = R.x - S.x; x1.normalize();
		x2 = R.y - S.y; x2.normalize();
		if(!x1.isZero() || !x2.isZero()) {
			printf("4P != (2P + P) + P !! where\n");
			PUT(P);
			break;
		}
	}
	if(i >= iter) printf("4P == (2P + P) + P   OK!!\n");
	
	// test Elliptic - power1
	// check (a + b)P = aP + bP
	for(i = 0; i < iter; ++i) {	
		rand_Em(E, F, P);
		rand_Fm(a);
		rand_Fm(b);
		c = a + b;
		E.pow(P, a, Q); E.pow(P, b, R);
		E.add(Q, R, S); E.affinize(S);
		E.pow(P, c, T); E.affinize(T);
		x1 = T.x - S.x; x1.normalize();
		x2 = T.y - S.y; x2.normalize();
		if(!x1.isZero() || !x2.isZero()) {
			printf("(a + b)P != aP + bP !! where\n");
			PUT(P);
			PUT(a);
			PUT(b);
			break;
		}
	}
	if(i >= iter) printf("(a + b)P == aP + bP   OK!!\n");
	
	// test Elliptic - power2
	// check a(bP) = (ab)P
	for(i = 0; i < iter; ++i) {	
		rand_Em(E, F, P);
		rand_Fm(a);
		rand_Fm(b);
		c = a * b;
		E.pow(P, b, Q); E.pow(Q, a, R); E.affinize(R);
		E.pow(P, c, S); E.affinize(S);
		x1 = R.x - S.x; x1.normalize();
		x2 = R.y - S.y; x2.normalize();
		if(!x1.isZero() || !x2.isZero()) {
			printf("(ab)P != a(bP) !! where\n");
			PUT(P);
			PUT(a);
			PUT(b);
			break;
		}
	}
	if(i >= iter) printf("(ab)P == a(bP)   OK!!\n");
	
	// test Pairing1
	// check e(P + Q, R) = e(P, R) * e(Q, R)
	for(i = 0; i < iter; ++i) {	
		rand_EmFp(E, F, P);
		rand_EmFp(E, F, Q);
		rand_EmFp(E, F, R);
		E.distort(R, R);
		MttPairing<ModP, ModM>::tateFast(E, F, P, R, x1);
		MttPairing<ModP, ModM>::tateFast(E, F, Q, R, x2);
		E.add(P, Q, S);
		MttPairing<ModP, ModM>::tateFast(E, F, S, R, x3);
		x3 -= x1 * x2;
		x3.normalize();
		if(!x3.isZero()) {
			printf("e(P + Q, R) != e(P, R) * e(Q, R)  !! where\n");
			PUT(P); PUT(Q); PUT(R);
			break;
		}
	}
	if(i >= iter) printf("e(P + Q, R) == e(P, R) * e(Q, R)  OK!!\n");
	
	// test Pairing1'
	// check e(P - Q, R) * e(Q, R) = e(P, R)
	for(i = 0; i < iter; ++i) {	
		rand_EmFp(E, F, P);
		rand_EmFp(E, F, Q);
		rand_EmFp(E, F, R);
		E.distort(R, R);
		MttPairing<ModP, ModM>::tateFast(E, F, P, R, x1);
		MttPairing<ModP, ModM>::tateFast(E, F, Q, R, x2);
		E.sub(P, Q, S);
		MttPairing<ModP, ModM>::tateFast(E, F, S, R, x3);
		x1 -= x3 * x2;
		x1.normalize();
		if(!x1.isZero()) {
			printf("e(P - Q, R) * e(Q, R) != e(P, R)  !! where\n");
			PUT(P); PUT(Q); PUT(R);
			break;
		}
	}
	if(i >= iter) printf("e(P - Q, R) * e(Q, R) != e(P, R)  OK!!\n");
	
	// test Pairing2
	// check e(P, Q + R) = e(P, Q) * e(P, R)
	for(i = 0; i < iter; ++i) {	
		rand_EmFp(E, F, P);
		rand_EmFp(E, F, Q);
		E.distort(Q, Q);
		rand_EmFp(E, F, R);
		E.distort(R, R);
		MttPairing<ModP, ModM>::tateFast(E, F, P, Q, x1);
		MttPairing<ModP, ModM>::tateFast(E, F, P, R, x2);
		E.add(Q, R, S);
		MttPairing<ModP, ModM>::tateFast(E, F, P, S, x3);
		x3 -= x1 * x2;
		x3.normalize();
		if(!x3.isZero()) {
			printf("e(P, Q + R) != e(P, Q) * e(P, R)  !! where\n");
			PUT(P); PUT(Q); PUT(R);
			break;
		}
	}
	if(i >= iter) printf("e(P + Q, R) == e(P, R) * e(Q, R)  OK!!\n");
	
	// test Pairing2'
	// check e(P, Q - R) * e(P, R) = e(P, Q)
	for(i = 0; i < iter; ++i) {	
		rand_EmFp(E, F, P);
		rand_EmFp(E, F, Q);
		E.distort(Q, Q);
		rand_EmFp(E, F, R);
		E.distort(R, R);
		MttPairing<ModP, ModM>::tateFast(E, F, P, Q, x1);
		MttPairing<ModP, ModM>::tateFast(E, F, P, R, x2);
		E.sub(Q, R, S);
		MttPairing<ModP, ModM>::tateFast(E, F, P, S, x3);
		x1 -= x3 * x2;
		x1.normalize();
		if(!x1.isZero()) {
			printf("e(P, Q - R) * e(P, R) != e(P, Q)  !! where\n");
			PUT(P); PUT(Q); PUT(R);
			break;
		}
	}
	if(i >= iter) printf("e(P, Q - R) * e(P, R) != e(P, Q)  OK!!\n");
	
	// test Pairing3
	// check e(aP, bQ) = e(P, Q)^(ab)
	for(i = 0; i < iter; ++i) {	
		rand_EmFp(E, F, P);
		rand_EmFp(E, F, Q);
		E.distort(Q, Q);
		rand_Fm(a);
		rand_Fm(b);
		c = a * b;
		MttPairing<ModP, ModM>::tateFast(E, F, P, Q, x1);
		E.pow(P, a, R);
		E.pow(Q, b, S);
		MttPairing<ModP, ModM>::tateFast(E, F, R, S, x2);
		F.pow(x1, c, x1);
		x1 -= x2;
		x1.normalize();
		if(!x1.isZero()) {
			printf("e(aP, bQ) != e(P, Q)^(ab)  !! where\n");
			PUT(P); PUT(Q); PUT(a); PUT(b);
			break;
		}
	}
	if(i >= iter) printf("(aP, bQ) != e(P, Q)^(ab)  OK!!\n");
	
	return 0;
}