/*!
	ÍW[
	@author $:Author$
	@version $Revision: 1.34 $
	@date $Date: 2007/08/31 05:59:05 $
*/
#define MIE_TC_EXPORT
#include "mie_timecapsule.h"
#include "mtt_config.h"
#include "mtt_file_access.h"
#include <fstream>
#include <ctime>

using namespace MTT;
using namespace MIE_TC;

#include "inner_util.h"
///////////////////////////////////////////////////////////////////////////////

void MIE_TC::init()
{
#ifdef __linux__
	/* Japanese ordinary time */
	setenv("TZ", "JST-9", 1);
#endif
	MTT::InitSystem();
}

/*!
	TimeInfoEm̕ۑ
*/
static bool SaveTimeInfoAndEm(std::string& data, const TimeInfo& timeInfo, const Em& P)
{
	std::ostringstream os;
	os << "time=" << timeInfo.get() << std::endl;
	os << P << std::endl;
	data = os.str();
	return true;
}

/*!
	TimeInfoEm̕
*/
static bool LoadTimeInfoAndEm(const std::string& data, TimeInfo *timeInfo, Em *P = 0)
{
	if (timeInfo == 0) return false;
	std::istringstream is(data);
	std::string line;
	std::getline(is, line);
	if (line.find("time=") != 0) return false;
	if (!timeInfo->set(line.c_str())) return false;
	if (P == 0) return true;
	is >> *P;
	return true;
}

///////////////////////////////////////////////////////////////////////////////

void PublicKey::deleteThis()
{
	deleteEm(&P_);
	deleteEm(&xP_);
}

void PublicKey::createThis()
{
	P_ = newEm();
	xP_ = newEm();
}

/*!
	̎x(PrivateKeyĂ)Ƃ(P, xP)ݒ肷
	@param seed [in] 
*/
bool PublicKey::init(const char *seed, const dummyFm *x)
{
	if (seed == 0) return false;
	MttPower<ModP, ModM>::ECurve E;
	MttPower<ModP, ModM>::Fq F;
	Em *P;

	// PKɐ
	Fq r;
	getRandInteger(&r, seed);

	P = _Em(P_);
	P->x = r;
	P->x.normalize();
	E.genMTorsionPoint(F, *P);
	E.affinize(*P);
	P->x.normalize();
	P->y.normalize();

	// xP = x * P;
	E.pow(*P, *_Fm(x), *_Em(xP_));
	return true;
}

/*!
	t@Cۑ
	@param name [in] Jt@C
*/
bool PublicKey::save(const char *name) const
{
	if (name == 0) return false;
	std::ofstream ofs(name, std::ios::binary);
	if (ofs) {
		const char *key = get();
		if (key) {
			int keySize = strlen(key);
			ofs.write(key, keySize);
			return ofs.good();
		}
	}
	return false;
}

bool PublicKey::load(const char *name)
{
	if (name == 0) return false;
	std::ifstream ifs(name, std::ios::binary);
	if (ifs) {
		std::string key;
		std::getline(ifs, key, '\0');
		return set(key.c_str());
	}
	return false;
}

const char *PublicKey::get() const
{
	if (P_ == 0 || xP_ == 0) return 0;
	std::ostringstream os;
	os << *_Em(P_) << *_Em(xP_);
	data_ = os.str();
	return data_.c_str();
}

bool PublicKey::set(const char *data)
{
	std::istringstream is(data);
	is >> *_Em(P_) >> *_Em(xP_);
	return true;
}

///////////////////////////////////////////////////////////////////////////////
/*
	time_t𕶎ɕϊ
*/
static bool ConvertTimeToString(char *data, int dataSize, std::time_t time)
{
	assert(data);
	const struct tm *tm = std::localtime(&time);
	if (tm == 0) return false;
	return std::strftime(data, dataSize, "%Y%m%d%H%M%S", tm) != 0;
}

/*
	strn܂strLen̐lminȏmaxȉȂ炻̒l(10i)Ԃ
	łȂꍇ-1Ԃ
*/
static int ConvertToNum(const char *str, int strLen, int min, int max)
{
	assert(str);
	int ret = 0;
	for (int i = 0; i < strLen; i++) {
		char c = str[i];
		if (c < '0' || c > '9') return -1;
		ret *= 10;
		ret += c - '0';
	}
	return (min <= ret && ret <= max) ? ret : -1;
}

static const time_t errTime = time_t(-1);

static time_t ConvertStringToTime(const char *str)
{
	/*
	    0    4  6  8  10 12
	    YYYY MM DD hh mm ss
	*/
	struct tm tm;
	memset(&tm, 0, sizeof(tm));
	int ret;
	tm.tm_isdst = 0; /* ĎԖ */
	ret = ConvertToNum(&str[0], 4, 1900, 2100);
	if (ret < 0) return errTime;
	tm.tm_year = ret - 1900; /* 0` (0 == 1900N) */

	ret = ConvertToNum(&str[4], 2, 1, 12);
	if (ret < 0) return errTime;
	tm.tm_mon =  ret - 1; /* 0`11(0 == 1) */

	ret = ConvertToNum(&str[6], 2, 1, 31);
	if (ret < 0) return errTime;
	tm.tm_mday = ret; /* 1`31 */

	ret = ConvertToNum(&str[8], 2, 0, 23);
	if (ret < 0) return errTime;
	tm.tm_hour = ret; /* 0`23 */

	ret = ConvertToNum(&str[10], 2, 0, 59);
	if (ret < 0) return errTime;
	tm.tm_min =  ret; /* 0`59 */

	ret = ConvertToNum(&str[12], 2, 0, 60);
	if (ret < 0) return errTime;
	tm.tm_sec =  ret; /* 0`60b */

	return mktime(&tm);
}

bool TimeInfo::set(const char *str)
{
	if (str == 0) return false;
	/* 擪"time="ȂXLbv */
	if (strncmp(str, "time=", 5) == 0) {
		str += 5; // 5 = sizeof("time=")
	}

	/* G[`FbN˂ */
	time_t time = ConvertStringToTime(str);
	if (time == errTime) return false;

	return ConvertTimeToString(data_, sizeof(data_), time);
}

/*!
	ݎɐݒ肷
*/
bool TimeInfo::setNow()
{
	time_t currTime;
	time(&currTime);

	return ConvertTimeToString(data_, sizeof(data_), currTime);
}

bool TimeInfo::isEqualTo(const TimeInfo& rhs) const
{
	return strcmp(data_, rhs.data_) == 0;
}

bool TimeInfo::isNewerThan(const TimeInfo& rhs, int plusSec) const
{
	time_t lhsTime, rhsTime;
	lhsTime = ConvertStringToTime(data_);
	rhsTime = ConvertStringToTime(rhs.data_);
	return lhsTime > rhsTime + plusSec;
}

///////////////////////////////////////////////////////////////////////////////
SignedTimeInfo::SignedTimeInfo(const char *str)
{
	if (str) set(str);
}

bool SignedTimeInfo::set(const char *str)
{
	if (str == 0) return false;
	data_ = str;
	return true;
}

bool SignedTimeInfo::getTimeInfo(TimeInfo *timeInfo) const
{
	return LoadTimeInfoAndEm(data_, timeInfo);
}

///////////////////////////////////////////////////////////////////////////////

void PrivateKey::deleteThis()
{
	deleteFm(&x_);
}

void PrivateKey::createThis()
{
	x_ = newFm();
}

bool PrivateKey::init(const char *seed, PublicKey *publicKey)
{
	Fm *x = _Fm(x_);

	getRandInteger(x, seed);
	return publicKey->init(seed, x_);
}

bool PrivateKey::sign(SignedTimeInfo *signedTimeInfo, const TimeInfo& timeInfo, const PublicKey& publicKey, Mode mode) const
{
	if (signedTimeInfo == 0) return false;
	if (x_ == 0) return false;

	if (mode == VerifyTime) {
		TimeInfo now;
		now.setNow();
		if (timeInfo.isNewerThan(now)) return false;
	}

	Fm t(timeInfo.get());
	t += *_Fm(x_);
	t.inverse();
	MttPower<ModP, ModM>::ECurve E;
	Em Q;
	E.pow(*_Em(publicKey.getP()), t, Q);
	return SaveTimeInfoAndEm(signedTimeInfo->data_, timeInfo, Q);
}

///////////////////////////////////////////////////////////////////////////////

/*
	(P, Q)̃yAǑʂgetR(x)vZdataɏo͂
	@param data [out] ZbV
	@param dataSize [in] ݉\TCY
	@param P [in]
	@param Q [in]
	@note yAOl̐Βl͕K1Ȃ̂getR()ŏ\
*/
static bool ComputePairing(std::string& data, const Em& P, const Em& Q)
{
	MttPower<ModP, ModM>::ECurve E;
	MttPower<ModP, ModM>::Fq F;

	Fq s;

	Em distortQ;
	E.distort(Q, distortQ);
	MttPairing<ModP, ModM>::tateFast(E, F, P, distortQ, s);

	/* Βl1ł邱ƂmF */
#if 0
	Fp a(s.getR()), b(s.getI());
	a *= a;
	b *= b;
	a += b;
	a -= Fp(1);
	a.normalize();
	assert(a.isZero());
#endif

	std::ostringstream os;
	os << "Fp=" << s.getR().toStr() << std::endl; /* getR()̂ݏo */
	data = os.str();
	return true;
}

const char* SessionKey::get() const
{
	/* "Fp=..."Ƃ`Ȃ̂+3 */
	return data_.c_str() + 3;
}

bool SessionKey::init(
	EncodedSessionKey *encodedSessionKey,
	const PublicKey& publicKey,
	const TimeInfo& timeInfo,
	const char* seed
)
{
	if (encodedSessionKey == 0 || seed == 0) return false;
	MttPower<ModP, ModM>::ECurve E;
	MttPower<ModP, ModM>::Fq F;

	Fm r;
	getRandInteger(&r, seed);

	/* r(xP + tP) */
	const Em *P = _Em(publicKey.getP());
	const Em *xP = _Em(publicKey.getxP());
	Em Q;
	Fm t(timeInfo.get());
	E.pow(*P, t, Q);
	E.add(Q, *xP, Q);
	E.pow(Q, r, Q); /* Q = r(xP + tP) */

	if (!SaveTimeInfoAndEm(encodedSessionKey->data_, timeInfo, Q)) {
		return false;
	}
	Em rP;
	E.pow(*P, r, rP);
	return ComputePairing(data_, rP, *P);
}

///////////////////////////////////////////////////////////////////////////////
EncodedSessionKey::EncodedSessionKey(const char *str)
{
	if (str) set(str);
}

bool EncodedSessionKey::set(const char *str)
{
	data_ = str;
	return true;
}

bool EncodedSessionKey::getTimeInfo(TimeInfo *timeInfo) const
{
	if (timeInfo == 0) return false;
	return LoadTimeInfoAndEm(data_, timeInfo);
}

bool EncodedSessionKey::decode(SessionKey *sessionKey, const SignedTimeInfo& signedTimeInfo, const PublicKey *publicKey) const
{
	if (sessionKey == 0) return false;
	Em Q; /* = r(xP + tP) */
	TimeInfo sessionTimeInfo;

	if (!LoadTimeInfoAndEm(data_, &sessionTimeInfo, &Q)) {
		return false;
	}

	Em R; /* = 1/(x + t) P */
	TimeInfo decTimeInfo;
	if (!LoadTimeInfoAndEm(signedTimeInfo.get(), &decTimeInfo, &R)) {
		return false;
	}

	/* ̈v`FbN */
	if (!sessionTimeInfo.isEqualTo(decTimeInfo)) {
		return false;
	}
	/*
		QQQ:F
		e(R, xP + tP) = e(P, P)mF
	*/
	if (publicKey) {
		const Em *P = _Em(publicKey->getP());
		const Em *xP = _Em(publicKey->getxP());
		Em S;
		Fm t(decTimeInfo.get());

		MttPower<ModP, ModM>::ECurve E;
		E.pow(*P, t, S); /* S = tP */
		E.add(S, *xP, S); /* S = xP + tP */

		Fq lhs, rhs;

		E.distort(S, S);
		MttPower<ModP, ModM>::Fq F;
		MttPairing<ModP, ModM>::tateFast(E, F, R, S, lhs);

		E.distort(*P, S);
		MttPairing<ModP, ModM>::tateFast(E, F, *P, S, rhs);
		lhs -= rhs;
		lhs.normalize();
		if (!lhs.isZero()) return false;
	}

	/* švZ */
	return ComputePairing(sessionKey->data_, Q, R);
}
